package com.nordicpeak.flowengine.alternativelist.provider;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.Breadcrumb;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.SystemStatus;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.hierarchy.core.interfaces.listeners.SystemStartupListener;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.CRUDCallback;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.MySQLRowLimiter;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.populators.PositiveStringIntegerPopulator;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.threads.ThreadUtils;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.validation.PositiveStringIntegerValidator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.alternativelist.AlternativeListHandlerModule;
import com.nordicpeak.flowengine.alternativelist.beans.AlternativeListDescriptor;
import com.nordicpeak.flowengine.alternativelist.exceptions.UnableToDeleteListInUseException;
import com.nordicpeak.flowengine.alternativelist.interfaces.AlternativeListInstance;
import com.nordicpeak.flowengine.alternativelist.interfaces.AlternativeListProvider;
import com.nordicpeak.flowengine.alternativelist.interfaces.AlternativeListQuery;
import com.nordicpeak.flowengine.interfaces.ImmutableFlow;
import com.nordicpeak.flowengine.interfaces.ImmutableStep;

public abstract class BaseAlternativeListProvider<BaseList extends BaseAlternativeList<BaseList, BaseListEvent>, ListInstance extends AlternativeListInstance, ListCRUD extends BaseAlternativeListCRUD<BaseList, ListInstance, ListCRUD, Provider, BaseListEvent>, Provider extends BaseAlternativeListProvider<BaseList, ListInstance, ListCRUD, Provider, BaseListEvent>, BaseListEvent extends BaseAlternativeListEvent<BaseList, BaseListEvent>> extends AnnotatedForegroundModule implements AlternativeListProvider, CRUDCallback<User>, SystemStartupListener, Runnable{
	
	public static final ValidationError LIST_IN_USE_VALIDATION_ERROR = new ValidationError("UnableToDeleteListInUse");
	
	@XSLVariable(prefix = "java.")
	private String eventListAddedMessage = "List added";

	@XSLVariable(prefix = "java.")
	private String eventListUpdatedMessage = "List updated";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Provider ID", description = "The unique identifier for this provider.", required = true)
	protected String providerID = getClass().getName();
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Recent events count", description = "Amount of list events to get for showList", required = true, formatValidator = PositiveStringIntegerValidator.class)
	protected Integer recentListEventsCount = 4;
	
	protected AnnotatedDAOWrapper<BaseList, Integer> listDAO;
	protected AnnotatedDAO<BaseListEvent> listEventDAO;
	
	protected QueryParameterFactory<BaseListEvent, BaseList> listEventListParamFactory;
	
	protected ListCRUD listCRUD;
	
	protected AlternativeListHandlerModule alternativeListHandlerModule;
	
	protected ConcurrentHashMap<Integer, ListInstance> alternativeListCacheMap = new ConcurrentHashMap<Integer, ListInstance>();
	
	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {
		
		super.init(moduleDescriptor, sectionInterface, dataSource);
		
		systemInterface.addSystemStartupListener(this);
	}
	
	@InstanceManagerDependency(required = true)
	public void setAlternativeListHandlerModule(AlternativeListHandlerModule alternativeListHandlerModule) {
		
		if (alternativeListHandlerModule != null) {
			
			alternativeListHandlerModule.addProvider(this);
			
		} else {
			
			this.alternativeListHandlerModule.removeProvider(this);
		}
		
		this.alternativeListHandlerModule = alternativeListHandlerModule;
	}
	
	@Override
	public void unload() throws Exception {
		
		if (alternativeListHandlerModule != null) {
			
			alternativeListHandlerModule.removeProvider(this);
		}
		
		super.unload();
	}
	
	public AlternativeListDescriptor getNewDescriptor(BaseList list){
		
		return new AlternativeListDescriptor(providerID, list.getListName());
	}
	
	protected abstract ListInstance getNewInstance(AlternativeListDescriptor descriptor, BaseList list);
	
	protected List<BaseList> getAllLists() throws SQLException {
		return listDAO.getAll();
	}
	
	protected void cacheAlternatives() throws SQLException {
		
		long startTime = System.currentTimeMillis();
		
		List<BaseList> lists = getAllLists();
		
		if (lists == null) {
			
			alternativeListCacheMap.clear();
			
		} else {
			
			for (BaseList list : lists) {
				
				if(systemInterface.getSystemStatus() != SystemStatus.STARTED) {
					
					return;
				}
				
				ListInstance listInstance = getNewInstance(getNewDescriptor(list), list);
				
				alternativeListCacheMap.put(list.getListID(), listInstance);
			}
			
			long end = System.currentTimeMillis() - startTime;
			
			log.info("Cached " + lists.size() + " lists in " + TimeUtils.millisecondsToString(end) + " ms");
		}
		
		if (alternativeListHandlerModule != null) {
			
			alternativeListHandlerModule.checkForListNameCollisions(this);
		}
	}

	public ListInstance cacheAlternatives(BaseList list) {
		
		ListInstance listInstance = getNewInstance(getNewDescriptor(list), list);
		
		alternativeListCacheMap.put(list.getListID(), listInstance);
		
		return listInstance;
	}
	
	public void unCacheList(BaseList list) {
		
		alternativeListCacheMap.remove(list.getListID());
	}
	
	public ListInstance getAlternativeInstance(BaseList list, AttributeHandler attributeHandler, User poster) {
		
		return alternativeListCacheMap.get(list.getListID());
	}
	
	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {
		
		return listCRUD.list(req, res, user, uriParser, null);
	}
	
	@WebPublic
	public ForegroundModuleResponse add(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {
		
		return listCRUD.add(req, res, user, uriParser);
	}
	
	@WebPublic
	public ForegroundModuleResponse show(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {
		
		return listCRUD.show(req, res, user, uriParser);
	}
	
	@WebPublic
	public ForegroundModuleResponse update(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {
		
		return listCRUD.update(req, res, user, uriParser);
	}
	
	@WebPublic
	public ForegroundModuleResponse delete(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {
		
		try {
			return listCRUD.delete(req, res, user, uriParser);
			
		} catch (UnableToDeleteListInUseException e){
			
			return listCRUD.show(req, res, user, uriParser, Collections.singletonList(LIST_IN_USE_VALIDATION_ERROR));
		}
	}
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((providerID == null) ? 0 : providerID.hashCode());
		return result;
	}
	
	@Override
	public boolean equals(AlternativeListProvider other) {
		if (this == other) {
			return true;
		}
		if (other == null) {
			return false;
		}
		if (getClass() != other.getClass()) {
			return false;
		}
		
		if (getProviderID() == null) {
			if (other.getProviderID() != null) {
				return false;
			}
		} else if (!getProviderID().equals(other.getProviderID())) {
			return false;
		}
		return true;
	}
	
	@Override
	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {
		
		Document doc = XMLUtils.createDomDocument();
		
		Element document = doc.createElement("Document");
		document.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		document.appendChild(this.moduleDescriptor.toXML(doc));
		document.appendChild(this.sectionInterface.getSectionDescriptor().toXML(doc));
		
		if (user != null) {
			document.appendChild(user.toXML(doc));
		}
		
		doc.appendChild(document);
		
		return doc;
	}
	
	@Override
	public String getTitlePrefix() {
		return getProviderName();
	}
	
	@Override
	public String getProviderName() {
		return moduleDescriptor.getName();
	}
	
	@Override
	public String getProviderID() {
		return providerID;
	}
	
	@Override
	public AlternativeListDescriptor getAlternativeList(String listName) {
		
		for (ListInstance listInstance : alternativeListCacheMap.values()) {
			
			if (listInstance.getDescriptor().getName().equals(listName)) {
				
				return listInstance.getDescriptor();
			}
		}
		
		return null;
	}
	
	public ListInstance getCachedInstance(String name) {
		
		for (ListInstance listInstance : alternativeListCacheMap.values()) {
			
			if (listInstance.getDescriptor().getName().equals(name)) {
				
				return listInstance;
			}
		}
		
		return null;
	}	
	
	@Override
	public List<AlternativeListDescriptor> getAlternativeLists() {
		
		if (alternativeListCacheMap.isEmpty()) {
			return null;
		}
		
		ArrayList<AlternativeListDescriptor> descriptors = new ArrayList<AlternativeListDescriptor>(alternativeListCacheMap.size());
		
		for (ListInstance listInstance : alternativeListCacheMap.values()) {
			
			descriptors.add(listInstance.getDescriptor());
		}
		
		Collections.sort(descriptors);
		
		return descriptors;
	}
	
	public Collection<ListInstance> getAlternativeInstances() {
		
		return alternativeListCacheMap.values();
	}
	
	public AlternativeListHandlerModule getAlternativeListHandlerModule() {

		return alternativeListHandlerModule;
	}
	
	protected abstract BaseListEvent getNewListEvent(BaseList list, Timestamp added, User poster, String message);
	
	public void addListEvent(String message, BaseList list, User user) {

		message = StringUtils.substring(message, 255, "...");

		BaseListEvent event = getNewListEvent(list, new Timestamp(System.currentTimeMillis()), user, message);

		try {
			listEventDAO.add(event);

		} catch (SQLException e) {
			log.error("Error adding new list event " + event, e);
		}
	}

	public String getEventListAddedMessage() {
		return eventListAddedMessage;
	}

	public String getEventListUpdatedMessage() {
		return eventListUpdatedMessage;
	}
	
	@WebPublic(alias = "usage")
	public ForegroundModuleResponse showListUsage(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Integer listID;
		BaseList list;

		if (uriParser.size() >= 3 && PositiveStringIntegerPopulator.getPopulator().validateFormat(uriParser.get(2)) && (listID = PositiveStringIntegerPopulator.getPopulator().getValue(uriParser.get(2))) != null && (list = listDAO.get(listID)) != null) {

			Document doc = createDocument(req, uriParser, user);
			Element showAlternativeListUsageElement = doc.createElement("ShowAlternativeListFlowUsage");
			doc.getDocumentElement().appendChild(showAlternativeListUsageElement);

			showAlternativeListUsageElement.appendChild(list.toXML(doc));
			
			List<AlternativeListQuery> queries = alternativeListHandlerModule.getQueriesUsingAlternativeList(getNewDescriptor(list));
			
			if (!CollectionUtils.isEmpty(queries)) {
				
				Element queriesElement = XMLUtils.appendNewElement(doc, showAlternativeListUsageElement, "Queries");
				
				Map<Integer, Element> flowElements = new HashMap<Integer, Element>();
				Map<Integer, Element> stepElements = new HashMap<Integer, Element>();
				
				for (AlternativeListQuery query : queries) {
					
					ImmutableFlow flow = query.getQueryDescriptor().getStep().getFlow();
					Element flowElement = flowElements.get(flow.getFlowID());
					
					if (flowElement == null) {
						
						flowElement = (Element) queriesElement.appendChild(flow.toXML(doc));
						flowElements.put(flow.getFlowID(), flowElement);
					}
					
					ImmutableStep step = query.getQueryDescriptor().getStep();
					Element stepElement = stepElements.get(step.getStepID());
					
					if (stepElement == null) {
						
						stepElement = (Element) flowElement.appendChild(step.toXML(doc));
						stepElements.put(step.getStepID(), stepElement);
					}
					
					Element queryElement = XMLUtils.appendNewElement(doc, stepElement, "Query");
					
					XMLUtils.appendNewElement(doc, queryElement, "ListName", query.getAlternativeListName());
					XMLUtils.appendNewElement(doc, queryElement, "ExtraUsageInfo", query.getExtraUsageInfo());
					queryElement.appendChild(query.getQueryDescriptor().toXML(doc));
				}
			}

			return new SimpleForegroundModuleResponse(doc, getTitlePrefix(), getDefaultBreadcrumb(), new Breadcrumb(this, list.getListName(), "/show/" + list.getListID()));
		}

		throw new URINotFoundException(uriParser);
	}
	
	public boolean isListInUse(BaseList list) throws Exception {
	
		return alternativeListHandlerModule.hasQueriesUsingAlternativeList(getNewDescriptor(list));
	}
	
	public List<BaseListEvent> getListEvents(BaseList list) throws SQLException {

		HighLevelQuery<BaseListEvent> query = new HighLevelQuery<BaseListEvent>();

		query.addParameter(listEventListParamFactory.getParameter(list));

		return listEventDAO.getAll(query);
	}
	
	public List<BaseListEvent> getRecentListEvents(BaseList list) throws SQLException {

		HighLevelQuery<BaseListEvent> query = new HighLevelQuery<BaseListEvent>();

		query.addParameter(listEventListParamFactory.getParameter(list));
		query.setRowLimiter(new MySQLRowLimiter(recentListEventsCount));

		return listEventDAO.getAll(query);
	}

	@WebPublic(alias = "events")
	public ForegroundModuleResponse showListEvents(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Integer listID;
		BaseList list;

		if (uriParser.size() >= 3 && PositiveStringIntegerPopulator.getPopulator().validateFormat(uriParser.get(2)) && (listID = PositiveStringIntegerPopulator.getPopulator().getValue(uriParser.get(2))) != null && (list = listDAO.get(listID)) != null) {

			Document doc = createDocument(req, uriParser, user);
			Element showAlternativeListEventsElement = doc.createElement("ShowAlternativeListEvents");
			doc.getDocumentElement().appendChild(showAlternativeListEventsElement);

			showAlternativeListEventsElement.appendChild(list.toXML(doc));
			XMLUtils.append(doc, showAlternativeListEventsElement, "Events", getListEvents(list));

			return new SimpleForegroundModuleResponse(doc, getTitlePrefix(), getDefaultBreadcrumb(), new Breadcrumb(this, list.getListName(), "/show/" + list.getListID()));
		}

		throw new URINotFoundException(uriParser);
	}

	@Override
	public void systemStarted() throws Exception {

		ThreadUtils.run(this, moduleDescriptor.toString() + " alternative cache thread", false);
	}
	
	@Override
	public void run(){
		
		try {
			cacheAlternatives();
			
		} catch (Exception e) {

			log.error("Error caching alternatives", e);
		}
	}
	
}
