package com.nordicpeak.flowengine.booking;

import java.io.IOException;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.json.JsonUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.webutils.http.HTTPUtils;

import com.nordicpeak.flowengine.booking.beans.BookingCalendar;

public class ManagerRequestProcessor {

	protected final Logger log = Logger.getLogger(this.getClass());

	private final String encoding;
	private final BookingSlotAdminModule bookingSlotAdminModule;

	public ManagerRequestProcessor(BookingSlotAdminModule bookingSlotAdminModule, String encoding) {

		this.bookingSlotAdminModule = bookingSlotAdminModule;
		this.encoding = encoding;
	}

	private static void sendEmptyJSONResponse(HttpServletResponse res) throws IOException {

		JsonObject jsonObject = new JsonObject(1);

		jsonObject.putField("hitCount", "0");

		HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);
	}

	private void sendJSONResponse(JsonArray jsonArray, HttpServletResponse res) throws IOException {

		JsonObject jsonObject = new JsonObject(2);

		jsonObject.putField("hitCount", Integer.toString(jsonArray.size()));
		jsonObject.putField("hits", jsonArray);

		HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);
	}

	public void processRequest(BookingCalendar calendar, HttpServletRequest req, HttpServletResponse res, User user) throws Exception {

		String searchString = req.getParameter("q");

		if (StringUtils.isEmpty(searchString)) {

			sendEmptyJSONResponse(res);
			return;
		}

		if (!encoding.equalsIgnoreCase("UTF-8")) {

			searchString = URLDecoder.decode(searchString, "UTF-8");
		}

		log.info("User " + user + " searching for managers of calendar " + calendar + " with search string: " + searchString);

		List<User> users = searchUsers(calendar, searchString);

		if (users == null) {

			sendEmptyJSONResponse(res);
			return;
		}

		JsonArray jsonArray = new JsonArray();

		for (User currentUser : users) {

			JsonObject instance = new JsonObject(4);
			instance.putField("ID", currentUser.getUserID());
			instance.putField("Name", currentUser.getFirstname() + " " + currentUser.getLastname());
			instance.putField("Enabled", currentUser.isEnabled());
			instance.putField("Username", currentUser.getUsername());

			AttributeHandler attributeHandler = currentUser.getAttributeHandler();

			if (attributeHandler != null && attributeHandler.isSet(BookingConstants.MANAGER_NAME_ATTRIBUTE)) {

				instance.putField("Email", attributeHandler.getString(BookingConstants.MANAGER_NAME_ATTRIBUTE));
			}

			jsonArray.addNode(instance);
		}

		sendJSONResponse(jsonArray, res);
	}

	private List<User> searchUsers(BookingCalendar calendar, String searchString) {

		Map<Integer, User> managingUsers = bookingSlotAdminModule.getAllowedManagers(calendar);

		if (CollectionUtils.isEmpty(managingUsers)) {
			return null;
		}

		String[] terms = searchString.toLowerCase().split("[ ]+");

		List<User> matchingUsers = new ArrayList<>(managingUsers.size());

		for (User potentialUser : managingUsers.values()) {

			List<String> fields = new ArrayList<>();

			if (potentialUser.getFirstname() != null) {

				fields.add(potentialUser.getFirstname().toLowerCase());
			}

			if (potentialUser.getLastname() != null) {

				fields.add(potentialUser.getLastname().toLowerCase());
			}

			if (potentialUser.getUsername() != null) {

				fields.add(potentialUser.getUsername().toLowerCase());
			}

			AttributeHandler attributeHandler = potentialUser.getAttributeHandler();

			if (attributeHandler != null && attributeHandler.isSet(BookingConstants.MANAGER_NAME_ATTRIBUTE)) {

				fields.add(attributeHandler.getString(BookingConstants.MANAGER_NAME_ATTRIBUTE).toLowerCase());
			}

			int matches = 0;

			for (String term : terms) {

				if (CollectionUtils.find(fields, field -> field.contains(term)) != null) {

					matches++;
				}
			}

			if (matches == terms.length) {

				matchingUsers.add(potentialUser);
			}
		}

		return matchingUsers;
	}

}
