package com.nordicpeak.flowengine.booking.beans;

import java.io.Serializable;
import java.lang.reflect.Field;
import java.sql.Time;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Objects;

import se.unlogic.emailutils.populators.EmailPopulator;
import se.unlogic.hierarchy.core.interfaces.AccessInterface;
import se.unlogic.standardutils.annotations.NoDuplicates;
import se.unlogic.standardutils.annotations.RequiredIfSet;
import se.unlogic.standardutils.annotations.SplitOnLineBreak;
import se.unlogic.standardutils.annotations.Templated;
import se.unlogic.standardutils.annotations.WebPopulate;
import se.unlogic.standardutils.collections.CaseInsensitiveStringComparator;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.annotations.DAOManaged;
import se.unlogic.standardutils.dao.annotations.Key;
import se.unlogic.standardutils.dao.annotations.OneToMany;
import se.unlogic.standardutils.dao.annotations.OrderBy;
import se.unlogic.standardutils.dao.annotations.SimplifiedRelation;
import se.unlogic.standardutils.dao.annotations.Table;
import se.unlogic.standardutils.enums.DayOfWeek;
import se.unlogic.standardutils.populators.IntegerPopulator;
import se.unlogic.standardutils.populators.NonNegativeStringIntegerPopulator;
import se.unlogic.standardutils.populators.PositiveStringIntegerPopulator;
import se.unlogic.standardutils.populators.TimePopulator;
import se.unlogic.standardutils.reflection.ReflectionUtils;
import se.unlogic.standardutils.string.StringTag;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.TimeStringyfier;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.GeneratedElementable;
import se.unlogic.standardutils.xml.XMLElement;
import se.unlogic.standardutils.xml.XMLParser;

import com.nordicpeak.flowengine.booking.enums.BookingCalendarViewMode;
import com.nordicpeak.flowengine.booking.populators.AtLeastTwoStringIntegerPopulator;
import com.nordicpeak.flowengine.booking.populators.PopulateByXMLParser;
import com.nordicpeak.flowengine.booking.populators.annotations.PopulateByXMLSettings;
import com.nordicpeak.flowengine.booking.queries.BookingQuery;

@Table(name = "booking_calendars")
@XMLElement(name = "Calendar")
public class BookingCalendar extends GeneratedElementable implements AccessInterface, Serializable {

	private static final long serialVersionUID = 4175567992506321877L;

	public static final Field MANAGER_GROUPS_RELATION = ReflectionUtils.getField(BookingCalendar.class, "managerGroupIDs");
	public static final Field MANAGER_USERS_RELATION = ReflectionUtils.getField(BookingCalendar.class, "managerUserIDs");
	public static final Field BOOKING_SLOTS_RELATION = ReflectionUtils.getField(BookingCalendar.class, "bookingSlots");
	public static final Field ADMIN_GROUPS_RELATION = ReflectionUtils.getField(BookingCalendar.class, "adminGroupIDs");

	private static final Field[] FIELDS_TO_POPULATE = BookingCalendar.class.getDeclaredFields();

	@DAOManaged(autoGenerated = true)
	@Key
	@XMLElement
	private Integer calendarID;

	@DAOManaged
	@WebPopulate(required = true)
	@OrderBy
	@StringTag
	@XMLElement
	@PopulateByXMLSettings(required = true)
	private String name;

	@DAOManaged
	@WebPopulate(required = true, maxLength = 65534)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65534, required = true)
	private String description;

	@DAOManaged
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String emailSenderName;

	@DAOManaged
	@WebPopulate(maxLength = 255, populator = EmailPopulator.class)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255, populator = EmailPopulator.class)
	private String emailSenderAddress;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean enabled;

	@DAOManaged
	@WebPopulate(required = true)
	@XMLElement
	@PopulateByXMLSettings(required = true)
	private BookingCalendarViewMode viewMode = BookingCalendarViewMode.CALENDAR;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean calendarWeekView;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean showDuration;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean showManagers;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean showBookings = true;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean requireConfirmation;

	@DAOManaged
	@WebPopulate(required = true)
	@XMLElement
	@PopulateByXMLSettings(required = true)
	private Integer price;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private Integer defaultSlotLength;

	@DAOManaged
	@WebPopulate(required = true)
	@XMLElement
	@PopulateByXMLSettings(required = true)
	private Integer hoursBeforeBookingLocked;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean disableLeaving;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean disableRebooking;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean hideAddedDateFromUser;

	@DAOManaged
	@WebPopulate(populator = PositiveStringIntegerPopulator.class)
	@XMLElement
	@PopulateByXMLSettings(populator = PositiveStringIntegerPopulator.class)
	private Integer minVisibleDaysRebooking;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean restrictRebookingToPreviousManager;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean restrictRebookingToCurrentBookingSize;

	@DAOManaged
	@WebPopulate(populator = NonNegativeStringIntegerPopulator.class)
	@RequiredIfSet(paramNames = "useReminder")
	@XMLElement
	@PopulateByXMLSettings(populator = NonNegativeStringIntegerPopulator.class)
	private Integer reminderMinutes;

	@DAOManaged
	@WebPopulate(populator = TimePopulator.class)
	@XMLElement(valueFormatter = TimeStringyfier.class)
	@PopulateByXMLSettings(populator = TimePopulator.class)
	private Time defaultStartTime;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean allowMultipleBookings = true;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean allowDuplicateBookings;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean bookingSizeSelectorEnabled;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean allowUserBookingSizeSelection;

	@DAOManaged
	@WebPopulate(populator = AtLeastTwoStringIntegerPopulator.class)
	@XMLElement
	@PopulateByXMLSettings(populator = AtLeastTwoStringIntegerPopulator.class)
	protected Integer userBookingMaxSize;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean allowPreliminaryBookings;

	@DAOManaged
	@WebPopulate(populator = PositiveStringIntegerPopulator.class)
	@RequiredIfSet(paramNames = "allowPreliminaryBookings")
	@XMLElement
	@PopulateByXMLSettings(populator = PositiveStringIntegerPopulator.class)
	protected Integer hoursPreliminaryBookingsAreValidFor;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendEnteredUserEmail;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendEnteredUserSMS;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendEnteredManagerEmail;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendExternalMessageReceivedUserSMS;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendExternalMessageReceivedUserEmail;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendExternalMessageReceivedManagerEmail;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendExternalMessageReceivedGlobalEmail;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendManualReminderSMS;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendRebookedUserSMS;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendRebookedUserEmail;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendUserRebookedManagerEmail = true;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendUserRebookedManagerEmailOnlyToBookingSlotManagers = false;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String manualEntranceSMS;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String externalMessageReceivedUserSMS;

	@DAOManaged
	@Templated(fieldName = "exitUserSMS")
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String userExitSMS;

	@DAOManaged
	@Templated(fieldName = "reminderSMS")
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String userReminderSMS;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String manualReminderSMS;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String manualEntranceEmailSubject;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String manualEntranceEmailMessage;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String manualEntranceManagerEmailSubject;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String manualEntranceManagerEmailMessage;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String externalMessageReceivedUserEmailSubject;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String externalMessageReceivedUserEmailMessage;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String manualReminderEmailSubject;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String manualReminderEmailMessage;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String rebookedUserSMS;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String rebookedUserEmailSubject;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String rebookedUserEmailMessage;

	@DAOManaged
	@Templated(fieldName = "exitUserEmailSubject")
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String userExitEmailSubject;

	@DAOManaged
	@Templated(fieldName = "exitUserEmailMessage")
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String userExitEmailMessage;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String externalMessageReceivedManagerEmailSubject;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String externalMessageReceivedManagerEmailMessage;

	@DAOManaged
	@Templated(fieldName = "exitManagerEmailSubject")
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String userExitManagerEmailSubject;

	@DAOManaged
	@Templated(fieldName = "exitManagerEmailMessage")
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String userExitManagerEmailMessage;

	@DAOManaged
	@Templated(fieldName = "rebookedManagerEmailSubject")
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String userRebookedManagerEmailSubject;

	@DAOManaged
	@Templated(fieldName = "rebookedManagerEmailMessage")
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String userRebookedManagerEmailMessage;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String externalMessageReceivedGlobalEmailSubject;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String externalMessageReceivedGlobalEmailMessage;

	@DAOManaged
	@OneToMany(autoAdd = true, autoGet = true, autoUpdate = true)
	@SimplifiedRelation(table = "booking_calendar_extmessage_global", remoteValueColumnName = "email")
	@WebPopulate(maxLength = 255, populator = EmailPopulator.class)
	@RequiredIfSet(paramNames = "sendExternalMessageReceivedGlobalEmail")
	@SplitOnLineBreak
	@XMLElement(fixCase = true, childName = "address")
	@PopulateByXMLSettings(maxLength = 255, populator = EmailPopulator.class, nodeName = "ExternalMessageReceivedGlobalEmailAddresses/address")
	private List<String> externalMessageReceivedGlobalEmailAddresses;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendExitUserSMS;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendExitManagerEmail;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendExitGlobalEmail;

	@DAOManaged
	@Templated(fieldName = "exitGlobalEmailSubject")
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String userExitGlobalEmailSubject;

	@DAOManaged
	@Templated(fieldName = "exitGlobalEmailMessage")
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String userExitGlobalEmailMessage;

	@DAOManaged
	@OneToMany(autoAdd = true, autoGet = true, autoUpdate = true)
	@SimplifiedRelation(table = "booking_calendar_exit_global", remoteValueColumnName = "email")
	@WebPopulate(maxLength = 255, populator = EmailPopulator.class)
	@RequiredIfSet(paramNames = "sendExitRequestGlobalEmail")
	@SplitOnLineBreak
	@XMLElement(fixCase = true, childName = "address")
	@PopulateByXMLSettings(maxLength = 255, populator = EmailPopulator.class, nodeName = "ExitGlobalEmailAddresses/address")
	private List<String> exitGlobalEmailAddresses;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendRebookedGlobalEmail;

	@DAOManaged
	@Templated(fieldName = "rebookedGlobalEmailSubject")
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String rebookedGlobalEmailSubject;

	@DAOManaged
	@Templated(fieldName = "rebookedGlobalEmailMessage")
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String rebookedGlobalEmailMessage;

	@DAOManaged
	@OneToMany(autoAdd = true, autoGet = true, autoUpdate = true)
	@SimplifiedRelation(table = "booking_calendar_rebooked_global", remoteValueColumnName = "email")
	@WebPopulate(maxLength = 255, populator = EmailPopulator.class)
	@RequiredIfSet(paramNames = "sendRebookedGlobalEmail")
	@SplitOnLineBreak
	@XMLElement(fixCase = true, childName = "address")
	@PopulateByXMLSettings(maxLength = 255, populator = EmailPopulator.class, nodeName = "RebookedGlobalEmailAddresses/address")
	private List<String> rebookedGlobalEmailAddresses;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendBookingSlotRebookedUserSMS;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String bookingSlotRebookedUserSMS;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendBookingSlotRebookedUserEmail;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String bookingSlotRebookedUserEmailSubject;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String bookingSlotRebookedUserEmailMessage;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendBookingSlotDeletedUserSMS;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String bookingSlotDeletedUserSMS;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean sendBookingSlotDeletedUserEmail;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 255)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String bookingSlotDeletedUserEmailSubject;

	@DAOManaged
	@Templated
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String bookingSlotDeletedUserEmailMessage;

	@DAOManaged
	@OneToMany
	@XMLElement(fixCase = true)
	private List<BookingSlot> bookingSlots;

	@DAOManaged
	@OneToMany
	@XMLElement(fixCase = true)
	private List<Repeating> repeatingSlots;

	@DAOManaged
	@OneToMany
	@XMLElement(fixCase = true)
	private List<BookingQuery> bookingQueries;

	@DAOManaged
	@OneToMany
	@SimplifiedRelation(table = "booking_calendar_manager_users", remoteValueColumnName = "userID")
	@WebPopulate(paramName = "user")
	@XMLElement(childName = "userID")
	private List<Integer> managerUserIDs;

	@DAOManaged
	@OneToMany
	@SimplifiedRelation(table = "booking_calendar_manager_groups", remoteValueColumnName = "groupID")
	@WebPopulate(paramName = "group")
	@XMLElement(childName = "groupID")
	private List<Integer> managerGroupIDs;

	@DAOManaged
	@OneToMany
	@SimplifiedRelation(table = "booking_calendar_admin_groups", remoteValueColumnName = "groupID")
	@WebPopulate(paramName = "adminGroup")
	@XMLElement(childName = "adminGroupID")
	private List<Integer> adminGroupIDs;

	@XMLElement
	@PopulateByXMLSettings(populator = IntegerPopulator.class)
	private Integer activeBookings;

	@XMLElement
	@PopulateByXMLSettings(populator = IntegerPopulator.class)
	private Integer bookingSlotCount;

	@DAOManaged
	@OneToMany(autoAdd = true, autoGet = true, autoUpdate = true)
	@SimplifiedRelation(table = "booking_calendar_export_attributes", remoteValueColumnName = "exportAttribute", preserveListOrder = true, indexColumn = "sortIndex")
	@WebPopulate(maxLength = 255)
	@NoDuplicates(comparator = CaseInsensitiveStringComparator.class)
	@SplitOnLineBreak
	@XMLElement(fixCase = true, childName = "exportAttribute")
	@PopulateByXMLSettings(maxLength = 255, nodeName = "ExportAttributes/exportAttribute")
	private List<String> exportAttributes;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean useCalendarForInternalUsers;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings(populator = IntegerPopulator.class)
	private Integer profileID;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean showFreeSlotsAsBoolean;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean showFullyBookedDaysInCalendar;

	@DAOManaged
	@WebPopulate(maxLength = 30)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 30)
	private String fullyBookedCalendarDayText;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean showFullyBookedSlotsInList;

	@DAOManaged
	@WebPopulate(maxLength = 30)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 30)
	private String fullyBookedListSlotText;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean hideBookingsPerWeekday;

	@DAOManaged
	@OneToMany(autoAdd = true, autoGet = true, autoUpdate = true)
	@SimplifiedRelation(table = "booking_calendar_weekdays_with_hidden_bookings", remoteValueColumnName = "dayOfWeek")
	@WebPopulate
	@XMLElement(childName = "dayOfWeek")
	@PopulateByXMLSettings(nodeName = "weekdaysWithHiddenBookings/dayOfWeek")
	private List<DayOfWeek> weekdaysWithHiddenBookings;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean showNotBookableSlots;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean hideDisabledManagersInFilter;

	@DAOManaged
	@WebPopulate
	@XMLElement
	@PopulateByXMLSettings
	private boolean hideExternalMessages;

	@DAOManaged
	@WebPopulate(populator = PositiveStringIntegerPopulator.class)
	@XMLElement
	@PopulateByXMLSettings(populator = PositiveStringIntegerPopulator.class)
	protected Integer newExternalMessagesAllowedDays;

	@DAOManaged
	@WebPopulate(maxLength = 65535)
	@XMLElement
	@PopulateByXMLSettings(maxLength = 65535)
	private String bookingSlotDefaultDescription;

	@DAOManaged
	@WebPopulate(maxLength = 255)
	@StringTag
	@XMLElement
	@PopulateByXMLSettings(maxLength = 255)
	private String bookingSlotDefaultLocation;

	@Override
	public boolean allowsAdminAccess() {

		return false;
	}

	@Override
	public boolean allowsUserAccess() {

		return false;
	}

	@Override
	public boolean allowsAnonymousAccess() {

		return false;
	}

	@Override
	public Collection<Integer> getAllowedGroupIDs() {

		if (managerGroupIDs == null) {

			return adminGroupIDs;
		}

		if (adminGroupIDs == null) {

			return managerGroupIDs;
		}

		return CollectionUtils.combine(managerGroupIDs, adminGroupIDs);
	}

	@Override
	public Collection<Integer> getAllowedUserIDs() {

		return getManagerUserIDs();
	}

	public BookingCalendarViewMode getViewMode() {

		return viewMode;
	}

	public void setViewMode(BookingCalendarViewMode viewMode) {

		this.viewMode = viewMode;
	}

	public boolean isCalendarWeekView() {

		return calendarWeekView;
	}

	public void setCalendarWeekView(boolean calendarWeekView) {

		this.calendarWeekView = calendarWeekView;
	}

	public List<Repeating> getRepeatingSlots() {

		return repeatingSlots;
	}

	public void setRepeatingSlots(List<Repeating> repeatingSlots) {

		this.repeatingSlots = repeatingSlots;
	}

	public Integer getCalendarID() {

		return calendarID;
	}

	public void setCalendarID(Integer bookingID) {

		this.calendarID = bookingID;
	}

	public String getName() {

		return name;
	}

	public void setName(String name) {

		this.name = name;
	}

	public String getDescription() {

		return description;
	}

	public void setDescription(String description) {

		this.description = description;
	}

	public boolean isEnabled() {

		return enabled;
	}

	public void setEnabled(boolean enabled) {

		this.enabled = enabled;
	}

	public boolean isShowDuration() {

		return showDuration;
	}

	public void setShowDuration(boolean showDuration) {

		this.showDuration = showDuration;
	}

	public boolean isShowManagers() {

		return showManagers;
	}

	public void setShowManagers(boolean showManagers) {

		this.showManagers = showManagers;
	}

	public boolean isShowBookings() {

		return showBookings;
	}

	public void setShowBookings(boolean showBookings) {

		this.showBookings = showBookings;
	}

	public Integer getPrice() {

		return price;
	}

	public void setPrice(Integer price) {

		this.price = price;
	}

	public Integer getHoursBeforeBookingLocked() {

		return hoursBeforeBookingLocked;
	}

	public void setHoursBeforeBookingLocked(Integer hoursBeforeBookingLocked) {

		this.hoursBeforeBookingLocked = hoursBeforeBookingLocked;
	}

	public List<BookingQuery> getBookingQueries() {

		return bookingQueries;
	}

	public void setBookingQueries(List<BookingQuery> bookingQueries) {

		this.bookingQueries = bookingQueries;
	}

	public List<BookingSlot> getBookingSlots() {

		return bookingSlots;
	}

	public void setBookingSlots(List<BookingSlot> bookingSlots) {

		this.bookingSlots = bookingSlots;
	}

	public List<Integer> getManagerGroupIDs() {

		return managerGroupIDs;
	}

	public void setManagerGroupIDs(List<Integer> managerGroupIDs) {

		this.managerGroupIDs = managerGroupIDs;
	}

	public List<Integer> getManagerUserIDs() {

		return managerUserIDs;
	}

	public void setManagerUserIDs(List<Integer> managerUserIDs) {

		this.managerUserIDs = managerUserIDs;
	}

	public List<Integer> getAdminGroupIDs() {

		return adminGroupIDs;
	}

	public void setAdminGroupIDs(List<Integer> adminGroupIDs) {

		this.adminGroupIDs = adminGroupIDs;
	}

	public String getEmailSenderName() {

		return emailSenderName;
	}

	public void setEmailSenderName(String emailSenderName) {

		this.emailSenderName = emailSenderName;
	}

	public String getEmailSenderAddress() {

		return emailSenderAddress;
	}

	public void setEmailSenderAddress(String emailSenderAddress) {

		this.emailSenderAddress = emailSenderAddress;
	}

	public Integer getDefaultSlotLength() {

		return defaultSlotLength;
	}

	public void setDefaultSlotLength(Integer defaultSlotLength) {

		this.defaultSlotLength = defaultSlotLength;
	}

	@Override
	public int hashCode() {

		return Objects.hash(calendarID);
	}

	@Override
	public boolean equals(Object obj) {

		if (this == obj) {
			return true;
		}
		if (!(obj instanceof BookingCalendar)) {
			return false;
		}
		BookingCalendar other = (BookingCalendar) obj;
		return Objects.equals(calendarID, other.calendarID);
	}

	@Override
	public String toString() {

		return StringUtils.toLogFormat(name, 30) + " (calendarID: " + calendarID + ")";
	}

	public boolean isSendExternalMessageReceivedUserSMS() {

		return sendExternalMessageReceivedUserSMS;
	}

	public void setSendExternalMessageReceivedUserSMS(boolean sendExternalMessageReceivedUserSMS) {

		this.sendExternalMessageReceivedUserSMS = sendExternalMessageReceivedUserSMS;
	}

	public boolean isSendExternalMessageReceivedUserEmail() {

		return sendExternalMessageReceivedUserEmail;
	}

	public void setSendExternalMessageReceivedUserEmail(boolean sendExternalMessageReceivedUserEmail) {

		this.sendExternalMessageReceivedUserEmail = sendExternalMessageReceivedUserEmail;
	}

	public boolean isSendExternalMessageReceivedManagerEmail() {

		return sendExternalMessageReceivedManagerEmail;
	}

	public void setSendExternalMessageReceivedManagerEmail(boolean sendExternalMessageReceivedManagerEmail) {

		this.sendExternalMessageReceivedManagerEmail = sendExternalMessageReceivedManagerEmail;
	}

	public boolean isSendExternalMessageReceivedGlobalEmail() {

		return sendExternalMessageReceivedGlobalEmail;
	}

	public void setSendExternalMessageReceivedGlobalEmail(boolean sendExternalMessageReceivedGlobalEmail) {

		this.sendExternalMessageReceivedGlobalEmail = sendExternalMessageReceivedGlobalEmail;
	}

	public List<String> getExternalMessageReceivedGlobalEmailAddresses() {

		return externalMessageReceivedGlobalEmailAddresses;
	}

	public void setExternalMessageReceivedGlobalEmailAddresses(List<String> externalMessageReceivedGlobalEmailAddresses) {

		this.externalMessageReceivedGlobalEmailAddresses = externalMessageReceivedGlobalEmailAddresses;
	}

	public boolean isSendEnteredUserEmail() {

		return sendEnteredUserEmail;
	}

	public void setSendEnteredUserEmail(boolean sendEnteredUserEmail) {

		this.sendEnteredUserEmail = sendEnteredUserEmail;
	}

	public boolean isSendEnteredUserSMS() {

		return sendEnteredUserSMS;
	}

	public void setSendEnteredUserSMS(boolean sendEnteredUserSMS) {

		this.sendEnteredUserSMS = sendEnteredUserSMS;
	}

	public boolean isSendEnteredManagerEmail() {

		return sendEnteredManagerEmail;
	}

	public void setSendEnteredManagerEmail(boolean sendEnteredManagerEmail) {

		this.sendEnteredManagerEmail = sendEnteredManagerEmail;
	}

	public boolean isSendExitUserSMS() {

		return sendExitUserSMS;
	}

	public void setSendExitUserSMS(boolean sendExitConfirmationUserSMS) {

		this.sendExitUserSMS = sendExitConfirmationUserSMS;
	}

	public boolean isSendExitManagerEmail() {

		return sendExitManagerEmail;
	}

	public void setSendExitManagerEmail(boolean sendExitManagerEmail) {

		this.sendExitManagerEmail = sendExitManagerEmail;
	}

	public boolean isSendExitGlobalEmail() {

		return sendExitGlobalEmail;
	}

	public void setSendExitGlobalEmail(boolean sendExitGlobalEmail) {

		this.sendExitGlobalEmail = sendExitGlobalEmail;
	}

	public List<String> getExitGlobalEmailAddresses() {

		return exitGlobalEmailAddresses;
	}

	public void setExitGlobalEmailAddresses(List<String> exitGlobalEmailAddresses) {

		this.exitGlobalEmailAddresses = exitGlobalEmailAddresses;
	}

	public Integer getActiveBookings() {

		return activeBookings;
	}

	public void setActiveBookings(Integer activeBookings) {

		this.activeBookings = activeBookings;
	}

	public Integer getBookingSlotCount() {

		return bookingSlotCount;
	}

	public void setBookingSlotCount(Integer bookingSlotCount) {

		this.bookingSlotCount = bookingSlotCount;
	}

	public boolean isRequireConfirmation() {

		return requireConfirmation;
	}

	public void setRequireConfirmation(boolean requireConfirmation) {

		this.requireConfirmation = requireConfirmation;
	}

	public Integer getReminderMinutes() {

		return reminderMinutes;
	}

	public void setReminderMinutes(Integer reminderMinutes) {

		this.reminderMinutes = reminderMinutes;
	}

	public String getManualEntranceEmailSubject() {

		return manualEntranceEmailSubject;
	}

	public void setManualEntranceEmailSubject(String manualEntranceEmailSubject) {

		this.manualEntranceEmailSubject = manualEntranceEmailSubject;
	}

	public String getManualEntranceEmailMessage() {

		return manualEntranceEmailMessage;
	}

	public void setManualEntranceEmailMessage(String manualEntranceEmailMessage) {

		this.manualEntranceEmailMessage = manualEntranceEmailMessage;
	}

	public String getManualEntranceManagerEmailSubject() {

		return manualEntranceManagerEmailSubject;
	}

	public void setManualEntranceManagerEmailSubject(String manualEntranceManagerEmailSubject) {

		this.manualEntranceManagerEmailSubject = manualEntranceManagerEmailSubject;
	}

	public String getManualEntranceManagerEmailMessage() {

		return manualEntranceManagerEmailMessage;
	}

	public void setManualEntranceManagerEmailMessage(String manualEntranceManagerEmailMessage) {

		this.manualEntranceManagerEmailMessage = manualEntranceManagerEmailMessage;
	}

	public String getExternalMessageReceivedUserEmailSubject() {

		return externalMessageReceivedUserEmailSubject;
	}

	public void setExternalMessageReceivedUserEmailSubject(String externalMessageReceivedUserEmailSubject) {

		this.externalMessageReceivedUserEmailSubject = externalMessageReceivedUserEmailSubject;
	}

	public String getExternalMessageReceivedUserEmailMessage() {

		return externalMessageReceivedUserEmailMessage;
	}

	public void setExternalMessageReceivedUserEmailMessage(String externalMessageReceivedUserEmailMessage) {

		this.externalMessageReceivedUserEmailMessage = externalMessageReceivedUserEmailMessage;
	}

	public String getExternalMessageReceivedManagerEmailSubject() {

		return externalMessageReceivedManagerEmailSubject;
	}

	public void setExternalMessageReceivedManagerEmailSubject(String externalMessageReceivedManagerEmailSubject) {

		this.externalMessageReceivedManagerEmailSubject = externalMessageReceivedManagerEmailSubject;
	}

	public String getExternalMessageReceivedManagerEmailMessage() {

		return externalMessageReceivedManagerEmailMessage;
	}

	public void setExternalMessageReceivedManagerEmailMessage(String externalMessageReceivedManagerEmailMessage) {

		this.externalMessageReceivedManagerEmailMessage = externalMessageReceivedManagerEmailMessage;
	}

	public String getUserExitManagerEmailSubject() {

		return userExitManagerEmailSubject;
	}

	public void setUserExitManagerEmailSubject(String userExitManagerEmailSubject) {

		this.userExitManagerEmailSubject = userExitManagerEmailSubject;
	}

	public String getUserExitManagerEmailMessage() {

		return userExitManagerEmailMessage;
	}

	public void setUserExitManagerEmailMessage(String userExitManagerEmailMessage) {

		this.userExitManagerEmailMessage = userExitManagerEmailMessage;
	}

	public String getExternalMessageReceivedGlobalEmailSubject() {

		return externalMessageReceivedGlobalEmailSubject;
	}

	public void setExternalMessageReceivedGlobalEmailSubject(String externalMessageReceivedGlobalEmailSubject) {

		this.externalMessageReceivedGlobalEmailSubject = externalMessageReceivedGlobalEmailSubject;
	}

	public String getExternalMessageReceivedGlobalEmailMessage() {

		return externalMessageReceivedGlobalEmailMessage;
	}

	public void setExternalMessageReceivedGlobalEmailMessage(String externalMessageReceivedGlobalEmailMessage) {

		this.externalMessageReceivedGlobalEmailMessage = externalMessageReceivedGlobalEmailMessage;
	}

	public String getUserExitGlobalEmailSubject() {

		return userExitGlobalEmailSubject;
	}

	public void setUserExitGlobalEmailSubject(String userExitGlobalEmailSubject) {

		this.userExitGlobalEmailSubject = userExitGlobalEmailSubject;
	}

	public String getUserExitGlobalEmailMessage() {

		return userExitGlobalEmailMessage;
	}

	public void setUserExitGlobalEmailMessage(String userExitGlobalEmailMessage) {

		this.userExitGlobalEmailMessage = userExitGlobalEmailMessage;
	}

	public String getUserExitEmailSubject() {

		return userExitEmailSubject;
	}

	public void setUserExitEmailSubject(String userExitEmailSubject) {

		this.userExitEmailSubject = userExitEmailSubject;
	}

	public String getUserExitEmailMessage() {

		return userExitEmailMessage;
	}

	public void setUserExitEmailMessage(String userExitEmailMessage) {

		this.userExitEmailMessage = userExitEmailMessage;
	}

	public String getManualEntranceSMS() {

		return manualEntranceSMS;
	}

	public void setManualEntranceSMS(String manualEntranceSMS) {

		this.manualEntranceSMS = manualEntranceSMS;
	}

	public String getExternalMessageReceivedUserSMS() {

		return externalMessageReceivedUserSMS;
	}

	public void setExternalMessageReceivedUserSMS(String externalMessageReceivedUserSMS) {

		this.externalMessageReceivedUserSMS = externalMessageReceivedUserSMS;
	}

	public String getUserExitSMS() {

		return userExitSMS;
	}

	public void setUserExitSMS(String userExitSMS) {

		this.userExitSMS = userExitSMS;
	}

	public String getUserReminderSMS() {

		return userReminderSMS;
	}

	public void setUserReminderSMS(String userReminderSMS) {

		this.userReminderSMS = userReminderSMS;
	}

	public boolean isAllowMultipleBookings() {

		return allowMultipleBookings;
	}

	public void setAllowMultipleBookings(boolean allowMultipleBookings) {

		this.allowMultipleBookings = allowMultipleBookings;
	}

	public boolean isAllowDuplicateBookings() {

		return allowDuplicateBookings;
	}

	public void setAllowDuplicateBookings(boolean allowDuplicateBookings) {

		this.allowDuplicateBookings = allowDuplicateBookings;
	}

	public boolean isBookingSizeSelectorEnabled() {

		return bookingSizeSelectorEnabled;
	}

	public void setBookingSizeSelectorEnabled(boolean bookingSizeSelectorEnabled) {

		this.bookingSizeSelectorEnabled = bookingSizeSelectorEnabled;
	}

	public boolean isAllowUserBookingSizeSelection() {

		return allowUserBookingSizeSelection;
	}

	public void setAllowUserBookingSizeSelection(boolean allowUserBookingSizeSelection) {

		this.allowUserBookingSizeSelection = allowUserBookingSizeSelection;
	}

	public Integer getUserBookingMaxSize() {

		return userBookingMaxSize;
	}

	public void setUserBookingMaxSize(Integer userBookingMaxSize) {

		this.userBookingMaxSize = userBookingMaxSize;
	}

	public String getUserRebookedManagerEmailSubject() {

		return userRebookedManagerEmailSubject;
	}

	public void setUserRebookedManagerEmailSubject(String userRebookedManagerEmailSubject) {

		this.userRebookedManagerEmailSubject = userRebookedManagerEmailSubject;
	}

	public String getUserRebookedManagerEmailMessage() {

		return userRebookedManagerEmailMessage;
	}

	public void setUserRebookedManagerEmailMessage(String userRebookedManagerEmailMessage) {

		this.userRebookedManagerEmailMessage = userRebookedManagerEmailMessage;
	}

	public List<String> getExportAttributes() {

		return exportAttributes;
	}

	public void setExportAttributes(List<String> exportAttributes) {

		this.exportAttributes = exportAttributes;
	}

	public boolean isSendManualReminderSMS() {

		return sendManualReminderSMS;
	}

	public void setSendManualReminderSMS(boolean sendManualReminderSMS) {

		this.sendManualReminderSMS = sendManualReminderSMS;
	}

	public String getManualReminderSMS() {

		return manualReminderSMS;
	}

	public void setManualReminderSMS(String manualReminderSMS) {

		this.manualReminderSMS = manualReminderSMS;
	}

	public String getManualReminderEmailSubject() {

		return manualReminderEmailSubject;
	}

	public void setManualReminderEmailSubject(String manualReminderEmailSubject) {

		this.manualReminderEmailSubject = manualReminderEmailSubject;
	}

	public String getManualReminderEmailMessage() {

		return manualReminderEmailMessage;
	}

	public void setManualReminderEmailMessage(String manualReminderEmailMessage) {

		this.manualReminderEmailMessage = manualReminderEmailMessage;
	}

	public Time getDefaultStartTime() {

		return defaultStartTime;
	}

	public void setDefaultStartTime(Time defaultStartTime) {

		this.defaultStartTime = defaultStartTime;
	}

	public boolean isDisableLeaving() {

		return disableLeaving;
	}

	public void setDisableLeaving(boolean disableLeaving) {

		this.disableLeaving = disableLeaving;
	}

	public boolean isDisableRebooking() {

		return disableRebooking;
	}

	public void setDisableRebooking(boolean disableRebooking) {

		this.disableRebooking = disableRebooking;
	}

	public boolean isRestrictRebookingToPreviousManager() {

		return restrictRebookingToPreviousManager;
	}

	public void setRestrictRebookingToPreviousManager(boolean restrictRebookingToPreviousManager) {

		this.restrictRebookingToPreviousManager = restrictRebookingToPreviousManager;
	}

	public boolean isRestrictRebookingToCurrentBookingSize() {

		return restrictRebookingToCurrentBookingSize;
	}

	public void setRestrictRebookingToCurrentBookingSize(boolean restrictRebookingToCurrentBookingSize) {

		this.restrictRebookingToCurrentBookingSize = restrictRebookingToCurrentBookingSize;
	}

	public Integer getMinVisibleDaysRebooking() {

		return minVisibleDaysRebooking;
	}

	public void setMinVisibleDaysRebooking(Integer minVisibleDaysRebooking) {

		this.minVisibleDaysRebooking = minVisibleDaysRebooking;
	}

	public boolean isUseCalendarForInternalUsers() {

		return useCalendarForInternalUsers;
	}

	public void setUseCalendarForInternalUsers(boolean useCalendarForInternalUsers) {

		this.useCalendarForInternalUsers = useCalendarForInternalUsers;
	}

	public boolean isAllowPreliminaryBookings() {

		return allowPreliminaryBookings;
	}

	public void setAllowPreliminaryBookings(boolean allowPreliminaryBookings) {

		this.allowPreliminaryBookings = allowPreliminaryBookings;
	}

	public Integer getHoursPreliminaryBookingsAreValidFor() {

		return hoursPreliminaryBookingsAreValidFor;
	}

	public void setHoursPreliminaryBookingsAreValidFor(Integer hoursPreliminaryBookingsAreValidFor) {

		this.hoursPreliminaryBookingsAreValidFor = hoursPreliminaryBookingsAreValidFor;
	}

	public boolean isHideAddedDateFromUser() {

		return hideAddedDateFromUser;
	}

	public void setHideAddedDateFromUser(boolean hideAddedDateFromUser) {

		this.hideAddedDateFromUser = hideAddedDateFromUser;
	}

	public Integer getProfileID() {

		return profileID;
	}

	public void setProfileID(Integer profileID) {

		this.profileID = profileID;
	}

	public boolean isShowFreeSlotsAsBoolean() {

		return showFreeSlotsAsBoolean;
	}

	public void setShowFreeSlotsAsBoolean(boolean showFreeSlotsAsBoolean) {

		this.showFreeSlotsAsBoolean = showFreeSlotsAsBoolean;
	}

	public boolean isShowFullyBookedDaysInCalendar() {

		return showFullyBookedDaysInCalendar;
	}

	public void setShowFullyBookedDaysInCalendar(boolean showFullyBookedDaysInCalendar) {

		this.showFullyBookedDaysInCalendar = showFullyBookedDaysInCalendar;
	}

	public String getFullyBookedCalendarDayText() {

		return fullyBookedCalendarDayText;
	}

	public void setFullyBookedCalendarDayText(String fullyBookedCalendarDayText) {

		this.fullyBookedCalendarDayText = fullyBookedCalendarDayText;
	}

	public boolean isShowFullyBookedSlotsInList() {

		return showFullyBookedSlotsInList;
	}

	public void setShowFullyBookedSlotsInList(boolean showFullyBookedSlotsInList) {

		this.showFullyBookedSlotsInList = showFullyBookedSlotsInList;
	}

	public String getFullyBookedListSlotText() {

		return fullyBookedListSlotText;
	}

	public void setFullyBookedListSlotText(String fullyBookedListSlotText) {

		this.fullyBookedListSlotText = fullyBookedListSlotText;
	}

	public boolean isHideBookingsPerWeekday() {

		return hideBookingsPerWeekday;
	}

	public void setHideBookingsPerWeekday(boolean hideBookingsPerWeekday) {

		this.hideBookingsPerWeekday = hideBookingsPerWeekday;
	}

	public List<DayOfWeek> getWeekdaysWithHiddenBookings() {

		return weekdaysWithHiddenBookings;
	}

	public void setWeekdaysWithHiddenBookings(List<DayOfWeek> weekdaysWithHiddenBookings) {

		this.weekdaysWithHiddenBookings = weekdaysWithHiddenBookings;
	}

	public boolean isSendRebookedUserSMS() {

		return sendRebookedUserSMS;
	}

	public void setSendRebookedUserSMS(boolean sendRebookedUserSMS) {

		this.sendRebookedUserSMS = sendRebookedUserSMS;
	}

	public boolean isSendRebookedUserEmail() {

		return sendRebookedUserEmail;
	}

	public void setSendRebookedUserEmail(boolean sendRebookedUserEmail) {

		this.sendRebookedUserEmail = sendRebookedUserEmail;
	}

	public boolean isSendUserRebookedManagerEmail() {

		return sendUserRebookedManagerEmail;
	}

	public void setSendUserRebookedManagerEmail(boolean sendUserRebookedManagerEmail) {

		this.sendUserRebookedManagerEmail = sendUserRebookedManagerEmail;
	}

	public boolean isSendUserRebookedManagerEmailOnlyToBookingSlotManagers() {

		return sendUserRebookedManagerEmailOnlyToBookingSlotManagers;
	}

	public void setSendUserRebookedManagerEmailOnlyToBookingSlotManagers(boolean sendUserRebookedManagerEmailOnlyToBookingSlotManagers) {

		this.sendUserRebookedManagerEmailOnlyToBookingSlotManagers = sendUserRebookedManagerEmailOnlyToBookingSlotManagers;
	}

	public String getRebookedUserSMS() {

		return rebookedUserSMS;
	}

	public void setRebookedUserSMS(String rebookedUserSMS) {

		this.rebookedUserSMS = rebookedUserSMS;
	}

	public String getRebookedUserEmailSubject() {

		return rebookedUserEmailSubject;
	}

	public void setRebookedUserEmailSubject(String rebookedUserEmailSubject) {

		this.rebookedUserEmailSubject = rebookedUserEmailSubject;
	}

	public String getRebookedUserEmailMessage() {

		return rebookedUserEmailMessage;
	}

	public void setRebookedUserEmailMessage(String rebookedUserEmailMessage) {

		this.rebookedUserEmailMessage = rebookedUserEmailMessage;
	}

	public boolean isSendRebookedGlobalEmail() {

		return sendRebookedGlobalEmail;
	}

	public void setSendRebookedGlobalEmail(boolean sendUserRebookedGlobalEmail) {

		this.sendRebookedGlobalEmail = sendUserRebookedGlobalEmail;
	}

	public String getRebookedGlobalEmailSubject() {

		return rebookedGlobalEmailSubject;
	}

	public void setRebookedGlobalEmailSubject(String userRebookedGlobalEmailSubject) {

		this.rebookedGlobalEmailSubject = userRebookedGlobalEmailSubject;
	}

	public String getRebookedGlobalEmailMessage() {

		return rebookedGlobalEmailMessage;
	}

	public void setRebookedGlobalEmailMessage(String userRebookedGlobalEmailMessage) {

		this.rebookedGlobalEmailMessage = userRebookedGlobalEmailMessage;
	}

	public List<String> getRebookedGlobalEmailAddresses() {

		return rebookedGlobalEmailAddresses;
	}

	public void setRebookedGlobalEmailAddresses(List<String> userRebookedGlobalEmailAddresses) {

		this.rebookedGlobalEmailAddresses = userRebookedGlobalEmailAddresses;
	}

	public boolean isSendBookingSlotRebookedUserSMS() {

		return sendBookingSlotRebookedUserSMS;
	}

	public void setSendBookingSlotRebookedUserSMS(boolean sendBookingSlotRebookedUserSMS) {

		this.sendBookingSlotRebookedUserSMS = sendBookingSlotRebookedUserSMS;
	}

	public String getBookingSlotRebookedUserSMS() {

		return bookingSlotRebookedUserSMS;
	}

	public void setBookingSlotRebookedUserSMS(String bookingSlotRebookedUserSMS) {

		this.bookingSlotRebookedUserSMS = bookingSlotRebookedUserSMS;
	}

	public boolean isSendBookingSlotRebookedUserEmail() {

		return sendBookingSlotRebookedUserEmail;
	}

	public void setSendBookingSlotRebookedUserEmail(boolean sendBookingSlotRebookedUserEmail) {

		this.sendBookingSlotRebookedUserEmail = sendBookingSlotRebookedUserEmail;
	}

	public String getBookingSlotRebookedUserEmailSubject() {

		return bookingSlotRebookedUserEmailSubject;
	}

	public void setBookingSlotRebookedUserEmailSubject(String bookingSlotRebookedUserEmailSubject) {

		this.bookingSlotRebookedUserEmailSubject = bookingSlotRebookedUserEmailSubject;
	}

	public String getBookingSlotRebookedUserEmailMessage() {

		return bookingSlotRebookedUserEmailMessage;
	}

	public void setBookingSlotRebookedUserEmailMessage(String bookingSlotRebookedUserEmailMessage) {

		this.bookingSlotRebookedUserEmailMessage = bookingSlotRebookedUserEmailMessage;
	}

	public boolean isSendBookingSlotDeletedUserSMS() {

		return sendBookingSlotDeletedUserSMS;
	}

	public void setSendBookingSlotDeletedUserSMS(boolean sendBookingSlotDeletedUserSMS) {

		this.sendBookingSlotDeletedUserSMS = sendBookingSlotDeletedUserSMS;
	}

	public String getBookingSlotDeletedUserSMS() {

		return bookingSlotDeletedUserSMS;
	}

	public void setBookingSlotDeletedUserSMS(String bookingSlotDeletedUserSMS) {

		this.bookingSlotDeletedUserSMS = bookingSlotDeletedUserSMS;
	}

	public boolean isSendBookingSlotDeletedUserEmail() {

		return sendBookingSlotDeletedUserEmail;
	}

	public void setSendBookingSlotDeletedUserEmail(boolean sendBookingSlotDeletedUserEmail) {

		this.sendBookingSlotDeletedUserEmail = sendBookingSlotDeletedUserEmail;
	}

	public String getBookingSlotDeletedUserEmailSubject() {

		return bookingSlotDeletedUserEmailSubject;
	}

	public void setBookingSlotDeletedUserEmailSubject(String bookingSlotDeletedUserEmailSubject) {

		this.bookingSlotDeletedUserEmailSubject = bookingSlotDeletedUserEmailSubject;
	}

	public String getBookingSlotDeletedUserEmailMessage() {

		return bookingSlotDeletedUserEmailMessage;
	}

	public void setBookingSlotDeletedUserEmailMessage(String bookingSlotDeletedUserEmailMessage) {

		this.bookingSlotDeletedUserEmailMessage = bookingSlotDeletedUserEmailMessage;
	}

	public boolean isShowNotBookableSlots() {

		return showNotBookableSlots;
	}

	public void setShowNotBookableSlots(boolean showNotBookableSlots) {

		this.showNotBookableSlots = showNotBookableSlots;
	}

	public boolean isHideDisabledManagersInFilter() {

		return hideDisabledManagersInFilter;
	}

	public void setHideDisabledManagersInFilter(boolean hideDisabledManagersInFilter) {

		this.hideDisabledManagersInFilter = hideDisabledManagersInFilter;
	}

	public boolean isHideExternalMessages() {

		return hideExternalMessages;
	}

	public void setHideExternalMessages(boolean hideExternalMessages) {

		this.hideExternalMessages = hideExternalMessages;
	}

	public Integer getNewExternalMessagesAllowedDays() {

		return newExternalMessagesAllowedDays;
	}

	public void setNewExternalMessagesAllowedDays(Integer newExternalMessagesAllowedDays) {

		this.newExternalMessagesAllowedDays = newExternalMessagesAllowedDays;
	}

	public String getBookingSlotDefaultDescription() {

		return bookingSlotDefaultDescription;
	}

	public void setBookingSlotDefaultDescription(String bookingSlotDefaultDescription) {

		this.bookingSlotDefaultDescription = bookingSlotDefaultDescription;
	}

	public String getBookingSlotDefaultLocation() {

		return bookingSlotDefaultLocation;
	}

	public void setBookingSlotDefaultLocation(String bookingSlotDefaultLocation) {

		this.bookingSlotDefaultLocation = bookingSlotDefaultLocation;
	}

	public void populate(XMLParser xmlParser) throws ValidationException {

		List<ValidationError> errors = new ArrayList<>();
		PopulateByXMLParser.populateByParser(FIELDS_TO_POPULATE, xmlParser, this, errors);

		if (!errors.isEmpty()) {

			throw new ValidationException(errors);
		}
	}

}
