package com.nordicpeak.flowengine.integration.cityworks;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.osgeo.proj4j.CRSFactory;
import org.osgeo.proj4j.CoordinateReferenceSystem;
import org.osgeo.proj4j.CoordinateTransform;
import org.osgeo.proj4j.CoordinateTransformFactory;
import org.osgeo.proj4j.ProjCoordinate;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.HTMLEditorSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.AttributeTagUtils;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.populators.DoublePopulator;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.string.TagReplacer;
import se.unlogic.standardutils.time.MillisecondTimeUnits;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.ByteArrayPart;
import se.unlogic.webutils.http.FilePart;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.SimpleMultipartRequest;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;

import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceAttribute;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.events.SubmitEvent;
import com.nordicpeak.flowengine.integration.base.BaseIntegrationProviderModule;
import com.nordicpeak.flowengine.integration.base.exceptions.IntegrationRequestException;
import com.nordicpeak.flowengine.integration.base.exceptions.IntegrationSettingsException;
import com.nordicpeak.flowengine.integration.cityworks.beans.AttachmentMode;
import com.nordicpeak.flowengine.integration.cityworks.beans.IntegrationEndpoint;
import com.nordicpeak.flowengine.integration.cityworks.beans.IntegrationSettings;
import com.nordicpeak.flowengine.integration.cityworks.beans.QueuedIntegrationRequest;
import com.nordicpeak.flowengine.integration.cityworks.exceptions.IncompleteRequestException;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.interfaces.PDFProvider;
import com.nordicpeak.flowengine.managers.FlowInstanceManager;
import com.nordicpeak.flowengine.utils.PDFByteAttachment;

import flexjson.JSONDeserializer;

public class CityworksIntegrationProviderModule extends BaseIntegrationProviderModule<QueuedIntegrationRequest, IntegrationSettings, IntegrationEndpoint> {

	private static final AnnotatedRequestPopulator<IntegrationSettings> POPULATOR = new AnnotatedRequestPopulator<>(IntegrationSettings.class);

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Error email subject", description = "The subject of emails sent when requests fail")
	@XSLVariable(prefix = "java.")
	private String errorEmailSubject = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Error email message", description = "The message of emails sent when errors are received", required = true)
	@XSLVariable(prefix = "java.")
	private String errorEmailMessage = "Not set";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "External ID prefix", description = "External ID prefix")
	private String externalIDPrefix = "FlowInstance-";

	@InstanceManagerDependency(required = true)
	private PDFProvider pdfProvider;

	private CoordinateTransform transformation;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);

		CoordinateTransformFactory ctFactory = new CoordinateTransformFactory();
		CRSFactory csFactory = new CRSFactory();

		// From EPSG:3006 to EPSG:3010 (SWEREF99_TM to SWEREF99_16_30)
		CoordinateReferenceSystem sourceCRS = csFactory.createFromName("EPSG:3006");
		CoordinateReferenceSystem targetCRS = csFactory.createFromName("EPSG:3010");

		transformation = ctFactory.createTransform(sourceCRS, targetCRS);
	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, CityworksIntegrationProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		settingsDAO = daoFactory.getDAO(IntegrationSettings.class);
		endpointDAO = daoFactory.getDAO(IntegrationEndpoint.class);
		queuedRequestDAO = daoFactory.getDAO(QueuedIntegrationRequest.class);

		flowFamilyIDParamFactory = settingsDAO.getParamFactory("flowFamilyID", Integer.class);
		endpointIDParamFactory = endpointDAO.getParamFactory("endpointID", Integer.class);
		lastRetryParamFactory = queuedRequestDAO.getParamFactory("lastRetry", Timestamp.class);
		eventIDParamFactory = queuedRequestDAO.getParamFactory("eventID", Integer.class);
		flowInstanceIDParamFactory = queuedRequestDAO.getParamFactory("flowInstanceID", Integer.class);
	}

	@WebPublic(alias = "update")
	public ForegroundModuleResponse updateFlowIntegrationSettings(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);

		if (flow == null) {

			return flowAdminModule.list(req, res, user, uriParser, new ValidationError("FlowNotFound"));
		}

		if (!flowAdminModule.hasFlowAccess(user, flow)) {

			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}

		IntegrationSettings integrationSettings = getIntegrationSettings(flow);

		ValidationException validationException = null;

		if (req.getMethod().equalsIgnoreCase("POST")) {

			if (integrationSettings == null) {

				integrationSettings = new IntegrationSettings();
				integrationSettings.setFlowFamilyID(flow.getFlowFamily().getFlowFamilyID());
			}

			try {
				POPULATOR.populate(integrationSettings, req);

				List<ValidationError> validationErrors = new ArrayList<>();

				IntegrationEndpoint endpoint = getEndpoint(req);

				if (endpoint == null) {

					validationErrors.add(new ValidationError("NoEndpointSelected"));

				} else {

					integrationSettings.setEndpoint(endpoint);
				}

				if (!validationErrors.isEmpty()) {

					throw new ValidationException(validationErrors);
				}

				log.info("User " + user + " updating integration settings for flow " + flow);

				if (integrationSettings.isPopulated()) {

					settingsDAO.addOrUpdate(integrationSettings, null);

				} else {

					settingsDAO.delete(integrationSettings);
				}

				flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + "#cityworksintegrationsettings");

				return null;

			} catch (ValidationException e) {

				validationException = e;
			}
		}

		log.info("User " + user + " requesting update integration settings form for flow " + flow);

		Document doc = createDocument(req, uriParser, user);

		Element integrationSettingsElement = doc.createElement("UpdateIntegrationSettings");
		doc.getDocumentElement().appendChild(integrationSettingsElement);

		integrationSettingsElement.appendChild(flow.toXML(doc));

		XMLUtils.append(doc, integrationSettingsElement, integrationSettings);
		XMLUtils.append(doc, integrationSettingsElement, "Endpoints", endpointDAO.getAll());

		if (validationException != null) {

			integrationSettingsElement.appendChild(validationException.toXML(doc));
			integrationSettingsElement.appendChild(RequestUtils.getRequestParameters(req, doc));
		}

		return new SimpleForegroundModuleResponse(doc, this.getDefaultBreadcrumb());
	}

	@EventListener(channel = FlowInstanceManager.class, priority = 49)
	public void processEvent(SubmitEvent event, EventSource eventSource) throws Exception {

		FlowInstance flowInstance = (FlowInstance) event.getFlowInstanceManager().getFlowInstance();
		IntegrationSettings settings = getIntegrationSettings(flowInstance.getFlow().getFlowFamily().getFlowFamilyID());

		if (settings == null || !settings.isEnabled() || !shouldProcessEvent(event)) {

			return;
		}

		FlowInstanceEvent flowInstanceEvent = event.getEvent();

		try {

			processIntegrationEvent(flowInstance, flowInstanceEvent);

		} catch (IncompleteRequestException e) {

			if (settings.isSendErrorEmail()) {

				log.info("Sending error emails");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

		} catch (IntegrationRequestException e) {

			if (maxRetryCount > 0) {

				log.info("Adding request to retry queue");

				addQueuedEvent(flowInstanceEvent, flowInstance);

			} else if (settings.isSendErrorEmail()) {

				log.info("Sending error emails");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

		} catch (IntegrationSettingsException e) {

			log.error("Settings error while processing event " + event, e);
		}
	}

	@Override
	public void retryEvent(FlowInstance flowInstance, FlowInstanceEvent event, QueuedIntegrationRequest queuedRequest, TransactionHandler transactionHandler) throws Exception {

		try {

			super.retryEvent(flowInstance, event, queuedRequest, transactionHandler);

		} catch (IncompleteRequestException e) {

			log.info("Deleting " + queuedRequest + " due to previous error");

			queuedRequestDAO.delete(queuedRequest, transactionHandler);

			IntegrationSettings settings;

			if (flowInstance != null && (settings = getIntegrationSettings(flowInstance.getFlow())) != null && settings.isSendErrorEmail()) {

				log.info("Sending error emails");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}
		}
	}

	@Override
	public void processIntegrationEvent(FlowInstance flowInstance, FlowInstanceEvent event) throws SQLException, IntegrationSettingsException, IntegrationRequestException {

		IntegrationSettings integrationSettings = getIntegrationSettings(flowInstance.getFlow());

		if (integrationSettings == null) {

			throw new IntegrationSettingsException("Settings are null");
		}

		if (!integrationSettings.isEnabled()) {

			throw new IntegrationSettingsException("Settings are disabled");
		}

		log.info("Triggering Cityworks integration for event " + event + " belonging to flow instance " + flowInstance);

		MutableAttributeHandler attributeHandler = flowInstance.getAttributeHandler();

		Long externalID = null;

		try {
			externalID = addServiceRequest(flowInstance, attributeHandler, integrationSettings);

		} catch (Exception e) {

			boolean shouldLogAsError = maxRetryCount == 0 && (!integrationSettings.isSendErrorEmail() || CollectionUtils.isEmpty(integrationSettings.getErrorEmailAddresses()));

			logResponse("Error creating cityworks service request for event " + event + " belonging to flow instance " + flowInstance, e, shouldLogAsError);

			throw new IntegrationRequestException();
		}

		if (externalID != null) {

			log.info("Got external ID " + externalID + " for flow instance " + flowInstance);

			FlowInstanceAttribute attribute = new FlowInstanceAttribute(Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE, externalID.toString());
			attribute.setFlowInstance(flowInstance);

			flowAdminModule.getDAOFactory().getFlowInstanceAttributeDAO().add(attribute);

			attributeHandler.setAttribute(Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE, externalID);
		}

		boolean shouldLogAsError = !integrationSettings.isSendErrorEmail() || CollectionUtils.isEmpty(integrationSettings.getErrorEmailAddresses());
		boolean incompleteRequest = false;

		if (externalID != null && integrationSettings.getAttachmentMode() != null) {

			try {
				File pdf = pdfProvider.getPDF(flowInstance.getFlowInstanceID(), event.getEventID());

				if (pdf == null) {

					log.error("No PDF found for event " + event + " belonging to flow instance " + flowInstance);
					return;
				}

				attachFileToServiceRequest(flowInstance, externalID, pdf, flowInstance.getFlowInstanceID() + ".pdf", "", integrationSettings);

				if (integrationSettings.getAttachmentMode() == AttachmentMode.PDF_AND_ATTACHMENTS) {

					try {
						List<PDFByteAttachment> pdfAttachments = pdfProvider.getPDFAttachments(pdf, true);

						for (PDFByteAttachment attachment : pdfAttachments) {
							attachFileToServiceRequest(externalID, attachment.getData(), flowInstance.getFlowInstanceID() + "-" + attachment.getFilename(), "", integrationSettings);
						}

					} catch (Exception e) {

						logResponse("Error getting attachments from PDF for event " + event + " belonging to flow instance " + flowInstance, e, shouldLogAsError);

						incompleteRequest = true;
					}
				}

				log.info("Attachments successfully sent");

			} catch (Exception e) {

				logResponse("Error sending attachments for event " + event + " belonging to flow instance " + flowInstance, e, shouldLogAsError);

				incompleteRequest = true;
			}
		}

		if (incompleteRequest) {

			throw new IncompleteRequestException();
		}
	}

	@SuppressWarnings("unchecked")
	private String getCityWorksSessionToken(IntegrationSettings settings) throws IOException {

		SimpleRequest request = new SimpleRequest(settings.getEndpoint().getServiceAddress() + "/General/Authentication/Authenticate");
		request.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
		request.setReadTimeout(requestTimeout * MillisecondTimeUnits.SECOND);

		JsonObject jsonNode = new JsonObject();
		jsonNode.putField("LoginName", settings.getUsername());
		jsonNode.putField("Password", settings.getPassword());

		request.addParameter("data", jsonNode.toJson());

		String response = HTTPUtils.sendHTTPGetRequest(request, StandardCharsets.UTF_8).getValue();
		Map<String, Object> responseJson = new JSONDeserializer<Map<String, Object>>().deserialize(response);

		long status = (Long) responseJson.get("Status");

		if (status != 0) {

			throw new IOException("Error getting session token, got status code " + parseStatusCode(status, responseJson));
		}

		return (String) ((Map<String, Object>) responseJson.get("Value")).get("Token");
	}

	@SuppressWarnings("unchecked")
	private long addServiceRequest(ImmutableFlowInstance flowInstance, MutableAttributeHandler attributeHandler, IntegrationSettings settings) throws IOException {

		log.info("Sending create service request for " + flowInstance);
		SimpleRequest addRequest = new SimpleRequest(settings.getEndpoint().getServiceAddress() + "/AMS/ServiceRequest/Create");
		addRequest.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
		addRequest.setReadTimeout(requestTimeout * MillisecondTimeUnits.SECOND);

		addRequest.addParameter("token", urlEncode(getCityWorksSessionToken(settings)));

		TagReplacer tagReplacer = getTagReplacer(flowInstance);

		JsonObject jsonNode = new JsonObject();
		jsonNode.putField("ProblemSid", settings.getProblemSid());
		jsonNode.putField("Address", replaceTags(settings.getAddress(), attributeHandler, tagReplacer));
		jsonNode.putField("Priority", settings.getPriorityID().toString());
		jsonNode.putField("Details", replaceTags(settings.getDetails(), attributeHandler, tagReplacer));
		jsonNode.putField("CallerFirstName", replaceTags(settings.getCallerFirstname(), attributeHandler, tagReplacer));
		jsonNode.putField("CallerLastName", replaceTags(settings.getCallerLastname(), attributeHandler, tagReplacer));
		jsonNode.putField("CallerEmail", replaceTags(settings.getCallerEmail(), attributeHandler, tagReplacer));
		jsonNode.putField("CallerCellPhone", replaceTags(settings.getCallerPhone(), attributeHandler, tagReplacer));
		jsonNode.putField("CallerIsFollowUpCall", "true".equalsIgnoreCase(settings.getCallerWantsFollowUpCall()));

		if (settings.getCoordinates() != null) {

			String coordinates = AttributeTagUtils.replaceTags(settings.getCoordinates(), attributeHandler, false);

			if (!StringUtils.isEmpty(coordinates)) {

				String[] splitCoordinates = coordinates.split(";")[0].split(",");

				DoublePopulator populator = DoublePopulator.getPopulator();

				if (splitCoordinates.length >= 2 && populator.validateFormat(splitCoordinates[0]) && populator.validateFormat(splitCoordinates[1])) {

					Double x = populator.getValue(splitCoordinates[0]);
					Double y = populator.getValue(splitCoordinates[1]);

					ProjCoordinate coordinate = new ProjCoordinate(x, y);
					transformation.transform(coordinate, coordinate);

					jsonNode.putField("X", coordinate.x);
					jsonNode.putField("Y", coordinate.y);

				} else {

					log.warn("Unable to parse coordinate for " + flowInstance + ", coordinate: " + coordinates);
				}
			}
		}

		addRequest.addParameter("data", urlEncode(jsonNode.toJson()));

		String createResponse = HTTPUtils.sendHTTPGetRequest(addRequest, StandardCharsets.UTF_8).getValue();

		Map<String, Object> createJson;
		try {
			createJson = new JSONDeserializer<Map<String, Object>>().deserialize(createResponse);

		} catch (ClassCastException e) {

			throw new IOException("Unable to parse response from CityWorks, got response: " + createResponse);
		}

		Long status = (Long) createJson.get("Status");

		if (status != 0) {

			throw new IOException("Error creating service request, got status code " + parseStatusCode(status, createJson));
		}

		return (Long) ((Map<String, Object>) createJson.get("Value")).get("RequestId");
	}

	private void attachFileToServiceRequest(ImmutableFlowInstance flowInstance, long requestID, File file, String filename, String comment, IntegrationSettings settings) throws Exception {

		log.info("Sending file attachment \"" + filename + "\" for service request ID " + requestID + " for " + flowInstance);
		SimpleMultipartRequest attachRequest = new SimpleMultipartRequest(settings.getEndpoint().getServiceAddress() + "/AMS/Attachments/AddRequestAttachment");
		attachRequest.addParameter("token", urlEncode(getCityWorksSessionToken(settings)));

		JsonObject jsonNode = new JsonObject();
		jsonNode.putField("RequestId", requestID);
		jsonNode.putField("Filename", filename);
		jsonNode.putField("Comments", comment);

		attachRequest.addParameter("data", urlEncode(jsonNode.toJson()));
		attachRequest.addPart(new FilePart(file));

		String attachResponse = HTTPUtils.sendHTTPPostRequest(attachRequest, StandardCharsets.UTF_8);

		Map<String, Object> attachJson = new JSONDeserializer<Map<String, Object>>().deserialize(attachResponse);

		Long status = (Long) attachJson.get("Status");

		if (status != 0) {

			throw new IOException("Error uploading attachment, got status code " + parseStatusCode(status, attachJson));
		}
	}

	private void attachFileToServiceRequest(long requestID, byte[] data, String filename, String comment, IntegrationSettings settings) throws Exception {

		log.info("Sending file attachment for request " + requestID);
		SimpleMultipartRequest attachRequest = new SimpleMultipartRequest(settings.getEndpoint().getServiceAddress() + "/AMS/Attachments/AddRequestAttachment");
		attachRequest.addParameter("token", urlEncode(getCityWorksSessionToken(settings)));

		JsonObject jsonNode = new JsonObject();
		jsonNode.putField("RequestId", requestID);
		jsonNode.putField("Filename", filename);
		jsonNode.putField("Comments", comment);

		attachRequest.addParameter("data", urlEncode(jsonNode.toJson()));
		attachRequest.addPart(new ByteArrayPart(filename, data));

		String attachResponse = HTTPUtils.sendHTTPPostRequest(attachRequest, StandardCharsets.UTF_8);

		Map<String, Object> attachJson = new JSONDeserializer<Map<String, Object>>().deserialize(attachResponse);

		Long status = (Long) attachJson.get("Status");

		if (status != 0) {

			throw new IOException("Error uploading attachment, got status code " + parseStatusCode(status, attachJson));
		}
	}

	private String parseStatusCode(long status, Map<String, Object> responseJson) {

		if (status == -1) {

			return "Connection error";
		}

		if (status == 0) {

			return "OK";
		}

		if (status == 1) {

			StringBuilder builder = new StringBuilder("Error");

			try {
				@SuppressWarnings("unchecked")
				List<Map<String, Object>> errors = (List<Map<String, Object>>) responseJson.get("ErrorMessages");

				builder.append(": ");

				for (Object it : errors) {
					builder.append(it.toString() + ", ");
				}
			} catch (ClassCastException ignore) {}

			return builder.toString();
		}

		if (status == 2) {

			return "Unauthorized";
		}

		if (status == 3) {

			return "InvalidCredentials";
		}

		return "unknown " + status;
	}

	private String urlEncode(String text) {

		try {
			return URLEncoder.encode(text, "UTF-8");
		} catch (UnsupportedEncodingException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	protected QueuedIntegrationRequest getNewQueuedIntegrationRequest() {

		return new QueuedIntegrationRequest();
	}

	@Override
	protected void handleMaxRetryCountExceeded(ImmutableFlowInstance flowInstance, FlowInstanceEvent event, QueuedIntegrationRequest queuedRequest, TransactionHandler transactionHandler, IntegrationRequestException lastRequestException) throws SQLException {

		super.handleMaxRetryCountExceeded(flowInstance, event, queuedRequest, transactionHandler, lastRequestException);

		IntegrationSettings settings;

		if (flowInstance != null && (settings = getIntegrationSettings(flowInstance.getFlow())) != null) {

			if (settings.isSendErrorEmail() && !CollectionUtils.isEmpty(settings.getErrorEmailAddresses())) {

				log.info("Sending error emails due to max retries exceeded");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));

			} else {

				log.error(queuedRequest + " hit its max retry count and error email is disabled");
			}
		}
	}

	@Override
	protected String getShowFlowModuleAnchorLink() {

		return "#cityworksintegrationsettings";
	}

}
