package com.nordicpeak.flowengine.coloredtag;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import javax.xml.transform.TransformerException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.LinkTag;
import se.unlogic.hierarchy.core.beans.ScriptTag;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.SimpleViewFragment;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.enums.ResponseType;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.AccessInterface;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.CRUDCallback;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.core.utils.ModuleViewFragmentTransformer;
import se.unlogic.hierarchy.core.utils.ViewFragmentModule;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.hierarchy.foregroundmodules.staticcontent.StaticContentModule;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.json.JsonLeaf;
import se.unlogic.standardutils.json.JsonNode;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.NonNegativeStringIntegerValidator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.standardutils.xsl.XSLVariableReader;
import se.unlogic.standardutils.xsl.XSLVariableReaderRenamer;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowAdminExtensionShowView;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.coloredtag.beans.ColoredTag;
import com.nordicpeak.flowengine.coloredtag.beans.ColoredTagMapping;
import com.nordicpeak.flowengine.coloredtag.beans.ColoredTagSettings;
import com.nordicpeak.flowengine.coloredtag.cruds.ColoredTagSettingsCRUD;
import com.nordicpeak.flowengine.coloredtag.daos.ColoredTagDAO;
import com.nordicpeak.flowengine.coloredtag.daos.ColoredTagMappingDAO;
import com.nordicpeak.flowengine.coloredtag.daos.ColoredTagSettingsDAO;
import com.nordicpeak.flowengine.flowinstanceadmin2.FlowInstanceAdmin2Module;
import com.nordicpeak.flowengine.flowinstanceadmin2.interfaces.ColoredTagProvider;
import com.nordicpeak.flowengine.flowinstanceadmin2.interfaces.ListOverviewButtonExtension;
import com.nordicpeak.flowengine.flowinstanceadmin2.interfaces.ListOverviewButtonExtensionProvider;
import com.nordicpeak.flowengine.interfaces.FlowAdminFragmentExtensionViewProvider;

public class ColoredTagProviderModule extends AnnotatedForegroundModule implements ColoredTagProvider, CRUDCallback<User>, ViewFragmentModule<ForegroundModuleDescriptor>, FlowAdminFragmentExtensionViewProvider, ListOverviewButtonExtensionProvider {

	private static final JsonLeaf EMPTY_LEAF = new JsonLeaf("");

	@XSLVariable(prefix = "java.")
	private String adminExtensionViewTitle = "not set";

	@XSLVariable(prefix = "java.")
	private String columnExtensionViewTitle = "not set";

	@XSLVariable(prefix = "java.")
	protected String noTagText;

	@XSLVariable(prefix = "java.")
	protected String noFilterText;

	@XSLVariable(prefix = "java.")
	protected String tableHeaderText;

	@XSLVariable(prefix = "java.")
	protected String extensionButtonName;

	@XSLVariable(prefix = "java.")
	protected String extensionButtonTitle;

	@XSLVariable(name = "i18n.Tag.Name")
	protected String tagNameDisplayName;

	@XSLVariable(name = "i18n.Tag.HexColor")
	protected String tagHexColorDisplayName;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Priority", description = "The priority of this extension provider compared to other providers. A low value means a higher priority. Valid values are 0 - " + Integer.MAX_VALUE + ".", required = true, formatValidator = NonNegativeStringIntegerValidator.class)
	protected int priority = 0;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Enable fragment XML debug", description = "Enables debugging of fragment XML")
	private boolean debugFragmentXML;

	//TODO add full read cache so no DB quries are nessecary for display
	private ColoredTagSettingsDAO settingsDAO;
	private ColoredTagDAO tagDAO;
	private ColoredTagMappingDAO mappingDAO;

	private ColoredTagSettingsCRUD settingsCRUD;

	private FlowAdminModule flowAdminModule;
	private FlowInstanceAdmin2Module flowInstanceAdminModule;

	private ModuleViewFragmentTransformer<ForegroundModuleDescriptor> viewFragmentTransformer;

	@InstanceManagerDependency(required = true)
	private StaticContentModule staticContentModule;

	private Set<Integer> flowInstanceIDCache;

	private List<ScriptTag> buttonExtensionScripts;
	private Map<String, String> listOverviewButtonDataAttributes;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		viewFragmentTransformer = new ModuleViewFragmentTransformer<>(sectionInterface.getForegroundModuleXSLTCache(), this, sectionInterface.getSystemInterface().getEncoding());

		super.init(moduleDescriptor, sectionInterface, dataSource);

		cacheFlowInstanceIDs();
		
		if (!systemInterface.getInstanceHandler().addInstance(ColoredTagProvider.class, this)) {

			throw new RuntimeException("Unable to register module in global instance handler using key " + ColoredTagProviderModule.class.getSimpleName() + ", another instance is already registered using this key.");
		}
	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, ColoredTagProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("daos/DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		settingsDAO = new ColoredTagSettingsDAO(dataSource, ColoredTagSettings.class, daoFactory);
		tagDAO = new ColoredTagDAO(dataSource, ColoredTag.class, daoFactory);
		mappingDAO = new ColoredTagMappingDAO(dataSource, ColoredTagMapping.class, daoFactory);

		AnnotatedDAOWrapper<ColoredTagSettings, Integer> settingsDaoWrapper = settingsDAO.getWrapper(ColoredTagSettings.FIELD_FLOW_FAMILY_ID.getName(), Integer.class);

		settingsDaoWrapper.addRelations(ColoredTagSettings.FIELD_TAGS, ColoredTag.FIELD_MAPPINGS);
		settingsDaoWrapper.setUseRelationsOnGet(true);
		settingsDaoWrapper.setUseRelationsOnAdd(true);
		settingsDaoWrapper.setUseRelationsOnUpdate(true);

		settingsCRUD = new ColoredTagSettingsCRUD(settingsDaoWrapper, this);
	}

	@Override
	protected void moduleConfigured() throws Exception {

		viewFragmentTransformer.setDebugXML(debugFragmentXML);
		viewFragmentTransformer.modifyScriptsAndLinks(true, null);

		XSLVariableReader variableReader = ModuleUtils.getXSLVariableReader(moduleDescriptor, sectionInterface.getSystemInterface());

		if (variableReader != null) {

			Map<String, String> map = new HashMap<String, String>();
			map.put("scripts", "scripts-buttonextension");

			XSLVariableReaderRenamer variableReaderRenamed = new XSLVariableReaderRenamer(variableReader, map);

			buttonExtensionScripts = ModuleUtils.getScripts(variableReaderRenamed, sectionInterface, "f", moduleDescriptor);
		}

		listOverviewButtonDataAttributes = new HashMap<String, String>(2);
		listOverviewButtonDataAttributes.put("url", systemInterface.getContextPath() + getFullAlias() + "/colortagflowinstances");
		listOverviewButtonDataAttributes.put("baseurl", systemInterface.getContextPath() + getFullAlias());
	}

	@Override
	public void unload() throws Exception {

		if (flowAdminModule != null) {

			setFlowAdminModule(null);
		}

		if (flowInstanceAdminModule != null) {

			setFlowInstanceAdminModule(null);
		}

		super.unload();
		
		systemInterface.getInstanceHandler().removeInstance(ColoredTagProvider.class, this);
	}

	@InstanceManagerDependency(required = true)
	public void setFlowAdminModule(FlowAdminModule flowAdminModule) {

		if (flowAdminModule == null && this.flowAdminModule != null) {

			this.flowAdminModule.removeFragmentExtensionViewProvider(this);
		}

		this.flowAdminModule = flowAdminModule;

		if (this.flowAdminModule != null) {

			this.flowAdminModule.addFragmentExtensionViewProvider(this);
		}
	}

	@InstanceManagerDependency(required = true)
	public void setFlowInstanceAdminModule(FlowInstanceAdmin2Module flowInstanceAdminModule) {

		if (flowInstanceAdminModule == null && this.flowInstanceAdminModule != null) {

			this.flowInstanceAdminModule.removeListOverviewButtonExtensionProvider(this);
		}

		this.flowInstanceAdminModule = flowInstanceAdminModule;

		if (this.flowInstanceAdminModule != null) {

			this.flowInstanceAdminModule.addListOverviewButtonExtensionProvider(this);
		}

	}

	private void cacheFlowInstanceIDs() {

		try {

			List<ColoredTagMapping> mappings = mappingDAO.getAll();

			if (mappings == null) {

				flowInstanceIDCache = null;

			} else {

				List<Integer> flowInstanceIDs = new ArrayList<>(mappings.size());

				for (ColoredTagMapping mapping : mappings) {

					flowInstanceIDs.add(mapping.getFlowInstanceID());
				}

				flowInstanceIDCache = ConcurrentHashMap.newKeySet(flowInstanceIDs.size());

				flowInstanceIDCache.addAll(flowInstanceIDs);
			}

		} catch (SQLException e) {

			log.error("Error caching flow instance ids", e);
		}
	}

	public void addToFlowInstanceIDCache(Integer flowInstanceID) {

		if (flowInstanceIDCache == null) {

			flowInstanceIDCache = ConcurrentHashMap.newKeySet();
		}

		flowInstanceIDCache.add(flowInstanceID);
	}

	public void removeFromFlowInstanceIDCache(Integer flowInstanceID) {

		if (flowInstanceIDCache == null) {

			return;
		}

		flowInstanceIDCache.remove(flowInstanceID);
	}

	@Override
	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {

		Document doc = XMLUtils.createDomDocument();

		Element documentElement = doc.createElement("Document");

		documentElement.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		documentElement.appendChild(this.moduleDescriptor.toXML(doc));

		doc.appendChild(documentElement);

		return doc;
	}

	@Override
	public void appendListElement(Document doc, Element targetElement, User user) throws SQLException {

		Element listElement = XMLUtils.appendNewElement(doc, targetElement, "ColoredTagProvider");

		XMLUtils.appendNewElement(doc, listElement, "ColumnID", "coloredTag");
		XMLUtils.appendNewElement(doc, listElement, "ColumnName", columnExtensionViewTitle);
		XMLUtils.appendNewElement(doc, listElement, "NoFilterText", noFilterText);
		XMLUtils.appendNewElement(doc, listElement, "TableHeaderText", tableHeaderText);
	}

	@Override
	public void appendOverviewElement(FlowInstance flowInstance, Document doc, Element targetElement) throws SQLException {

		ColoredTagSettings settings = settingsDAO.getByFlowFamilyID(flowInstance.getFlow().getFlowFamily().getFlowFamilyID(), ColoredTagSettings.FIELD_TAGS, ColoredTag.FIELD_MAPPINGS);

		if (settings == null || !settings.isEnabled() || settings.getTags() == null) {

			return;
		}

		Element overviewElement = XMLUtils.appendNewElement(doc, targetElement, "ColoredTagProvider");

		XMLUtils.appendNewElement(doc, overviewElement, "NoTagText", noTagText);

		XMLUtils.append(doc, overviewElement, "Tags", settings.getTags());

		for (ColoredTag tag : settings.getTags()) {

			if (tag.hasMapping(flowInstance.getFlowInstanceID())) {

				XMLUtils.appendNewElement(doc, overviewElement, "SelectedTagID", tag.getTagID());
				return;
			}
		}
	}

	@Override
	public JsonNode getFlowInstanceTableCellData(FlowInstance flowInstance) throws SQLException {

		if (flowInstanceIDCache != null && flowInstanceIDCache.contains(flowInstance.getFlowInstanceID())) {

			ColoredTagMapping mapping = mappingDAO.getByFlowInstanceID(flowInstance.getFlowInstanceID(), ColoredTagMapping.FIELD_TAG, ColoredTag.FIELD_SETTINGS);

			if (mapping != null && mapping.getTag().getSettings().isEnabled()) {

				return mapping.getTag().toJson();
			}
		}

		return EMPTY_LEAF;
	}

	@Override
	public ForegroundModuleResponse selectColoredTag(FlowInstance flowInstance, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws SQLException, IOException {

		Integer tagID = NumberUtils.toInt(uriParser.get(4));

		ColoredTagMapping mapping = mappingDAO.getByFlowInstanceID(flowInstance.getFlowInstanceID(), ColoredTagMapping.FIELD_TAG);

		if (tagID == null) {

			if (mapping != null) {

				log.info("User " + user + " removing colored tag for instance " + flowInstance);

				mappingDAO.delete(mapping);

				removeFromFlowInstanceIDCache(flowInstance.getFlowInstanceID());
			}

		} else if (mapping == null || !mapping.getTag().getTagID().equals(tagID)) {

			ColoredTag tag = tagDAO.get(tagID);

			if (tag != null) {

				log.info("User " + user + " selecting colored tag " + tag + " for instance " + flowInstance);

				if (mapping == null) {

					mapping = new ColoredTagMapping();
					mapping.setFlowInstanceID(flowInstance.getFlowInstanceID());
				}

				mapping.setTag(tag);

				mappingDAO.addOrUpdate(mapping, null);

				addToFlowInstanceIDCache(flowInstance.getFlowInstanceID());
			}
		}

		HTTPUtils.sendReponse(tagID != null ? tagID.toString() : "0", "text/html", res);

		return null;
	}

	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse colorTagFlowinstances(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		List<ValidationError> validationErrors = new ArrayList<ValidationError>();

		List<FlowInstance> flowInstances = flowInstanceAdminModule.getFlowInstanceSelection(true, false, req, res, user, uriParser, validationErrors);

		if (!validationErrors.isEmpty()) {

			return flowInstanceAdminModule.list(req, res, user, uriParser, validationErrors);

		} else if (flowInstances == null) {

			flowInstanceAdminModule.redirectToDefaultMethod(req, res);

			return null;
		}

		Flow flow = flowInstances.get(0).getFlow();
		FlowFamily flowFamily = flow.getFlowFamily();

		ColoredTagSettings settings = settingsDAO.getByFlowFamilyID(flowFamily.getFlowFamilyID(), ColoredTagSettings.FIELD_TAGS, ColoredTag.FIELD_MAPPINGS);

		if (settings == null || !settings.isEnabled()) {

			validationErrors.add(new ValidationError("NoColoredTagSetting"));
		}

		if (req.getMethod().equalsIgnoreCase("POST") && req.getParameter("directFromInstanceList") == null && validationErrors.isEmpty()) {

			Integer tagID = NumberUtils.toInt(req.getParameter("tagID"));

			ColoredTag tag = null;

			if (tagID != null && (tag = tagDAO.get(tagID)) == null) {

				validationErrors.add(new ValidationError("ColorTagNotFound"));

			} else if (tag != null) {

				if (settings.getTag(tag.getTagID()) == null) {

					validationErrors.add(new ValidationError("InvalidColorTag"));
				}

			}

			if (validationErrors.isEmpty()) {

				TransactionHandler transactionHandler = null;

				try {

					transactionHandler = mappingDAO.createTransaction();

					for (FlowInstance flowInstance : flowInstances) {

						ColoredTagMapping mapping = mappingDAO.getByFlowInstanceID(flowInstance.getFlowInstanceID(), ColoredTagMapping.FIELD_TAG);

						boolean deleted = false;

						if (mapping != null && (tagID == null || !mapping.getTag().getTagID().equals(tagID))) {

							mappingDAO.delete(mapping, transactionHandler);

							removeFromFlowInstanceIDCache(flowInstance.getFlowInstanceID());

							deleted = true;
						}

						if (tag != null && (mapping == null || deleted)) {

							mapping = new ColoredTagMapping();
							mapping.setFlowInstanceID(flowInstance.getFlowInstanceID());
							mapping.setTag(tag);

							mappingDAO.addOrUpdate(mapping, transactionHandler, null);

							addToFlowInstanceIDCache(flowInstance.getFlowInstanceID());

						}

					}

					transactionHandler.commit();

					log.info("User " + user + " added/updated color tags for flow instanes IDs: " + StringUtils.toCommaSeparatedString(req.getParameterValues("flowInstanceID")));

					flowInstanceAdminModule.redirectToDefaultMethod(req, res);

					return null;

				} finally {

					TransactionHandler.autoClose(transactionHandler);
				}

			}

		}

		log.info("User " + user + " requested form for mass changing flow instance color tags");

		Document doc = createDocument(req, uriParser, user);

		Element colorTagElement = XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "ColorTagFlowInstances");

		XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "fullAlias", flowInstanceAdminModule.getFullAlias());

		if (!validationErrors.isEmpty()) {

			XMLUtils.append(doc, colorTagElement, "ValidationErrors", validationErrors);
			colorTagElement.appendChild(RequestUtils.getRequestParameters(req, doc));
		}

		if (settings != null) {

			XMLUtils.append(doc, colorTagElement, "Tags", settings.getTags());
		}

		List<Integer> flowInstanceIDs = new ArrayList<Integer>(flowInstances.size());

		for (FlowInstance flowInstance : flowInstances) {
			flowInstanceIDs.add(flowInstance.getFlowInstanceID());
		}

		XMLUtils.append(doc, colorTagElement, "FlowInstances", "flowInstanceID", flowInstanceIDs);

		colorTagElement.appendChild(flow.toXML(doc));

		return new SimpleForegroundModuleResponse(doc, moduleDescriptor.getName(), getDefaultBreadcrumb());

	}

	@EventListener(channel = FlowInstance.class)
	public void processFlowInstanceEvent(CRUDEvent<FlowInstance> event, EventSource eventSource) {

		if (event.getAction() == CRUDAction.DELETE) {

			try (TransactionHandler transactionHandler = mappingDAO.createTransaction()) {
			
				for (FlowInstance flowInstance : event.getBeans()) {

					log.info("Deleting color tags for " + flowInstance);
					
					mappingDAO.deleteByFlowInstanceID(flowInstance.getFlowInstanceID(), transactionHandler);
				}
				
				transactionHandler.commit();
				
			} catch (Exception e) {
				
				log.error("Error deleting color tags for flow instances " + event.getBeans(), e);
			}
		}
	}

	@Override
	public ForegroundModuleDescriptor getModuleDescriptor() {

		return moduleDescriptor;
	}

	@Override
	public List<LinkTag> getLinkTags() {

		return links;
	}

	@Override
	public List<ScriptTag> getScriptTags() {

		return scripts;
	}

	@Override
	public String getTitlePrefix() {

		return moduleDescriptor.getName();
	}

	public ModuleViewFragmentTransformer<ForegroundModuleDescriptor> getViewFragmentTransformer() {

		return viewFragmentTransformer;
	}

	public ColoredTagSettingsDAO getSettingsDAO() {

		return settingsDAO;
	}

	public String getProviderID() {

		return this.moduleDescriptor.getModuleID().toString();
	}

	public String getColumnExtensionViewTitle() {

		return columnExtensionViewTitle;
	}

	//Fragment

	@Override
	public FlowAdminExtensionShowView getShowView(String extensionRequestURL, Flow flow, HttpServletRequest req, User user, URIParser uriParser) throws TransformerException, SQLException {

		return settingsCRUD.getShowView(extensionRequestURL, flow, req, user, uriParser);
	}

	@Override
	public ViewFragment processRequest(String extensionRequestURL, Flow flow, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		String method = uriParser.get(4);

		req.setAttribute("extensionRequestURL", extensionRequestURL);
		req.setAttribute("flow", flow);

		if ("updateflowsettings".equals(method)) {

			return getViewFragmentResponse(settingsCRUD.update(req, res, user, uriParser));

		} else if ("deleteflowsettings".equals(method)) {

			return getViewFragmentResponse(settingsCRUD.delete(req, res, user, uriParser));
		}

		throw new URINotFoundException(uriParser);
	}

	private ViewFragment getViewFragmentResponse(ForegroundModuleResponse foregroundModuleResponse) throws TransformerException {

		if (foregroundModuleResponse != null) {

			if (foregroundModuleResponse.getResponseType() == ResponseType.XML_FOR_SEPARATE_TRANSFORMATION) {

				return viewFragmentTransformer.createViewFragment(foregroundModuleResponse.getDocument());
			}

			log.warn("Scripts and links have not been modified for FlowAdminFragmentExtensionViewProviderProcessRequest");
			return new SimpleViewFragment(foregroundModuleResponse.getHtml(), debugFragmentXML ? foregroundModuleResponse.getDocument() : null, foregroundModuleResponse.getScripts(), foregroundModuleResponse.getLinks());
		}

		return null;
	}

	@Override
	public int getPriority() {

		return priority;
	}

	@Override
	public String getExtensionViewTitle() {

		return adminExtensionViewTitle;
	}

	@Override
	public String getExtensionViewLinkName() {

		return "coloredtags";
	}

	@Override
	public int getModuleID() {

		return moduleDescriptor.getModuleID();
	}

	public String getStaticContentURL() {

		return staticContentModule.getModuleContentURL(moduleDescriptor);
	}

	public String getTagNameDisplayName() {

		return tagNameDisplayName;
	}

	public String getTagHexColorDisplayName() {

		return tagHexColorDisplayName;
	}

	@Override
	public AccessInterface getAccessInterface() {

		return moduleDescriptor;
	}

	@Override
	public ListOverviewButtonExtension getListOverviewButtonExtension(HttpServletRequest req, User user) {

		return new ListOverviewButtonExtension(extensionButtonName, extensionButtonTitle, "colored-tag-btn", listOverviewButtonDataAttributes, buttonExtensionScripts);
	}

	@Override
	public Map<Integer, String> getFlowInstanceIDAndHexColorByFlowInstance(List<FlowInstance> flowInstance) throws SQLException {

		Map<Integer, String> map = new HashMap<>();
		List<ColoredTagMapping> mappings = mappingDAO.getByFlowInstanceID(flowInstance.stream().map(FlowInstance::getFlowInstanceID).collect(Collectors.toList()), ColoredTagMapping.FIELD_TAG);
		
		if(mappings != null) {
			for (ColoredTagMapping coloredTagMapping : mappings) {
				map.put(coloredTagMapping.getFlowInstanceID(), coloredTagMapping.getTag().getHexColor());
			}
		}
		
		return map;
	}

}