package com.nordicpeak.flowengine.queries;

import static java.time.temporal.TemporalAdjusters.firstDayOfMonth;
import static java.time.temporal.TemporalAdjusters.lastDayOfMonth;

import java.sql.Date;
import java.sql.SQLException;
import java.sql.Time;
import java.text.SimpleDateFormat;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.Period;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.hierarchy.core.utils.FCKUtils;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.enums.DayOfWeek;
import se.unlogic.standardutils.object.ObjectUtils;
import se.unlogic.standardutils.populators.DatePopulator;
import se.unlogic.standardutils.populators.TimePopulator;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLGenerator;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.url.URLRewriter;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.beans.InstanceRequestMetadata;
import com.nordicpeak.flowengine.enums.QueryState;
import com.nordicpeak.flowengine.interfaces.ImmutableQueryDescriptor;
import com.nordicpeak.flowengine.interfaces.ImmutableQueryInstanceDescriptor;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.InstanceMetadata;
import com.nordicpeak.flowengine.interfaces.MutableQueryDescriptor;
import com.nordicpeak.flowengine.interfaces.MutableQueryInstanceDescriptor;
import com.nordicpeak.flowengine.interfaces.Query;
import com.nordicpeak.flowengine.interfaces.QueryContentFilter;
import com.nordicpeak.flowengine.interfaces.QueryInstance;
import com.nordicpeak.flowengine.queries.basequery.BaseQueryCRUDCallback;
import com.nordicpeak.flowengine.queries.basequery.BaseQueryProviderModule;
import com.nordicpeak.flowengine.queries.enums.DateIntervalRestrictionType;
import com.nordicpeak.flowengine.queries.enums.DateIntervalRestrictionUnit;
import com.nordicpeak.flowengine.queries.enums.DateRestrictionType;
import com.nordicpeak.flowengine.queries.enums.TomorrowRestrictionMode;
import com.nordicpeak.flowengine.queries.validationerrors.DateIntervallValidationError;
import com.nordicpeak.flowengine.queries.validationerrors.TimeNotMultipleOfInterval;
import com.nordicpeak.flowengine.utils.JTidyUtils;
import com.nordicpeak.flowengine.utils.TextTagReplacer;

public class DateTimeQueryProviderModule extends BaseQueryProviderModule<DateTimeQueryInstance> implements BaseQueryCRUDCallback {

	private static final DatePopulator DATE_POPULATOR = new DatePopulator(true);
	private static final TimePopulator TIME_POPULATOR = new TimePopulator();
	
	private static final SimpleDateFormat TIME_FORMATTER = new SimpleDateFormat("HH:mm");

	@XSLVariable(prefix = "java.")
	private String mondayString = "MONDAY";
	
	@XSLVariable(prefix = "java.")
	private String tuesdayString = "TUESDAY";

	@XSLVariable(prefix = "java.")
	private String wednesdayString = "WEDNESDAY";

	@XSLVariable(prefix = "java.")
	private String thursdayString = "THURSDAY";

	@XSLVariable(prefix = "java.")
	private String fridayString = "FRIDAY";

	@XSLVariable(prefix = "java.")
	private String saturdayString = "SATURDAY";

	@XSLVariable(prefix = "java.")
	private String sundayString = "SUNDAY";

	private AnnotatedDAO<DateTimeQuery> queryDAO;
	private AnnotatedDAO<DateTimeQueryInstance> queryInstanceDAO;

	private DateTimeQueryCRUD queryCRUD;

	private QueryParameterFactory<DateTimeQuery, Integer> queryIDParamFactory;
	private QueryParameterFactory<DateTimeQueryInstance, Integer> queryInstanceIDParamFactory;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, DateTimeQueryProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		queryDAO = daoFactory.getDAO(DateTimeQuery.class);
		queryInstanceDAO = daoFactory.getDAO(DateTimeQueryInstance.class);

		queryCRUD = new DateTimeQueryCRUD(queryDAO.getWrapper(Integer.class), new AnnotatedRequestPopulator<DateTimeQuery>(DateTimeQuery.class), "DateTimeQuery", "query", null, this);

		queryIDParamFactory = queryDAO.getParamFactory("queryID", Integer.class);
		queryInstanceIDParamFactory = queryInstanceDAO.getParamFactory("queryInstanceID", Integer.class);
	}

	@Override
	public Query createQuery(MutableQueryDescriptor descriptor, TransactionHandler transactionHandler) throws SQLException {

		DateTimeQuery query = new DateTimeQuery();

		query.setQueryID(descriptor.getQueryID());
		
		List<DayOfWeek> allDays = new ArrayList<DayOfWeek>(7);
		
		allDays.add(DayOfWeek.MONDAY);
		allDays.add(DayOfWeek.TUESDAY);
		allDays.add(DayOfWeek.WEDNESDAY);
		allDays.add(DayOfWeek.THURSDAY);
		allDays.add(DayOfWeek.FRIDAY);
		allDays.add(DayOfWeek.SATURDAY);
		allDays.add(DayOfWeek.SUNDAY);
		
		query.setAllowedDaysOfWeek(allDays);

		this.queryDAO.add(query, transactionHandler, null);

		query.init(descriptor, getFullAlias() + "/config/" + descriptor.getQueryID());

		return query;
	}

	@Override
	public Query importQuery(MutableQueryDescriptor descriptor, TransactionHandler transactionHandler, Map<Integer, ImmutableStatus> statusConversionMap, QueryContentFilter contentFilter) throws Throwable {

		DateTimeQuery query = new DateTimeQuery();

		query.setQueryID(descriptor.getQueryID());

		query.populate(descriptor.getImportParser().getNode(XMLGenerator.getElementName(query.getClass())));
		
		contentFilter.filterHTML(query);

		this.queryDAO.add(query, transactionHandler, null);

		return query;
	}

	@Override
	public Query getQuery(MutableQueryDescriptor descriptor, boolean extraData) throws SQLException {

		DateTimeQuery query = this.getQuery(descriptor.getQueryID());

		if (query == null) {

			return null;
		}

		query.init(descriptor, getFullAlias() + "/config/" + descriptor.getQueryID());

		return query;
	}

	@Override
	public Query getQuery(MutableQueryDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		DateTimeQuery query = this.getQuery(descriptor.getQueryID(), transactionHandler);

		if (query == null) {

			return null;
		}

		query.init(descriptor, getFullAlias() + "/config/" + descriptor.getQueryID());

		return query;
	}

	@Override
	public QueryInstance getQueryInstance(MutableQueryInstanceDescriptor descriptor, String instanceManagerID, HttpServletRequest req, User user, User poster, InstanceMetadata instanceMetadata) throws SQLException {

		DateTimeQueryInstance queryInstance;

		//Check if we should create a new instance or get an existing one
		if (descriptor.getQueryInstanceID() == null) {

			queryInstance = new DateTimeQueryInstance();

		} else {

			queryInstance = getQueryInstance(descriptor.getQueryInstanceID());

			if (queryInstance == null) {

				return null;
			}
		}

		queryInstance.setQuery(getQuery(descriptor.getQueryDescriptor().getQueryID()));

		if (queryInstance.getQuery() == null) {

			return null;
		}

		if (req != null) {

			FCKUtils.setAbsoluteFileUrls(queryInstance.getQuery(), RequestUtils.getFullContextPathURL(req) + ckConnectorModuleAlias);

			URLRewriter.setAbsoluteLinkUrls(queryInstance.getQuery(), req, true);
		}

		queryInstance.getQuery().scanAttributeTags();

		TextTagReplacer.replaceTextTags(queryInstance.getQuery(), instanceMetadata.getSiteProfile());

		queryInstance.set(descriptor);

		//If this is a new query instance copy the default values
		if (descriptor.getQueryInstanceID() == null) {

			queryInstance.copyQueryValues();
		}

		return queryInstance;
	}

	private DateTimeQuery getQuery(Integer queryID) throws SQLException {

		HighLevelQuery<DateTimeQuery> query = new HighLevelQuery<DateTimeQuery>();

		query.addParameter(queryIDParamFactory.getParameter(queryID));

		return queryDAO.get(query);
	}

	private DateTimeQuery getQuery(Integer queryID, TransactionHandler transactionHandler) throws SQLException {

		HighLevelQuery<DateTimeQuery> query = new HighLevelQuery<DateTimeQuery>();

		query.addParameter(queryIDParamFactory.getParameter(queryID));

		return queryDAO.get(query, transactionHandler);
	}

	private DateTimeQueryInstance getQueryInstance(Integer queryInstanceID) throws SQLException {

		HighLevelQuery<DateTimeQueryInstance> query = new HighLevelQuery<DateTimeQueryInstance>(DateTimeQueryInstance.QUERY_RELATION);

		query.addParameter(queryInstanceIDParamFactory.getParameter(queryInstanceID));

		return queryInstanceDAO.get(query);
	}

	@Override
	public void save(DateTimeQueryInstance queryInstance, TransactionHandler transactionHandler, InstanceRequestMetadata requestMetadata) throws Throwable {

		if (queryInstance.getQueryInstanceID() == null || !queryInstance.getQueryInstanceID().equals(queryInstance.getQueryInstanceDescriptor().getQueryInstanceID())) {

			queryInstance.setQueryInstanceID(queryInstance.getQueryInstanceDescriptor().getQueryInstanceID());

			this.queryInstanceDAO.add(queryInstance, transactionHandler, null);

		} else {

			this.queryInstanceDAO.update(queryInstance, transactionHandler, null);
		}
	}
	
	@Override
	protected void appendQueryInstance(DateTimeQueryInstance queryInstance, Document doc, Element targetElement, AttributeHandler attributeHandler) {
		
		super.appendQueryInstance(queryInstance, doc, targetElement, attributeHandler);
		
		if (queryInstance.getQuery().isLockOnOwnershipTransfer() && attributeHandler.getPrimitiveBoolean("OwnershipTransfered")) {
			
			XMLUtils.appendNewElement(doc, targetElement, "Locked", "true");
		}
	}

	@Override
	public void populate(DateTimeQueryInstance queryInstance, HttpServletRequest req, User user, User poster, boolean allowPartialPopulation, MutableAttributeHandler attributeHandler, InstanceRequestMetadata requestMetadata) throws ValidationException {

		DateTimeQuery query = queryInstance.getQuery();
		Integer queryID = query.getQueryID();
		
		if (query.isLockOnOwnershipTransfer() && attributeHandler.getPrimitiveBoolean("OwnershipTransfered")) {
			return;
		}
		
		List<ValidationError> errors = new ArrayList<ValidationError>();

		Date startDate = null;
		Date endDate = null;
		Time startTime = null;
		Time endTime = null;
		
		String queryPrefix = "q" + queryID;
		String startTimeParameter = queryPrefix + "_startTime";
		String endTimeParameter = queryPrefix + "_endTime";
		
		if (queryInstance.getQuery().isUseDateFields()) {

			startDate = ValidationUtils.validateParameter("q" + queryID + "_startDate", req, !allowPartialPopulation, DATE_POPULATOR, errors);

			if (queryInstance.getQuery().isUseMultiDate()) {

				endDate = ValidationUtils.validateParameter("q" + queryID + "_endDate", req, !allowPartialPopulation, DATE_POPULATOR, errors);
			}
		}

		if (queryInstance.getQuery().isUseTimeFields()) {

			if (queryInstance.getQuery().isUseMultiTime()) {

				endTime = ValidationUtils.validateParameter(endTimeParameter, req, !allowPartialPopulation, TIME_POPULATOR, errors);
			}

			startTime = ValidationUtils.validateParameter(startTimeParameter, req, !allowPartialPopulation, TIME_POPULATOR, errors);
		}

		if (queryInstance.getQueryInstanceDescriptor().getQueryState() != QueryState.VISIBLE_REQUIRED && ObjectUtils.isNull(startDate, endDate, startTime, endTime)) {

			queryInstance.reset(attributeHandler);
			return;
		}

		if (!errors.isEmpty()) {

			throw new ValidationException(errors);
		}

		if (startDate != null && endDate != null) {

			if (startDate.after(endDate)) {

				errors.add(new ValidationError("StartDateAfterEndDate"));

			} else if (startTime != null && endTime != null) {

				if (!queryInstance.getQuery().isAllowStartTimeAfterEndTime() && startDate.equals(endDate) && startTime.after(endTime)) {

					errors.add(new ValidationError("StartTimeAfterEndTime"));
				}
			}

		} else if (startTime != null && endTime != null) {

			if (!queryInstance.getQuery().isAllowStartTimeAfterEndTime() && startTime.after(endTime)) {

				errors.add(new ValidationError("StartTimeAfterEndTime"));
			}
		}
		
		if(!requestMetadata.isManager()) {
		
			validateDate(startDate, endDate, startTime, endTime, queryInstance, errors);
		}

		if (startTime != null) {
			
			boolean otherStartTimeErrors = false;
			
			if (queryInstance.getQuery().getEarliestTime() != null && startTime.before(queryInstance.getQuery().getEarliestTime())) {
				
				otherStartTimeErrors = true;
				errors.add(new ValidationError("StartTimeTooEarly", TimeUtils.TIME_FORMATTER.format(queryInstance.getQuery().getEarliestTime())));
			}
			
			if (queryInstance.getQuery().getLatestTime() != null && startTime.after(queryInstance.getQuery().getLatestTime())) {
				
				otherStartTimeErrors = true;
				errors.add(new ValidationError("StartTimeTooLate", TimeUtils.TIME_FORMATTER.format(queryInstance.getQuery().getLatestTime())));
			}
			
			if (!otherStartTimeErrors && queryInstance.getQuery().getTimeInterval() != null && queryInstance.getQuery().getTimeInterval() >= 1) {
				
				Calendar calendar = Calendar.getInstance();
				calendar.setTime(startTime);
				
				int minutes = calendar.get(Calendar.MINUTE);
				int interval = queryInstance.getQuery().getTimeInterval();
				int mod = minutes % interval;
				
				if (mod != 0) {
					
					calendar.set(Calendar.MINUTE, minutes - mod);
					Time newStartTime = new Time(calendar.getTimeInMillis());
					
					errors.add(new TimeNotMultipleOfInterval("startTime", interval, startTime, newStartTime));
					
					HashMap<String, String[]> paramMap = (HashMap<String, String[]>) req.getParameterMap();
					paramMap.get(startTimeParameter)[0] = TimeUtils.TIME_FORMATTER.format(newStartTime);
				}
			}
		}
		
		if (endTime != null) {
			
			boolean otherEndTimeErrors = false;
			
			if (queryInstance.getQuery().getEarliestTime() != null && endTime.before(queryInstance.getQuery().getEarliestTime())) {
				
				otherEndTimeErrors = true;
				errors.add(new ValidationError("EndTimeTooEarly", TimeUtils.TIME_FORMATTER.format(queryInstance.getQuery().getEarliestTime())));
			}
			
			if (queryInstance.getQuery().getLatestTime() != null && endTime.after(queryInstance.getQuery().getLatestTime())) {
				
				otherEndTimeErrors = true;
				errors.add(new ValidationError("EndTimeTooLate", TimeUtils.TIME_FORMATTER.format(queryInstance.getQuery().getLatestTime())));
			}
			
			if (!otherEndTimeErrors && queryInstance.getQuery().getTimeInterval() != null && queryInstance.getQuery().getTimeInterval() >= 1) {
				
				Calendar calendar = Calendar.getInstance();
				calendar.setTime(endTime);
				
				int minutes = calendar.get(Calendar.MINUTE);
				int interval = queryInstance.getQuery().getTimeInterval();
				int mod = minutes % interval;
				
				if (mod != 0) {
					
					calendar.set(Calendar.MINUTE, minutes + (interval - mod));
					Time newEndTime = new Time(calendar.getTimeInMillis());
					
					errors.add(new TimeNotMultipleOfInterval("endTime", interval, endTime, newEndTime));
					
					HashMap<String, String[]> paramMap = (HashMap<String, String[]>) req.getParameterMap();
					paramMap.get(endTimeParameter)[0] = TimeUtils.TIME_FORMATTER.format(newEndTime);
				}
			}
		}
		
		
		if (!errors.isEmpty()) {

			throw new ValidationException(errors);
		}

		queryInstance.setStartDate(startDate);
		queryInstance.setEndDate(endDate);
		queryInstance.setStartTime(startTime);
		queryInstance.setEndTime(endTime);
		
		if (queryInstance.getQuery().isSetAsAttribute()) {
			
			queryInstance.setAttribute(attributeHandler);
		}
		
		queryInstance.getQueryInstanceDescriptor().setPopulated(true);

	}

	private void validateDate(Date startDate, Date endDate, Time startTime, Time endTime, DateTimeQueryInstance queryInstance, List<ValidationError> errors) {

		DateTimeQuery query = queryInstance.getQuery();
		
		//Check day of week
		if(startDate != null && query.getAllowedDaysOfWeek() != null) {

			Calendar startDateCalendar = Calendar.getInstance();
			startDateCalendar.setTime(startDate);

			int dayOfWeekInt = startDateCalendar.get(Calendar.DAY_OF_WEEK);
			
			DayOfWeek currentDayOfWeek = DayOfWeek.fromInt(dayOfWeekInt, true);
			
			if(!query.getAllowedDaysOfWeek().contains(currentDayOfWeek)) {
				
				errors.add(new ValidationError("StartDatOnNotAllowedWeekday", getAllowedWeekdayString(query.getAllowedDaysOfWeek())));
			}
		}
		
		//Check tomorrow-restriction
		if(startDate != null && query.isRestrictTomorrowOnTime()) {

			//First check that chosen date is after today
			LocalDate currentDate = LocalDate.now();
			
			long daysBetween = ChronoUnit.DAYS.between(currentDate, startDate.toLocalDate());
			
			if(TomorrowRestrictionMode.RESTRICT_TOMORROW.equals(query.getTomorrowRestrictionMode())) {

				if(daysBetween < 1) {
					
					errors.add(new ValidationError("RestrictTomorrowOnTimeNeedsStartDateAfterToday"));
				}

				//Check the case when tomorrows date is chosen
				if(daysBetween == 1) {
				
					//Check if chosen time is after allowed time
					LocalTime currentTime = LocalTime.now();
					
					if(currentTime.compareTo(query.getRestrictTomorrowTime().toLocalTime()) > 0 ) {

						errors.add(new ValidationError("TomorrowIsNotAllowedNow", TIME_FORMATTER.format(query.getRestrictTomorrowTime())));
					}
				}
			} else { //if(TomorrowRestrictionMode.RESTRICT_TODAY.equals(query.getTomorrowRestrictionMode())) {

				if(daysBetween < 0) {
					
					errors.add(new ValidationError("RestrictTomorrowOnTimeNeedsStartDateTodayOrAfter"));
				}

				//Check the case when today's date is chosen
				if(daysBetween == 0) {
				
					//Check if chosen time is after allowed time
					LocalTime currentTime = LocalTime.now();
					
					if(currentTime.compareTo(query.getRestrictTomorrowTime().toLocalTime()) > 0 ) {

						errors.add(new ValidationError("TodayIsNotAllowedNow", TIME_FORMATTER.format(query.getRestrictTomorrowTime())));
					}
				}
			}
		}

		//Check month-restrictions for start date
		if(startDate != null && query.isRestrictCurrentMonth()) {
			
			LocalDate compareDate = LocalDate.now();
		
			if(!query.isRestrictUpcomingMonths()) { //Check current-month restriction

				LocalDate lowerLimit = compareDate.with(firstDayOfMonth());
				LocalDate upperLimit = compareDate.with(lastDayOfMonth());
				
				if(!startDate.toLocalDate().isBefore(lowerLimit) && !startDate.toLocalDate().isAfter(upperLimit)) {
					
					errors.add(new ValidationError("CurrentMonthIsNotAllowed"));
				}
				
			} else  { //Check current and upcoming month restriction
			
				LocalDate lowerLimit = compareDate.with(firstDayOfMonth());
				
				LocalDate upperLimit = compareDate.plusMonths(query.getNumberOfUpcomingMonths());
				upperLimit = upperLimit.with(lastDayOfMonth());
				
				if(!startDate.toLocalDate().isBefore(lowerLimit) && !startDate.toLocalDate().isAfter(upperLimit)) {

					if(query.getNumberOfUpcomingMonths() == 1) {
					
						errors.add(new ValidationError("CurrentAndOneUpcomingMonthAreNotAllowed", query.getNumberOfUpcomingMonths().toString()));
					
					} else {
				
						errors.add(new ValidationError("CurrentAndSomeUpcomingMonthsAreNotAllowed", query.getNumberOfUpcomingMonths().toString()));
					}
				}
			}
		}

		//Check month-restrictions for end date
		if(endDate != null && query.isEndDateRestrictCurrentMonth()) {
			
			LocalDate compareDate = LocalDate.now();
		
			if(!query.isRestrictUpcomingEndMonths()) { //Check current-month restriction

				LocalDate lowerLimit = compareDate.with(firstDayOfMonth());
				LocalDate upperLimit = compareDate.with(lastDayOfMonth());
				
				if(!endDate.toLocalDate().isBefore(lowerLimit) && !endDate.toLocalDate().isAfter(upperLimit)) {
					
					errors.add(new ValidationError("CurrentEndMonthIsNotAllowed"));
				}
				
			} else  { //Check current and upcoming month restriction
			
				LocalDate lowerLimit = compareDate.with(firstDayOfMonth());
				
				LocalDate upperLimit = compareDate.plusMonths(query.getNumberOfUpcomingEndMonths());
				upperLimit = upperLimit.with(lastDayOfMonth());
				
				if(!endDate.toLocalDate().isBefore(lowerLimit) && !endDate.toLocalDate().isAfter(upperLimit)) {

					if(query.getNumberOfUpcomingEndMonths() == 1) {
					
						errors.add(new ValidationError("CurrentAndOneUpcomingEndMonthAreNotAllowed", query.getNumberOfUpcomingEndMonths().toString()));
					
					} else {
				
						errors.add(new ValidationError("CurrentAndSomeUpcomingEndMonthsAreNotAllowed", query.getNumberOfUpcomingEndMonths().toString()));
					}
				}
			}
		}
		
		//Minimum start date check
		if (startDate != null && queryInstance.getQuery().getMinStartDate() != null && startDate.before(queryInstance.getQuery().getMinStartDate())) {

			errors.add(new ValidationError("StartDateBeforeMinDate", DateUtils.DATE_FORMATTER.format(queryInstance.getQuery().getMinStartDate())));
		}

		//Maximum start date check
		if (startDate != null && queryInstance.getQuery().getMaxStartDate() != null && startDate.after(queryInstance.getQuery().getMaxStartDate())) {

			errors.add(new ValidationError("StartDateAfterMaxDate", DateUtils.DATE_FORMATTER.format(queryInstance.getQuery().getMaxStartDate())));
		}
		
		//Minimum end date check
		if (endDate != null && queryInstance.getQuery().getMinEndDate() != null && endDate.before(queryInstance.getQuery().getMinEndDate())) {

			errors.add(new ValidationError("EndDateBeforeMinDate", DateUtils.DATE_FORMATTER.format(queryInstance.getQuery().getMinEndDate())));
		}		

		//Maximum end date check
		if (endDate != null && queryInstance.getQuery().getMaxEndDate() != null && endDate.after(queryInstance.getQuery().getMaxEndDate())) {

			errors.add(new ValidationError("EndDateAfterMaxDate", DateUtils.DATE_FORMATTER.format(queryInstance.getQuery().getMaxEndDate())));
		}		
		
		//Date interval check
		if(queryInstance.getQuery().isRestrictDateInterval() && startDate != null && endDate != null) {

			LocalDate startDateLocal = startDate.toLocalDate();
			LocalDate endDateLocal = endDate.toLocalDate();
			
			long diffDays = ChronoUnit.DAYS.between(startDateLocal, endDateLocal);
			long diffMonths = ChronoUnit.MONTHS.between(startDateLocal, endDateLocal);
			long remainderDays = Period.between(startDateLocal, endDateLocal).getDays();
			
			if(queryInstance.getQuery().getDateIntervalRestrictionType().equals(DateIntervalRestrictionType.INTERVAL_MINIMUM)) {

				if(queryInstance.getQuery().getDateIntervalRestrictionUnit().equals(DateIntervalRestrictionUnit.MONTHS)) {
					
					if(queryInstance.getQuery().getDateIntervalRestriction() > diffMonths) {
						
						if(queryInstance.getQuery().getDateIntervalRestriction() > 11) {
							
							int years = queryInstance.getQuery().getDateIntervalRestriction() / 12;
							int months = queryInstance.getQuery().getDateIntervalRestriction() % 12;
							
							errors.add(new DateIntervallValidationError("DateIntervalYearsToSmall", years, months));
							
						} else {
							
							errors.add(new ValidationError("DateIntervalMonthsToSmall", "" + queryInstance.getQuery().getDateIntervalRestriction()));
						}
					}
					
				} else if(queryInstance.getQuery().getDateIntervalRestrictionUnit().equals(DateIntervalRestrictionUnit.DAYS)) {
					
					if(queryInstance.getQuery().getDateIntervalRestriction() > diffDays) {
						
						errors.add(new ValidationError("DateIntervalDaysToSmall", "" + queryInstance.getQuery().getDateIntervalRestriction()));
					}
				}
				
			} else if(queryInstance.getQuery().getDateIntervalRestrictionType().equals(DateIntervalRestrictionType.INTERVAL_MAXIMUM)) {
				
				if(queryInstance.getQuery().getDateIntervalRestrictionUnit().equals(DateIntervalRestrictionUnit.MONTHS)) {
					
					if(queryInstance.getQuery().getDateIntervalRestriction() < diffMonths || (queryInstance.getQuery().getDateIntervalRestriction().longValue() == diffMonths && remainderDays > 0)) {

						if(queryInstance.getQuery().getDateIntervalRestriction() > 11) {
							
							int years = queryInstance.getQuery().getDateIntervalRestriction() / 12;
							int months = queryInstance.getQuery().getDateIntervalRestriction() % 12;
							
							errors.add(new DateIntervallValidationError("DateIntervalYearsToBig", years, months));
							
						} else {
							
							errors.add(new ValidationError("DateIntervalMonthsToBig", "" + queryInstance.getQuery().getDateIntervalRestriction()));
						}
						
					}
					
				} else if(queryInstance.getQuery().getDateIntervalRestrictionUnit().equals(DateIntervalRestrictionUnit.DAYS)) {
					
					if(queryInstance.getQuery().getDateIntervalRestriction() < diffDays) {
						
						errors.add(new ValidationError("DateIntervalDaysToBig", "" + queryInstance.getQuery().getDateIntervalRestriction()));
					}
				}
			}
		}
		
		//Start date restriction check
		if (startDate != null && queryInstance.getQuery().getRestrictedStartDays() != null && queryInstance.getQuery().getStartDateRestrictionType() != null) {
			
			LocalDate tempStartDate = startDate.toLocalDate();
			LocalDate now = LocalDate.now();
			
			if(queryInstance.getQuery().getStartDateRestrictionType().equals(DateRestrictionType.RESTRICT_FUTURE_DATES)) {
				
				long daysBetween = ChronoUnit.DAYS.between(now, tempStartDate);
				
				if(daysBetween < queryInstance.getQuery().getRestrictedStartDays()) {
				
					if(queryInstance.getQuery().getRestrictedStartDays() > 0) {
					
						errors.add(new ValidationError("StartDateNotBigEnoughPositive", "" + queryInstance.getQuery().getRestrictedStartDays()));
						
					} else if(queryInstance.getQuery().getRestrictedStartDays() == 0) {
						
						errors.add(new ValidationError("StartDateNotBigEnoughToday"));
						
					} else { //Negative
						
						errors.add(new ValidationError("StartDateNotBigEnoughNegative", "" + Math.abs(queryInstance.getQuery().getRestrictedStartDays())));
					}
				}
			} else if(queryInstance.getQuery().getStartDateRestrictionType().equals(DateRestrictionType.RESTRICT_PAST_DATES)) {
				
				long daysBetween = ChronoUnit.DAYS.between(tempStartDate, now);
				
				if(daysBetween < queryInstance.getQuery().getRestrictedStartDays()) {
					
					if(queryInstance.getQuery().getRestrictedStartDays() > 0) {
						
						errors.add(new ValidationError("StartDateTooBigPositive", "" + queryInstance.getQuery().getRestrictedStartDays()));
					}
					else if(queryInstance.getQuery().getRestrictedStartDays() == 0) {
						
						errors.add(new ValidationError("StartDateTooBigToday"));
						
					} else { //Negative
						
						errors.add(new ValidationError("StartDateTooBigNegative", "" + Math.abs(queryInstance.getQuery().getRestrictedStartDays())));
					}
				}
				
			}
		}

		//End date restriction check
		if (endDate != null && queryInstance.getQuery().getRestrictedEndDays() != null && queryInstance.getQuery().getEndDateRestrictionType() != null) {

			LocalDate tempEndDate = endDate.toLocalDate();
			LocalDate now = LocalDate.now();
			
			if(queryInstance.getQuery().getEndDateRestrictionType().equals(DateRestrictionType.RESTRICT_FUTURE_DATES)) {
				
				long daysBetween = ChronoUnit.DAYS.between(now, tempEndDate);
				
				if(daysBetween < queryInstance.getQuery().getRestrictedEndDays()) {
				
					if(queryInstance.getQuery().getRestrictedEndDays() > 0) {
					
						errors.add(new ValidationError("EndDateNotBigEnoughPositive", "" + queryInstance.getQuery().getRestrictedEndDays()));
						
					} else if(queryInstance.getQuery().getRestrictedEndDays() == 0) {
						
						errors.add(new ValidationError("EndDateNotBigEnoughToday"));
						
					} else { //Negative
						
						errors.add(new ValidationError("EndDateNotBigEnoughNegative", "" + Math.abs(queryInstance.getQuery().getRestrictedEndDays())));
					}
				}
			} else if(queryInstance.getQuery().getEndDateRestrictionType().equals(DateRestrictionType.RESTRICT_PAST_DATES)) {
				
				long daysBetween = ChronoUnit.DAYS.between(tempEndDate, now);
				
				if(daysBetween < queryInstance.getQuery().getRestrictedEndDays()) {
					
					if(queryInstance.getQuery().getRestrictedEndDays() > 0) {
						
						errors.add(new ValidationError("EndDateTooBigPositive", "" + queryInstance.getQuery().getRestrictedEndDays()));
					}
					else if(queryInstance.getQuery().getRestrictedEndDays() == 0) {
						
						errors.add(new ValidationError("EndDateTooBigToday"));
						
					} else { //Negative
						
						errors.add(new ValidationError("EndDateTooBigNegative", "" + Math.abs(queryInstance.getQuery().getRestrictedEndDays())));
					}
				}
			}
		}
	}
	
	private String getAllowedWeekdayString(List<DayOfWeek> allowedDaysOfWeek) {

		StringBuilder allowedDaysOfWeekString = new StringBuilder();
		
		if(allowedDaysOfWeek.contains(DayOfWeek.MONDAY)) {
			
			allowedDaysOfWeekString.append(mondayString);
		}
		if(allowedDaysOfWeek.contains(DayOfWeek.TUESDAY)) {
			
			if(allowedDaysOfWeekString.length() > 0) {
				
				allowedDaysOfWeekString.append(", ");
			}
			
			allowedDaysOfWeekString.append(tuesdayString);
		}
		if(allowedDaysOfWeek.contains(DayOfWeek.WEDNESDAY)) {
			
			if(allowedDaysOfWeekString.length() > 0) {
				
				allowedDaysOfWeekString.append(", ");
			}
			
			allowedDaysOfWeekString.append(wednesdayString);
		}
		if(allowedDaysOfWeek.contains(DayOfWeek.THURSDAY)) {
			
			if(allowedDaysOfWeekString.length() > 0) {
				
				allowedDaysOfWeekString.append(", ");
			}
			
			allowedDaysOfWeekString.append(thursdayString);
		}
		if(allowedDaysOfWeek.contains(DayOfWeek.FRIDAY)) {
			
			if(allowedDaysOfWeekString.length() > 0) {
				
				allowedDaysOfWeekString.append(", ");
			}
			
			allowedDaysOfWeekString.append(fridayString);
		}
		if(allowedDaysOfWeek.contains(DayOfWeek.SATURDAY)) {
			
			if(allowedDaysOfWeekString.length() > 0) {
				
				allowedDaysOfWeekString.append(", ");
			}
			
			allowedDaysOfWeekString.append(saturdayString);
		}
		if(allowedDaysOfWeek.contains(DayOfWeek.SUNDAY)) {
			
			if(allowedDaysOfWeekString.length() > 0) {
				
				allowedDaysOfWeekString.append(", ");
			}
			
			allowedDaysOfWeekString.append(sundayString);
		}
		
		return allowedDaysOfWeekString.toString();
	}

	@WebPublic(alias = "config")
	public ForegroundModuleResponse configureQuery(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return this.queryCRUD.update(req, res, user, uriParser);
	}

	@Override
	public boolean deleteQuery(ImmutableQueryDescriptor descriptor, TransactionHandler transactionHandler) throws SQLException {

		DateTimeQuery query = getQuery(descriptor.getQueryID());

		if (query == null) {

			return false;
		}

		this.queryDAO.delete(query, transactionHandler);

		return true;
	}

	@Override
	public boolean deleteQueryInstance(ImmutableQueryInstanceDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		DateTimeQueryInstance queryInstance = this.getQueryInstance(descriptor.getQueryInstanceID());

		if (queryInstance == null) {

			return false;
		}

		this.queryInstanceDAO.delete(queryInstance, transactionHandler);

		return true;
	}

	@Override
	public String getTitlePrefix() {

		return this.moduleDescriptor.getName();
	}

	@Override
	public void copyQuery(MutableQueryDescriptor sourceQueryDescriptor, MutableQueryDescriptor copyQueryDescriptor, TransactionHandler transactionHandler, Map<Integer, ImmutableStatus> statusConversionMap) throws SQLException {

		DateTimeQuery query = getQuery(sourceQueryDescriptor.getQueryID(), transactionHandler);

		query.setQueryID(copyQueryDescriptor.getQueryID());

		queryDAO.add(query, transactionHandler, null);
	}

	@Override
	protected void appendPDFData(Document doc, Element showQueryValuesElement, DateTimeQueryInstance queryInstance, AttributeHandler attributeHandler) {

		super.appendPDFData(doc, showQueryValuesElement, queryInstance, attributeHandler);

		if (queryInstance.getQuery().getDescription() != null) {

			XMLUtils.appendNewCDATAElement(doc, showQueryValuesElement, "Description", JTidyUtils.getXHTML(queryInstance.getQuery().getDescription(attributeHandler), systemInterface.getEncoding()));
			XMLUtils.appendNewCDATAElement(doc, showQueryValuesElement, "isHTMLDescription", queryInstance.getQuery().getDescription().contains("<") && queryInstance.getQuery().getDescription().contains(">"));
		}
	}

	@Override
	protected Class<DateTimeQueryInstance> getQueryInstanceClass() {

		return DateTimeQueryInstance.class;
	}

	public boolean isValidForSubmit(MutableAttributeHandler attributeHandler, DateTimeQueryInstance dateTimeQueryInstance, User poster) {

		List<ValidationError> errors = new ArrayList<ValidationError>(0);
		
		validateDate(dateTimeQueryInstance.getStartDate(), dateTimeQueryInstance.getEndDate(), dateTimeQueryInstance.getStartTime(), dateTimeQueryInstance.getEndTime(), dateTimeQueryInstance, errors);
		
		if(CollectionUtils.isEmpty(errors)) {
			
			return true;
		}
		
		return false;
	}
}
