package com.nordicpeak.flowengine.integration.efecte;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.regex.Pattern;

import javax.mail.BodyPart;
import javax.mail.Flags.Flag;
import javax.mail.Folder;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Multipart;
import javax.mail.Part;
import javax.mail.Session;
import javax.mail.Store;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMultipart;
import javax.sql.DataSource;

import org.apache.log4j.Level;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.safety.Whitelist;

import se.unlogic.cron4jutils.CronStringValidator;
import se.unlogic.emailutils.framework.SimpleAuthenticator;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.PasswordSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.openhierarchy.foregroundmodules.siteprofile.interfaces.SiteProfile;
import se.unlogic.openhierarchy.foregroundmodules.siteprofile.interfaces.SiteProfileHandler;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.fileattachments.FileAttachmentHandler;
import se.unlogic.standardutils.streams.StreamUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.PositiveStringIntegerValidator;
import se.unlogic.standardutils.validation.TCPPortStringFormatValidator;

import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.MessageHandler;
import com.nordicpeak.flowengine.beans.ExternalMessage;
import com.nordicpeak.flowengine.beans.ExternalMessageAttachment;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceAttribute;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.utils.FlowEngineFileAttachmentUtils;

import it.sauronsoftware.cron4j.Scheduler;

public class EfecteMailCollectorModule extends AnnotatedForegroundModule implements Runnable {

	public static final Field[] LIST_EXCLUDED_FIELDS = { FlowInstanceEvent.FLOW_INSTANCE_RELATION };

	private static final Pattern SUBJECT_PATTERN_1 = Pattern.compile(".*\\[#ID:CASE-[0-9]+#\\].*");
	private static final Pattern SUBJECT_PATTERN_2 = Pattern.compile(".*\\[#ID:IT-[0-9]+#\\].*");
	private static final Pattern SUBJECT_PATTERN_3 = Pattern.compile(".*\\[#ID:CHG-[0-9]+#\\].*");
	private static final Pattern SUBJECT_PATTERN_4 = Pattern.compile(".*\\[#ID:PKG-[0-9]+#\\].*");

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Email check interval", description = "How often this module should check for new email messages from Efecte (specified in crontab format)", required = true, formatValidator = CronStringValidator.class)
	private String emailCheckInterval = "*/5 * * * *";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "IMAP server address", description = "The address to the IMAP server", required = true)
	private String imapServerAddress;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "IMAP server port", description = "The port of the IMAP server", formatValidator = TCPPortStringFormatValidator.class, required = true)
	private Integer imapServerPort = 993;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "IMAP server protocol", description = "The protocol of the IMAP server", required = true)
	private String imapServerProtocol = "imaps";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "IMAP folder", description = "The folder name on the IMAP server", required = true)
	private String imapFolderName = "INBOX";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "IMAP server username", description = "The username for the IMAP server", required = true)
	private String imapServerUsername;

	@ModuleSetting
	@PasswordSettingDescriptor(name = "IMAP server password", description = "The password for the IMAP server", required = true)
	private String imapServerPassword;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Allowed error count", description = "The max number of allowed errors before logging an error instead of a warning", formatValidator = PositiveStringIntegerValidator.class, required = true)
	private Integer allowedErrorCount = 5;

	@InstanceManagerDependency
	protected SiteProfileHandler profileHandler;

	@InstanceManagerDependency
	private FileAttachmentHandler fileAttachmentHandler;

	@InstanceManagerDependency(required = true)
	protected MessageHandler messageHandler;

	private AnnotatedDAO<FlowInstanceAttribute> flowInstanceAttributeDAO;

	private QueryParameterFactory<FlowInstanceAttribute, String> flowInstanceAttributeNameParamFactory;
	private QueryParameterFactory<FlowInstanceAttribute, String> flowInstanceAttributeValueParamFactory;

	private Scheduler scheduler;

	private SimpleAuthenticator authenticator;

	private int currentErrorCount;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		FlowEngineDAOFactory daoFactory = new FlowEngineDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler());

		flowInstanceAttributeDAO = daoFactory.getFlowInstanceAttributeDAO();

		flowInstanceAttributeNameParamFactory = flowInstanceAttributeDAO.getParamFactory("name", String.class);
		flowInstanceAttributeValueParamFactory = flowInstanceAttributeDAO.getParamFactory("value", String.class);
	}

	@Override
	protected void moduleConfigured() throws Exception {

		stopScheduler();

		if (ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.WARN)) {

			authenticator = new SimpleAuthenticator(imapServerUsername, imapServerPassword);

			initScheduler();

		}
	}

	protected synchronized void initScheduler() {

		log.info("Starting scheduler...");

		scheduler = new Scheduler(systemInterface.getApplicationName() + " - " + moduleDescriptor.toString());
		scheduler.setDaemon(true);

		scheduler.schedule(this.emailCheckInterval, this);
		scheduler.start();
	}

	protected synchronized void stopScheduler() {

		log.info("Stopping scheduler...");

		try {

			if (scheduler != null) {

				scheduler.stop();
				scheduler = null;
			}

		} catch (IllegalStateException e) {
			log.error("Error stopping scheduler", e);
		}
	}

	@Override
	public void unload() throws Exception {

		stopScheduler();

		super.unload();
	}

	@Override
	public synchronized void run() {

		log.debug("Module " + moduleDescriptor + " checking for new emails...");

		//		if(!hasRequiredDependencies) {
		//			
		//			log.warn("Required dependencies are not set, aborting check for new emails");
		//			return;
		//		}

		try {
			Properties smtpProps = new Properties();
			smtpProps.put("mail.host", imapServerAddress);
			smtpProps.put("mail.port", imapServerPort);
			smtpProps.put("mail.transport.protocol", imapServerProtocol);
			smtpProps.put("mail.smtp.timeout", "5000");
			smtpProps.put("mail.smtps.timeout", "5000");
			smtpProps.put("mail.smtp.connectiontimeout", "5000");
			smtpProps.put("mail.smtps.connectiontimeout", "5000");
			smtpProps.put("mail.imap.timeout", "5000");
			smtpProps.put("mail.imaps.timeout", "5000");

			Session session = Session.getInstance(smtpProps, authenticator);

			Store store = session.getStore(imapServerProtocol);
			store.connect();

			Folder inbox = null;

			try {

				inbox = store.getFolder(imapFolderName);
				inbox.open(Folder.READ_WRITE);

				Message messages[] = inbox.getMessages();

				if (messages != null) {

					for (Message message : messages) {

						if (message.getSubject() != null && (SUBJECT_PATTERN_1.matcher(message.getSubject()).matches() || SUBJECT_PATTERN_2.matcher(message.getSubject()).matches() || SUBJECT_PATTERN_3.matcher(message.getSubject()).matches() || SUBJECT_PATTERN_4.matcher(message.getSubject()).matches())) {

							String efecteID = StringUtils.substringBetween(message.getSubject(), "[#ID:", "#]");

							FlowInstance flowInstance = getFlowInstanceByAttribute(efecteID);

							if (flowInstance == null) {

								log.warn("Unable to find flow instance with " + Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE + " attribute matching " + efecteID);

								message.setFlag(Flag.DELETED, true);

								continue;
							}
							
							String messageText = getMessageText(message);

							if (StringUtils.isEmpty(messageText)) {

								log.warn("Unable to text content from message with subject " + message.getSubject());

								continue;
							}
							
							log.info("Processing message " + message.getSubject());

							ExternalMessage externalMessage = new ExternalMessage();

							externalMessage.setAdded(new Timestamp(message.getReceivedDate().getTime()));
							externalMessage.setFlowInstance(flowInstance);
							externalMessage.setMessage(messageText);
							externalMessage.setPostedByManager(true);
							externalMessage.setPoster(null);
							externalMessage.setSystemMessage(true);

							if (message.getContent() instanceof Multipart) {

								List<ExternalMessageAttachment> attachments = new ArrayList<ExternalMessageAttachment>();

								Multipart multiPart = (Multipart) message.getContent();

								for (int i = 0; i < multiPart.getCount(); i++) {

									MimeBodyPart part = (MimeBodyPart) multiPart.getBodyPart(i);

									if (Part.ATTACHMENT.equalsIgnoreCase(part.getDisposition())) {

										if (StringUtils.isEmpty(part.getFileName())) {

											log.warn("Mime body part without filename received, skipping.");

											continue;
										}

										log.info("Downloading attachment " + part.getFileName());
										ByteArrayOutputStream outputStream = new ByteArrayOutputStream();

										StreamUtils.transfer(part.getInputStream(), outputStream);

										byte[] data = outputStream.toByteArray();

										ExternalMessageAttachment attachment = FlowEngineFileAttachmentUtils.createAttachment(fileAttachmentHandler, part.getFileName(), data.length, data, ExternalMessageAttachment.class);

										attachments.add(attachment);
									}
								}

								if (!attachments.isEmpty()) {

									externalMessage.setAttachments(attachments);
								}
							}

							messageHandler.add(externalMessage);

							message.setFlag(Flag.DELETED, true);
						}
					}
				}

				inbox.close(true);
				inbox = null;

				currentErrorCount = 0;

			} finally {

				if (inbox != null && inbox.isOpen()) {
					inbox.close(false);
				}

				store.close();
			}

		} catch (SQLException e) {

			log.error("Error processing emails from Efecte inbox", e);

		} catch (MessagingException | IOException e) {

			currentErrorCount++;

			if (currentErrorCount > allowedErrorCount) {

				log.error("Error getting emails from Efecte inbox", e);

			} else {

				log.warn("Error getting emails from Efecte inbox", e);
			}

		} catch (Throwable e) {

			log.error("Error processing emails from Efecte inbox", e);
		}
	}

	private FlowInstance getFlowInstanceByAttribute(String efecteID) throws SQLException {

		HighLevelQuery<FlowInstanceAttribute> query = new HighLevelQuery<FlowInstanceAttribute>(FlowInstanceAttribute.FLOW_INSTANCE_RELATION, FlowInstance.STATUS_RELATION, FlowInstance.FLOW_RELATION, FlowInstance.OWNERS_RELATION, Flow.FLOW_FAMILY_RELATION);

		query.addParameter(flowInstanceAttributeNameParamFactory.getParameter(Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE));
		query.addParameter(flowInstanceAttributeValueParamFactory.getParameter(efecteID));

		FlowInstanceAttribute flowInstanceAttribute = flowInstanceAttributeDAO.get(query);

		if (flowInstanceAttribute != null) {

			return flowInstanceAttribute.getFlowInstance();
		}

		return null;
	}

	private static String getMessageText(Message message) throws MessagingException, IOException {

		String result = "";

		if (message.isMimeType("text/plain")) {

			result = message.getContent().toString();

		}
		if (message.isMimeType("text/html")) {

			result = result + "\n" + parseHTML(message.getContent().toString());

		} else if (message.isMimeType("multipart/*")) {

			MimeMultipart mimeMultipart = (MimeMultipart) message.getContent();
			result = getTextFromMimeMultipart(mimeMultipart);
		}

		return StringUtils.substringBefore(result, "---------- Original Message ----------").replace("&nbsp;", "").replace("Vid svar p detta mail s uppdateras ditt rende via automatik. Viktigt att tnka p r att ha rendenumret kvar i mnesraden Tex: [IT-xxxx] .", "");
	}

	private static String getTextFromMimeMultipart(MimeMultipart mimeMultipart) throws MessagingException, IOException {

		String result = "";

		int count = mimeMultipart.getCount();

		for (int i = 0; i < count; i++) {

			BodyPart bodyPart = mimeMultipart.getBodyPart(i);

			if (bodyPart.isMimeType("text/plain")) {

				result = result + "\n" + bodyPart.getContent();
				break;

			} else if (bodyPart.isMimeType("text/html")) {

				result = result + "\n" + parseHTML(bodyPart.getContent().toString());

			} else if (bodyPart.getContent() instanceof MimeMultipart) {

				result = result + getTextFromMimeMultipart((MimeMultipart) bodyPart.getContent());
			}
		}

		return result;
	}

	private static String parseHTML(String html) {

		Document document = Jsoup.parse(html);
		document.outputSettings(new Document.OutputSettings().prettyPrint(false));
		document.select("br").append("\\r");
		String s = document.html().replaceAll("\\\\r", "\r");
		return Jsoup.clean(s, "", Whitelist.none(), new Document.OutputSettings().prettyPrint(false));
	}

	public SiteProfile getSiteProfile(ImmutableFlowInstance flowInstance) {

		if (profileHandler != null) {

			SiteProfile profile = profileHandler.getProfile(flowInstance.getProfileID());

			if (profile != null && flowInstance.getFlow().getFlowFamily() != null) {

				SiteProfile subProfile = profileHandler.getMatchingSubProfile(profile.getProfileID(), "flowfamily-" + flowInstance.getFlow().getFlowFamily().getFlowFamilyID());

				if (subProfile != null) {

					return subProfile;
				}
			}

			return profile;
		}

		return null;
	}
}
