package com.nordicpeak.flowengine.externalalternativelist;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.hierarchy.core.interfaces.menu.MenuItemDescriptor;
import se.unlogic.hierarchy.core.utils.AttributeTagUtils;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.standardutils.annotations.SplitOnLineBreak;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAOFactory;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.string.AnnotatedBeanTagSourceFactory;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.string.TagReplacer;
import se.unlogic.standardutils.string.URLEncodingTagReplacer;
import se.unlogic.standardutils.time.MillisecondTimeUnits;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;

import com.nordicpeak.flowengine.alternativelist.beans.AlternativeListDescriptor;
import com.nordicpeak.flowengine.alternativelist.interfaces.AlternativeListInstance;
import com.nordicpeak.flowengine.alternativelist.provider.BaseAlternativeListProvider;
import com.nordicpeak.flowengine.interfaces.APISource;
import com.nordicpeak.flowengine.interfaces.APISourceHandler;
import com.nordicpeak.flowengine.interfaces.APISourceProvider;
import com.nordicpeak.flowengine.utils.UserAttributeTagUtils;

import flexjson.JSONDeserializer;

public class ExternalAlternativeListProviderModule extends BaseAlternativeListProvider<ListSource, ExternalAlternativeListInstance, ListSourceCRUD, ExternalAlternativeListProviderModule, ListSourceEvent> implements APISourceProvider {

	private static final AnnotatedBeanTagSourceFactory<User> USER_TAG_SOURCE_FACTORY = new AnnotatedBeanTagSourceFactory<User>(User.class, "$user.");
	
	@XSLVariable(prefix = "java.")
	private String apiSourceShortTypeName;
	
	@XSLVariable(prefix = "java.")
	private String apiSourceFullTypeName;
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "The maximum time in seconds to wait for a connection")
	private Integer connectionTimeout = 10;
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Read timeout", description = "The maximum time in seconds to wait for a response")
	private Integer readTimeout = 10;

	@ModuleSetting
	@SplitOnLineBreak
	@TextAreaSettingDescriptor(name = "Allowed encodings", description = "Encodings allowed to use")
	private List<String> allowedEncodings = new ArrayList<String>(Arrays.asList("ISO-8859-1", "UTF-8"));
	
	private APISourceHandler apiSourceHandler;
	
	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {
		
		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, ExternalAlternativeListProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));
		
		if (upgradeResult.isUpgrade()) {
			
			log.info(upgradeResult.toString());
		}
		
		AnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface);
		
		listDAO = daoFactory.getDAO(ListSource.class).getWrapper(Integer.class);
		listEventDAO = daoFactory.getDAO(ListSourceEvent.class);
		listEventListParamFactory = listEventDAO.getParamFactory(ListSourceEvent.LIST_RELATION, ListSource.class);
		
		listCRUD = new ListSourceCRUD(listDAO, new AnnotatedRequestPopulator<ListSource>(ListSource.class), this);
	}
	
	@InstanceManagerDependency
	public void setAPISourceHandler(APISourceHandler apiSourceHandler) throws SQLException {
		
		if (apiSourceHandler != null) {
			
			apiSourceHandler.addAPISourceProvider(this);
			
		} else if (this.apiSourceHandler != null) {
			
			this.apiSourceHandler.removeAPISourceProvider(this);
		}
		
		this.apiSourceHandler = apiSourceHandler;
		
		//Check if this module is fully cached, then trigger an update of the menu items
		if(sectionInterface.getForegroundModuleCache().isCached(moduleDescriptor)) {
			
			this.sectionInterface.getMenuCache().moduleUpdated(moduleDescriptor, this);
		}
	}	
	
	@Override
	public AlternativeListInstance getAlternativeListInstance(String name, AttributeHandler attributeHandler, User poster) {

		ExternalAlternativeListInstance listInstance = getCachedInstance(name);
		
		if(listInstance != null){
			
			return getAlternativeInstance(listInstance.getListSource(), attributeHandler, poster);
		}
		
		return null;
	}
	
	@Override
	public ExternalAlternativeListInstance getAlternativeInstance(ListSource source, AttributeHandler attributeHandler, User poster) {
		
		ExternalAlternativeListInstance listInstance = alternativeListCacheMap.get(source.getSourceID());
		
		if (listInstance != null) {

			if (source.isEnableAttributeFiltering() && attributeHandler != null) {
				
				String attributedURL = source.getDynamicURL();
				
				TagReplacer tagReplacer = new URLEncodingTagReplacer();
				
				tagReplacer.addTagSource(USER_TAG_SOURCE_FACTORY.getTagSource(poster));
				
				attributedURL = tagReplacer.replace(attributedURL);
				
				attributedURL = AttributeTagUtils.replaceTags(attributedURL, attributeHandler, false, true);
				
				if (poster != null) {
				
					attributedURL = UserAttributeTagUtils.replaceTags(attributedURL, poster, true);
				}
				
				List<ExternalAlternative> alternatives = getAlternatives(source, attributedURL);
				
				if (alternatives == null) {
					return null;
				}
				
				listInstance = new ExternalAlternativeListInstance(new AlternativeListDescriptor(providerID, source.getListName()), alternatives, System.currentTimeMillis(), source);
				
			} else if (source.getCacheRefreshTime() == null) {
				
				List<ExternalAlternative> alternatives = getAlternatives(source, null);
				
				if (alternatives == null) {
					return null;
				}
				
				listInstance = new ExternalAlternativeListInstance(new AlternativeListDescriptor(providerID, source.getListName()), alternatives, System.currentTimeMillis(), source);
				
			} else if (System.currentTimeMillis() - listInstance.getRetrivedTime() > source.getCacheRefreshTime() * MillisecondTimeUnits.MINUTE) {
				
				listInstance = cacheAlternatives(source);
			}
		}
		
		return listInstance;
	}
	
	protected List<ExternalAlternative> getAlternatives(ListSource source, String attributedURL) {

		try {

			SimpleRequest listRequest = null;

			if (source.isEnableAttributeFiltering() && attributedURL != null) {

				listRequest = new SimpleRequest(attributedURL);

			} else {

				listRequest = new SimpleRequest(source.getUrl());
			}

			log.info("Getting alternatives for list source " + source + " using URL " + listRequest.getEndpoint());
			
			listRequest.setConnectionTimeout(connectionTimeout * 1000);
			listRequest.setReadTimeout(readTimeout * 1000);

			if (source.getUseAuthentication()) {

				listRequest.setUsername(source.getUsername());
				listRequest.setPassword(source.getPassword());
			}

			String response = HTTPUtils.sendHTTPGetRequest(listRequest, source.getCharset()).getValue();

			if (!StringUtils.isEmpty(response)) {

				if (response.trim().startsWith("<")) {

					log.info("Parsing list " + source + " as XML");

					try {
						XMLParser parser = new XMLParser(XMLUtils.parseXML(response, source.getCharset().toString(), false, false));

						List<XMLParser> items;

						if (source.getAlternativeXPath() == null) {

							items = parser.getNodes("/Alternatives/Alternative");

						} else {

							items = parser.getNodes(source.getAlternativeXPathExpression(), true);
						}

						List<ExternalAlternative> alternatives = new ArrayList<ExternalAlternative>(items.size());

						if (items.size() > 0) {

							for (XMLParser item : items) {

								String key;

								if (source.getKeyXPath() == null) {

									key = item.getString("Key");

								} else {

									key = item.getString(source.getKeyXPathExpression());
								}

								String value;

								if (source.getValueXPath() == null) {

									value = item.getString("Value");

								} else {

									value = item.getString(source.getValueXPathExpression());
								}
								
								if (value != null) {
									value = StringUtils.substring(value, 255);
								}

								alternatives.add(new ExternalAlternative(key, value));
							}
						}

						return alternatives;

					} catch (Exception e) {

						log.warn("Error parsing alternatives XML for " + source, e);
					}

				} else {

					try {
						log.info("Parsing list " + source + " as JSON");

						JSONDeserializer<ArrayList<ExternalAlternative>> deserializer = new JSONDeserializer<ArrayList<ExternalAlternative>>();
						deserializer.use("values", ExternalAlternative.class);

						List<ExternalAlternative> alternatives = deserializer.deserialize(response, ArrayList.class);

						@SuppressWarnings("unused")
						ExternalAlternative alternative = alternatives.get(0); // test if valid list

						return alternatives;

					} catch (Exception e) {

						log.warn("Error parsing alternatives JSON for " + source, e);
					}
				}

			} else {

				log.warn("Got empty response for list " + source);
			}

		} catch (Exception e) {

			log.warn("Error getting alternatives for " + source, e);
		}

		if (source.getListID() != null && source.isKeepOldListOnRefreshFail()) {

			ExternalAlternativeListInstance listInstance = alternativeListCacheMap.get(source.getListID());

			if (listInstance != null) {

				return listInstance.getAlternatives();
			}
		}

		return null;
	}
	
	public boolean testSource(ListSource source) {
		
		List<ExternalAlternative> alternatives = getAlternatives(source, null);
		
		if (!CollectionUtils.isEmpty(alternatives)) {
			
			for (ExternalAlternative alternative : alternatives) {

				if (alternative.getAlternativeID() == null) {

					log.warn("Alternative found with empty alternative found in list " + source);
					
					return false;
				}
				
				if (StringUtils.isEmpty(alternative.getName())) {

					log.warn("Alternative with empty name found in list " + source);
					
					return false;
				}
			}
			
			return true;
		}
		
		log.warn("No alternatives found in list " + source);
		
		return false;
	}
	
	@Override
	protected ExternalAlternativeListInstance getNewInstance(AlternativeListDescriptor descriptor, ListSource list) {
		
 		return new ExternalAlternativeListInstance(descriptor, getAlternatives(list, null), System.currentTimeMillis(), list);
	}
	
	@Override
	protected ListSourceEvent getNewListEvent(ListSource list, Timestamp added, User poster, String message) {
		
		return new ListSourceEvent(list, added, poster, message);
	}

	public List<String> getAllowedEncodings() {

		return allowedEncodings;
	}

	@Override
	public List<? extends APISource> getAPISources() throws Exception {

		return listDAO.getAll();
	}
	
	@Override
	public String getShortTypeName() {

		return apiSourceShortTypeName;
	}
	
	@Override
	public String getFullTypeName() {

		return apiSourceFullTypeName;
	}
	
	@Override
	public String getAddURL() {

		return this.getFullAlias() + "/add";
	}

	@Override
	public String getBaseShowURL() {

		return this.getFullAlias() + "/show/";
	}
	
	@Override
	public String getBaseUpdateURL() {

		return this.getFullAlias() + "/update/";
	}

	@Override
	public String getBaseDeleteURL() {

		return this.getFullAlias() + "/delete/";
	}

	@Override
	public boolean isInUse(APISource apiSource) throws Exception {

		ListSource endpoint = (ListSource)apiSource;
		
		return super.isListInUse(endpoint);
	}

	public APISourceHandler getAPISourceHandler() {

		return apiSourceHandler;
	}
	
	@Override
	public List<? extends MenuItemDescriptor> getVisibleMenuItems() {

		if(this.apiSourceHandler != null) {
			
			return null;
		
		}else {
		
			return super.getVisibleMenuItems();
		}
	}
}
