package com.nordicpeak.flowengine.api;

import java.io.IOException;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.Date;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.foregroundmodules.rest.RESTMethod;
import se.unlogic.hierarchy.foregroundmodules.rest.URIParam;
import se.unlogic.hierarchy.foregroundmodules.rest.responsehandlers.XMLResponseHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.LowLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.fileattachments.FileAttachment;
import se.unlogic.standardutils.fileattachments.FileAttachmentHandler;
import se.unlogic.standardutils.xml.ElementableListener;
import se.unlogic.standardutils.xml.XMLGeneratorDocument;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.http.enums.ContentDisposition;

import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.beans.ExternalMessage;
import com.nordicpeak.flowengine.beans.ExternalMessageAttachment;
import com.nordicpeak.flowengine.beans.ExternalMessageReadReceipt;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.utils.FlowEngineFileAttachmentUtils;

public class ExternalMessageAPIModule extends APIModule implements ElementableListener<ExternalMessage>{

	@ModuleSetting
	@CheckboxSettingDescriptor(name="Force UTF-8 encoding", description="Forces UTF-8 encoding to be used instead of the system default encoding.")
	protected boolean forceUTF8;
	
	@InstanceManagerDependency(required = true)
	private FlowAdminModule flowAdminModule;

	@InstanceManagerDependency
	private FileAttachmentHandler fileAttachmentHandler;
	
	protected AnnotatedDAO<ExternalMessage> externalMessageDAO;
	protected AnnotatedDAO<ExternalMessageAttachment> attachmentDAO;

	protected QueryParameterFactory<ExternalMessage, Integer> messageIDParamFactory;
	protected QueryParameterFactory<ExternalMessageAttachment, Integer> attachmentIDParamFactory;

	@Override
	protected void moduleConfigured() throws Exception {

		String encoding;
		
		if(forceUTF8) {
			
			encoding = "UTF-8";
			
		}else {
			
			encoding = sectionInterface.getSystemInterface().getEncoding();
		}
		
		setResponseHandlers(CollectionUtils.getGenericSingletonList(new XMLResponseHandler(encoding)));
	}	

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		FlowEngineDAOFactory daoFactory = new FlowEngineDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler());

		externalMessageDAO = daoFactory.getExternalMessageDAO();
		attachmentDAO = daoFactory.getExternalMessageAttachmentDAO();

		messageIDParamFactory = externalMessageDAO.getParamFactory("messageID", Integer.class);
		attachmentIDParamFactory = attachmentDAO.getParamFactory("attachmentID", Integer.class);
	}

	@RESTMethod(alias = "getmessages/family/{flowFamilyID}", method = "get")
	public Document getMessagesByFamily(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowFamilyID") Integer flowFamilyID) throws Throwable {

		FlowFamily family = getFlowFamily(flowFamilyID);

		if (family == null) {

			throw new URINotFoundException(uriParser);
		}

		Date fromDate = getDate(req.getParameter("fromDate"), true);

		Date toDate = getDate(req.getParameter("toDate"), false);

		log.info("User " + user + " requesting external messages for flow family " + family + " from date " + fromDate + " to date " + toDate);

		List<ExternalMessage> messages = getMessages(family, fromDate, toDate);

		Document doc = XMLUtils.createDomDocument();

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(ExternalMessage.FLOWINSTANCE_RELATION);
		generatorDocument.addIgnoredField(ExternalMessageAttachment.DATA_FIELD);
		generatorDocument.addFieldElementableListener(ExternalMessage.class, this);
		
		doc.appendChild(new MessageResponse(messages).toXML(generatorDocument));
		
		return doc;
	}	
	
	@RESTMethod(alias = "getmessages/{flowID}", method = "get")
	public Document getMessages(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowID") Integer flowID) throws Throwable {

		Flow flow = (Flow) flowAdminModule.getFlow(flowID);

		if (flow == null) {

			throw new URINotFoundException(uriParser);
		}

		Date fromDate = getDate(req.getParameter("fromDate"), true);

		Date toDate = getDate(req.getParameter("toDate"), false);

		log.info("User " + user + " requesting external messages for flow " + flow + " from date " + fromDate + " to date " + toDate);

		List<ExternalMessage> messages = getMessages(flow, fromDate, toDate);

		Document doc = XMLUtils.createDomDocument();

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(ExternalMessage.FLOWINSTANCE_RELATION);
		generatorDocument.addIgnoredField(ExternalMessageAttachment.DATA_FIELD);
		generatorDocument.addFieldElementableListener(ExternalMessage.class, this);
		
		doc.appendChild(new MessageResponse(messages).toXML(generatorDocument));
		
		return doc;
	}
	
	@RESTMethod(alias = "getmessage/{messageID}", method = "get")
	public Document getMessage(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "messageID") Integer messageID) throws Throwable {

		ExternalMessage message = getMessage(messageID);

		if (message == null) {

			throw new URINotFoundException(uriParser);
		}

		log.info("User " + user + " requesting external message " + message);

		Document doc = XMLUtils.createDomDocument();

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(ExternalMessage.FLOWINSTANCE_RELATION);
		generatorDocument.addIgnoredField(ExternalMessageAttachment.DATA_FIELD);
		generatorDocument.addFieldElementableListener(ExternalMessage.class, this);
		
		doc.appendChild(message.toXML(generatorDocument));
		
		return doc;
	}

	protected ExternalMessage getMessage(Integer messageID) throws SQLException {

		HighLevelQuery<ExternalMessage> query = new HighLevelQuery<>();
		
		query.addRelations(ExternalMessage.FLOWINSTANCE_RELATION,ExternalMessage.ATTACHMENTS_RELATION, ExternalMessage.READ_RECEIPTS_RELATION, ExternalMessageReadReceipt.ATTACHMENT_DOWNLOADS_RELATION);
		
		query.addParameter(messageIDParamFactory.getParameter(messageID));
		
		return externalMessageDAO.get(query);
	}

	private List<ExternalMessage> getMessages(FlowFamily family, Date fromDate, Date toDate) throws SQLException {
	
		LowLevelQuery<ExternalMessage> query = new LowLevelQuery<ExternalMessage>();
		
		StringBuilder sqlBuilder = new StringBuilder("SELECT * FROM flowengine_external_messages INNER JOIN flowengine_flow_instances ON flowengine_external_messages.flowInstanceID = flowengine_flow_instances.flowInstanceID INNER JOIN flowengine_flows ON flowengine_flow_instances.flowID = flowengine_flows.flowID WHERE flowFamilyID = ?");
		
		query.addParameter(family.getFlowFamilyID());
		
		addMessageQueryParameters(sqlBuilder, query, fromDate, toDate);
		
		sqlBuilder.append(" ORDER BY flowengine_external_messages.added ASC");
		
		query.setSql(sqlBuilder.toString());
		
		query.addRelations(ExternalMessage.FLOWINSTANCE_RELATION,ExternalMessage.ATTACHMENTS_RELATION, ExternalMessage.READ_RECEIPTS_RELATION, ExternalMessageReadReceipt.ATTACHMENT_DOWNLOADS_RELATION);
		
		query.addExcludedRelations(ExternalMessageAttachment.DATA_FIELD);
		
		return externalMessageDAO.getAll(query);
	}
	
	private List<ExternalMessage> getMessages(Flow flow, Date fromDate, Date toDate) throws SQLException {

		LowLevelQuery<ExternalMessage> query = new LowLevelQuery<ExternalMessage>();
		
		StringBuilder sqlBuilder = new StringBuilder("SELECT * FROM flowengine_external_messages INNER JOIN flowengine_flow_instances ON flowengine_external_messages.flowInstanceID = flowengine_flow_instances.flowInstanceID WHERE flowID = ?");
		
		query.addParameter(flow.getFlowID());
		
		addMessageQueryParameters(sqlBuilder, query, fromDate, toDate);
		
		sqlBuilder.append(" ORDER BY flowengine_external_messages.added ASC");
		
		query.setSql(sqlBuilder.toString());
		
		query.addRelations(ExternalMessage.FLOWINSTANCE_RELATION,ExternalMessage.ATTACHMENTS_RELATION, ExternalMessage.READ_RECEIPTS_RELATION, ExternalMessageReadReceipt.ATTACHMENT_DOWNLOADS_RELATION);
		
		query.addExcludedRelations(ExternalMessageAttachment.DATA_FIELD);
		
		return externalMessageDAO.getAll(query);
	}

	protected void addMessageQueryParameters(StringBuilder sqlBuilder, LowLevelQuery<ExternalMessage> query, Date fromDate, Date toDate) {

		if (fromDate != null) {

			sqlBuilder.append(" AND flowengine_external_messages.added >= ?");
			query.addParameter(new Timestamp(fromDate.getTime()));
		}

		if (toDate != null) {

			sqlBuilder.append(" AND flowengine_external_messages.added <= ?");
			query.addParameter(new Timestamp(toDate.getTime()));
		}				
	}

	@RESTMethod(alias = "getattachment/{attachmentID}", method = "get")
	public Document getAttachment(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "attachmentID") Integer attachmentID) throws Throwable {

		ExternalMessageAttachment attachment = getAttachment(attachmentID);
		
		if(attachment == null){
			
			throw new URINotFoundException(uriParser);
		}
		
		log.info("User " + user + " requesting external message attachment " + attachment);
		
		if(attachment.getData() != null) {
			
			HTTPUtils.sendBlob(attachment.getData(), attachment.getFilename(), attachment.getAdded(), req, res, ContentDisposition.ATTACHMENT);
		
		}else {
		
			if(fileAttachmentHandler == null) {
				
				throw new IOException("Unable to find attachment data for external message attachment " + attachment + " requested by " + user + ", no file attachment handler available.");
			}
			
			FileAttachment fileAttachment = fileAttachmentHandler.getFileAttachment(FlowEngineFileAttachmentUtils.getAttachmentPath(attachment.getMessage(), attachment));
			
			if(fileAttachment == null) {
				
				throw new IOException("Unable to find attachment data for external message attachment " + attachment + " requested by " + user + ", no file found in file attachment handler.");
			}
			
			HTTPUtils.sendFile(fileAttachment.getInputStream(), attachment.getFilename(), attachment.getAdded(), req, res, ContentDisposition.ATTACHMENT, fileAttachment.getLength());
			
		}
		
		return null;
	}

	protected ExternalMessageAttachment getAttachment(Integer attachmentID) throws SQLException {

		HighLevelQuery<ExternalMessageAttachment> query = new HighLevelQuery<ExternalMessageAttachment>(attachmentIDParamFactory.getParameter(attachmentID));
		
		addAttachmentQueryParameters(query);
		
		return attachmentDAO.get(query);
	}

	protected void addAttachmentQueryParameters(HighLevelQuery<ExternalMessageAttachment> query) {

		query.addRelation(ExternalMessageAttachment.MESSAGE_RELATION);
		query.addRelation(ExternalMessage.FLOWINSTANCE_RELATION);
		query.addRelation(FlowInstance.FLOW_RELATION);
	}

	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		throw new URINotFoundException(uriParser);
	}

	@Override
	public void elementGenerated(Document doc, Element element, ExternalMessage externalMessage) {

		XMLUtils.appendNewElement(doc, element, "flowInstanceID", externalMessage.getFlowInstance().getFlowInstanceID());
	}
	
	protected FlowFamily getFlowFamily(Integer flowFamilyID) throws AccessDeniedException {

		return flowAdminModule.getFlowFamily(flowFamilyID);
	}
}
