package com.nordicpeak.flowengine.api;

import java.lang.reflect.Field;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.List;

import javax.naming.OperationNotSupportedException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;

import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.ModuleConfigurationException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceAttribute;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;

public class FamilyFilteredFlowInstanceAPIModule extends FlowInstanceAPIModule {

	@ModuleSetting(allowsNull = true)
	@TextAreaSettingDescriptor(name="Allowed flow families", description="Flow families that are allowed to be accessed via this module")
	protected List<Integer> allowedFlowFamilies;
	
	@Override
	protected FlowFamily getFlowFamily(Integer flowFamilyID) throws AccessDeniedException {

		FlowFamily family = super.getFlowFamily(flowFamilyID);
		
		if(family != null && !allowedFlowFamilies.contains(family.getFlowFamilyID())){
			
			throw new AccessDeniedException("Access to flow family " + family + " is not allowed via this module");
		}
		
		return family;
	}

	@Override
	protected Flow getFlow(Integer flowID) throws AccessDeniedException {

		Flow flow = super.getFlow(flowID);
		
		if(flow != null && !allowedFlowFamilies.contains(flow.getFlowFamily().getFlowFamilyID())){
			
			throw new AccessDeniedException("Access to flow family " + flow.getFlowFamily() + " is not allowed via this module");
		}
		
		return flow;
	}

	@Override
	protected FlowInstance getFlowInstance(Integer flowInstanceID, Field... relations) throws SQLException, AccessDeniedException {

		HighLevelQuery<FlowInstance> query = new HighLevelQuery<FlowInstance>();

		if(relations.length > 0) {
			
			query.addRelations(relations);
			query.addRelations(FlowInstance.FLOW_RELATION, Flow.FLOW_FAMILY_RELATION);
			
		}else {
			
			query.addRelations(FlowInstance.EVENTS_RELATION, FlowInstanceEvent.ATTRIBUTES_RELATION, FlowInstance.FLOW_RELATION, Flow.FLOW_FAMILY_RELATION, FlowInstance.ATTRIBUTES_RELATION);
		}		
		
		query.addParameter(flowInstanceIDParamFactory.getParameter(flowInstanceID));
		query.addParameter(firstSubmittedParamFactory.getIsNotNullParameter());

		FlowInstance flowInstance = flowInstanceDAO.get(query);
		
		if(flowInstance != null && !allowedFlowFamilies.contains(flowInstance.getFlow().getFlowFamily().getFlowFamilyID())){
			
			throw new AccessDeniedException("Access to flow family " + flowInstance.getFlow().getFlowFamily() + " is not allowed via this module");
		}
		
		return flowInstance;
	}

	@Override
	protected List<FlowInstanceAttribute> getFlowInstanceAttributes(String attribute, String value, Field... relations) throws SQLException {

		List<FlowInstanceAttribute> flowInstanceAttributes = super.getFlowInstanceAttributes(attribute, value, FlowInstanceAttribute.FLOW_INSTANCE_RELATION, FlowInstance.FLOW_RELATION, Flow.FLOW_FAMILY_RELATION);

		if(flowInstanceAttributes == null) {
			
			return null;
		}
		
		Iterator<FlowInstanceAttribute> iterator = flowInstanceAttributes.iterator();
		
		while(iterator.hasNext()) {
			
			FlowInstanceAttribute flowInstanceAttribute = iterator.next();
			
			if (!allowedFlowFamilies.contains(flowInstanceAttribute.getFlowInstance().getFlow().getFlowFamily().getFlowFamilyID())) {

				iterator.remove();
			}
		}
		
		if(!flowInstanceAttributes.isEmpty()){
			
			return flowInstanceAttributes;
		}
		
		return null;
	}

	@Override
	public ForegroundModuleResponse processRequest(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		if(allowedFlowFamilies == null){
			
			throw new ModuleConfigurationException("No allowed flow families have beent set.");
		}
		
		return super.processRequest(req, res, user, uriParser);
	}

	@Override
	public Document getFlowInstanceEventsByFlow(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, Integer flowID) throws Throwable {

		throw new OperationNotSupportedException("Method not supported in this version of the API");
	}

}
