package com.nordicpeak.flowengine.api;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.foregroundmodules.rest.RESTMethod;
import se.unlogic.hierarchy.foregroundmodules.rest.URIParam;
import se.unlogic.hierarchy.foregroundmodules.rest.responsehandlers.XMLResponseHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.LowLevelQuery;
import se.unlogic.standardutils.dao.QueryOperators;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.db.DBUtils;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.xml.XMLGeneratorDocument;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.http.enums.ContentDisposition;

import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.api.listeners.FlowInstanceEventTimestampMillisecondElementableListener;
import com.nordicpeak.flowengine.api.listeners.FlowInstanceTimestampMillisecondElementableListener;
import com.nordicpeak.flowengine.api.populators.ChainedFlowInstanceEventFamilyResultSetPopulator;
import com.nordicpeak.flowengine.api.populators.ChainedFlowInstanceEventResultSetPopulator;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceAttribute;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.beans.Status;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.interfaces.APIAccessController;
import com.nordicpeak.flowengine.interfaces.PDFProvider;
import com.nordicpeak.flowengine.interfaces.XMLProvider;
import com.nordicpeak.flowengine.utils.APIAccessUtils;

public class FlowInstanceAPIModule extends APIModule {

	private static final FlowInstanceTimestampMillisecondElementableListener MILLISECOND_LISTENER = new FlowInstanceTimestampMillisecondElementableListener();
	
	protected static final ChainedFlowInstanceEventResultSetPopulator CHAINED_FLOW_INSTANCE_EVENT_RESULT_SET_POPULATOR = new ChainedFlowInstanceEventResultSetPopulator();
	protected static final ChainedFlowInstanceEventFamilyResultSetPopulator CHAINED_FLOW_INSTANCE_EVENT_FAMILY_RESULT_SET_POPULATOR = new ChainedFlowInstanceEventFamilyResultSetPopulator();
	
	@ModuleSetting
	@CheckboxSettingDescriptor(name="Force UTF-8 encoding", description="Forces UTF-8 encoding to be used instead of the system default encoding.")
	protected boolean forceUTF8;
	
	@InstanceManagerDependency(required = true)
	private FlowAdminModule flowAdminModule;

	@InstanceManagerDependency(required = true)
	private XMLProvider xmlProvider;

	@InstanceManagerDependency(required = true)
	private PDFProvider pdfProvider;
	
	@InstanceManagerDependency(required = false)
	private APIAccessController apiAccessModule;

	protected AnnotatedDAO<FlowInstance> flowInstanceDAO;

	protected AnnotatedDAO<FlowInstanceAttribute> flowInstanceAttributeDAO;
	
	protected AnnotatedDAO<FlowInstanceEvent> flowInstanceEventDAO;
	
	protected AnnotatedDAO<Status> statusDAO;

	protected QueryParameterFactory<FlowInstance, Flow> flowQueryParamFactory;
	protected QueryParameterFactory<FlowInstance, Timestamp> firstSubmittedParamFactory;
	protected QueryParameterFactory<FlowInstance, Timestamp> lastStatusChangeParamFactory;
	protected QueryParameterFactory<FlowInstance, Integer> flowInstanceIDParamFactory;
	protected QueryParameterFactory<FlowInstance, Status> flowInstanceStatusParamFactory;
	
	protected QueryParameterFactory<FlowInstanceAttribute, String> attributeNameQueryParamFactory;
	protected QueryParameterFactory<FlowInstanceAttribute, String> attributeValueQueryParamFactory;
	
	protected QueryParameterFactory<Status, String> statusNameParamFactory;
	protected QueryParameterFactory<Status, Flow> statusFlowParamFactory;

	protected CopyOnWriteArrayList<XMLDocumentFilter> xmlDocumentFilters = new CopyOnWriteArrayList<XMLDocumentFilter>();

	@Override
	protected void moduleConfigured() throws Exception {

		String encoding;
		
		if(forceUTF8) {
			
			encoding = "UTF-8";
			
		}else {
			
			encoding = sectionInterface.getSystemInterface().getEncoding();
		}
		
		setResponseHandlers(CollectionUtils.getGenericSingletonList(new XMLResponseHandler(encoding)));
	}
	
	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		FlowEngineDAOFactory daoFactory = new FlowEngineDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler());

		flowInstanceDAO = daoFactory.getFlowInstanceDAO();

		flowQueryParamFactory = flowInstanceDAO.getParamFactory("flow", Flow.class);
		firstSubmittedParamFactory = flowInstanceDAO.getParamFactory("firstSubmitted", Timestamp.class);
		lastStatusChangeParamFactory = flowInstanceDAO.getParamFactory("lastStatusChange", Timestamp.class);
		flowInstanceIDParamFactory = flowInstanceDAO.getParamFactory("flowInstanceID", Integer.class);
		flowInstanceStatusParamFactory = flowInstanceDAO.getParamFactory("status", Status.class);

		flowInstanceAttributeDAO = daoFactory.getFlowInstanceAttributeDAO();

		attributeNameQueryParamFactory = flowInstanceAttributeDAO.getParamFactory("name", String.class);
		attributeValueQueryParamFactory = flowInstanceAttributeDAO.getParamFactory("value", String.class);
		
		flowInstanceEventDAO = daoFactory.getFlowInstanceEventDAO();
		
		statusDAO = daoFactory.getStatusDAO();
		
		statusNameParamFactory = statusDAO.getParamFactory("name", String.class);
		statusFlowParamFactory = statusDAO.getParamFactory("flow", Flow.class);
	}

	@RESTMethod(alias = "getinstances/family/{flowFamilyID}", method = "get")
	public Document getFlowInstancesByFamily(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowFamilyID") Integer flowFamilyID) throws Throwable {

		FlowFamily family = getFlowFamily(flowFamilyID);

		if (family == null) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, family, user);

		Date fromDate = getDate(req.getParameter("fromDate"), true);

		Date toDate = getDate(req.getParameter("toDate"), false);

		log.info("User " + user + " requesting flow instances for flow family " + family + " from date " + fromDate + " to date " + toDate);

		List<FlowInstance> flowInstances = getFlowInstances(flowAdminModule.getFlowVersions(family), fromDate, toDate);

		Document doc = XMLUtils.createDomDocument();

		Element flowInstancesElement = doc.createElement("FlowInstances");
		doc.appendChild(flowInstancesElement);

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(FlowInstance.SKIP_USER_ATTRIBUTES);
		generatorDocument.addRootElementableListener(FlowInstance.class, MILLISECOND_LISTENER);

		XMLUtils.append(generatorDocument, flowInstancesElement, flowInstances);

		return doc;
	}
	
	@RESTMethod(alias = "getinstances/family/{flowFamilyID}/{status}", method = "get")
	public Document getFlowInstancesByFamilyAndStatus(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowFamilyID") Integer flowFamilyID, @URIParam(name = "status") String statusName) throws Throwable {

		FlowFamily family = getFlowFamily(flowFamilyID);

		if (family == null || statusName == null) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, family, user);

		List<Status> statuses = getMatchingStatuses(statusName, family);

		if (statuses == null) {

			throw new URINotFoundException(uriParser);
		}

		Date fromDate = getDate(req.getParameter("fromDate"), true);

		Date toDate = getDate(req.getParameter("toDate"), false);

		log.info("User " + user + " requesting flow instances filtered by status for flow family " + family + " from date " + fromDate + " to date " + toDate);

		List<FlowInstance> flowInstances = getFlowInstances(flowAdminModule.getFlowVersions(family), statuses, fromDate, toDate);

		Document doc = XMLUtils.createDomDocument();

		Element flowInstancesElement = doc.createElement("FlowInstances");
		doc.appendChild(flowInstancesElement);

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(FlowInstance.SKIP_USER_ATTRIBUTES);
		generatorDocument.addRootElementableListener(FlowInstance.class, MILLISECOND_LISTENER);

		XMLUtils.append(generatorDocument, flowInstancesElement, flowInstances);

		return doc;
	}

	@RESTMethod(alias = "getinstances/{flowID}", method = "get")
	public Document getFlowInstancesByFlow(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowID") Integer flowID) throws Throwable {

		Flow flow = getFlow(flowID);

		if (flow == null) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, flow.getFlowFamily(), user);

		Date fromDate = getDate(req.getParameter("fromDate"), true);

		Date toDate = getDate(req.getParameter("toDate"), false);

		log.info("User " + user + " requesting flow instances for flow " + flow + " from date " + fromDate + " to date " + toDate);

		List<FlowInstance> flowInstances = getFlowInstances(flow, fromDate, toDate);

		Document doc = XMLUtils.createDomDocument();

		Element flowInstancesElement = doc.createElement("FlowInstances");
		doc.appendChild(flowInstancesElement);

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(FlowInstance.SKIP_USER_ATTRIBUTES);
		generatorDocument.addRootElementableListener(FlowInstance.class, MILLISECOND_LISTENER);
		
		XMLUtils.append(generatorDocument, flowInstancesElement, flowInstances);

		return doc;
	}
	
	@RESTMethod(alias = "getinstances/{flowID}/{status}", method = "get")
	public Document getFlowInstancesByFlowAndStatus(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowID") Integer flowID, @URIParam(name = "status") String statusName) throws Throwable {

		Flow flow = getFlow(flowID);

		if (flow == null || statusName == null) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, flow.getFlowFamily(), user);
		
		List<Status> statuses = getMatchingStatuses(statusName, flow);

		if(statuses == null) {
			
			throw new URINotFoundException(uriParser);
		}
		
		Date fromDate = getDate(req.getParameter("fromDate"), true);

		Date toDate = getDate(req.getParameter("toDate"), false);

		log.info("User " + user + " requesting flow instances filtered by status for flow " + flow + " from date " + fromDate + " to date " + toDate);

		List<FlowInstance> flowInstances = getFlowInstances(flow, statuses, fromDate, toDate);

		Document doc = XMLUtils.createDomDocument();

		Element flowInstancesElement = doc.createElement("FlowInstances");
		doc.appendChild(flowInstancesElement);

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(FlowInstance.SKIP_USER_ATTRIBUTES);
		generatorDocument.addRootElementableListener(FlowInstance.class, MILLISECOND_LISTENER);
		
		XMLUtils.append(generatorDocument, flowInstancesElement, flowInstances);

		return doc;
	}

	private List<Status> getMatchingStatuses(String statusName, Flow flow) throws SQLException {

		HighLevelQuery<Status> query = new HighLevelQuery<Status>();
		
		query.addParameter(statusFlowParamFactory.getParameter(flow));
		query.addParameter(statusNameParamFactory.getParameter(statusName));
		
		return statusDAO.getAll(query);
	}
	
	private List<Status> getMatchingStatuses(String statusName, FlowFamily flowFamily) throws SQLException {
		
		List<Flow> flows = flowAdminModule.getFlowVersions(flowFamily);
		
		if (CollectionUtils.isEmpty(flows)) {
			
			return null;
		}
		
		HighLevelQuery<Status> query = new HighLevelQuery<Status>();
		
		query.addParameter(statusFlowParamFactory.getWhereInParameter(flows));
		query.addParameter(statusNameParamFactory.getParameter(statusName));

		return statusDAO.getAll(query);
	}

	@RESTMethod(alias = "getinstancesbyattribute", method = "get")
	public Document getFlowInstancesByAttribute(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		String attribute = req.getParameter("attribute");
		String value = req.getParameter("value");

		if (attribute == null || value == null) {

			throw new URINotFoundException(uriParser);
		}

		log.info("User " + user + " requesting flow instances by attribute " + attribute + " with value " + value);

		List<FlowInstanceAttribute> flowInstanceAttributes = getFlowInstanceAttributes(attribute, value, FlowInstanceAttribute.FLOW_INSTANCE_RELATION, FlowInstance.FLOW_RELATION, Flow.FLOW_FAMILY_RELATION);

		Document doc = XMLUtils.createDomDocument();

		Element flowInstancesElement = doc.createElement("FlowInstances");
		doc.appendChild(flowInstancesElement);

		if (flowInstanceAttributes != null) {

			XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
			generatorDocument.addIgnoredFields(FlowInstance.SKIP_USER_ATTRIBUTES, FlowInstance.FLOW_RELATION);
			generatorDocument.addRootElementableListener(FlowInstance.class, MILLISECOND_LISTENER);

			for (FlowInstanceAttribute flowInstanceAttribute : flowInstanceAttributes) {

				if (flowInstanceAttribute.getFlowInstance().getFirstSubmitted() != null) {
					
					if (!APIAccessUtils.hasAccess(apiAccessModule, flowInstanceAttribute.getFlowInstance().getFlow().getFlowFamily(), user)) {
						continue;
					}

					flowInstancesElement.appendChild(flowInstanceAttribute.getFlowInstance().toXML(generatorDocument));
				}
			}
		}

		return doc;
	}
	
	@RESTMethod(alias = "getinstanceattributes/{flowInstanceID}", method = "get")
	public Document getFlowInstanceAttributes(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowInstanceID") Integer flowInstanceID) throws Throwable {

		FlowInstance flowInstance = getFlowInstance(flowInstanceID, FlowInstance.ATTRIBUTES_RELATION, FlowInstance.FLOW_RELATION, Flow.FLOW_FAMILY_RELATION);

		if (flowInstance == null) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, flowInstance.getFlow().getFlowFamily(), user);

		log.info("User " + user + " requesting flow instance attributes for flow instance " + flowInstance);

		Document doc = XMLUtils.createDomDocument();

		Element attributesElement = doc.createElement("FlowInstanceAttributes");
		doc.appendChild(attributesElement);

		XMLUtils.append(doc, attributesElement, flowInstance.getAttributes());
		
		return doc;
	}
	
	@RESTMethod(alias = "getinstanceattribute/{flowInstanceID}/{attribute}", method = "get")
	public Document getFlowInstanceAttribute(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowInstanceID") Integer flowInstanceID, @URIParam(name = "attribute") String attribute) throws Throwable {

		FlowInstance flowInstance = getFlowInstance(flowInstanceID, FlowInstance.ATTRIBUTES_RELATION, FlowInstance.FLOW_RELATION, Flow.FLOW_FAMILY_RELATION);

		if (flowInstance == null || flowInstance.getFirstSubmitted() == null || CollectionUtils.isEmpty(flowInstance.getAttributes())) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, flowInstance.getFlow().getFlowFamily(), user);
		
		for(FlowInstanceAttribute flowInstanceAttribute : flowInstance.getAttributes()) {
			
			if(flowInstanceAttribute.getName().equals(attribute)) {
				
				Document doc = XMLUtils.createDomDocument();
				
				Element attributeElement = doc.createElement("FlowInstanceAttribute");
				doc.appendChild(attributeElement);
				
				attributeElement.appendChild(flowInstanceAttribute.toXML(doc));
				
				return doc;
				
			}
			
		}
		
		throw new URINotFoundException(uriParser);
		
	}

	@RESTMethod(alias = "getinstance/{flowInstanceID}/xml", method = "get")
	public void getFlowInstanceXML(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowInstanceID") Integer flowInstanceID) throws Throwable {

		FlowInstance flowInstance = getFlowInstance(flowInstanceID);

		if (flowInstance == null) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, flowInstance.getFlow().getFlowFamily(), user);

		log.info("User " + user + " requesting XML for flow instance " + flowInstance);

		FlowInstanceEvent event = getLatestXMLEvent(flowInstance);

		if (event == null) {

			log.info("Unable to find any XML events for flow instance " + flowInstance + " requested by user " + user);

			throw new URINotFoundException(uriParser);
		}

		File xmlFile = xmlProvider.getXML(flowInstanceID, event.getEventID());

		if (xmlFile == null) {

			log.warn("Unable to find XML file for event " + event + " for flow instance " + flowInstance + " requested by user " + user);

			throw new URINotFoundException(uriParser);
		}

		if (hasMatachingFilters(flowInstance)) {

			Document doc = XMLUtils.parseXMLFile(xmlFile, false, false);

			for (XMLDocumentFilter filter : xmlDocumentFilters) {

				try {
					filter.processXML(flowInstance, doc);

				} catch (Exception e) {

					log.error("Error while running XML document filter " + filter, e);
				}
			}

			String processedXML = XMLUtils.toString(doc, xmlProvider.getEncoding(), false);

			try {
				HTTPUtils.sendReponse(processedXML, "text/xml", res);
				
			} catch (IOException e) {
				
				log.info("Error sending XML file for event " + event + " for flow instance " + flowInstance + " requested by user " + user);
			}

		} else {

			try {
				HTTPUtils.sendFile(xmlFile, "flowinstance-" + flowInstanceID + "-event-" + event.getEventID() + ".xml", req, res, ContentDisposition.INLINE);
				
			} catch (IOException e) {
				
				log.info("Error sending XML file for event " + event + " for flow instance " + flowInstance + " requested by user " + user);
			}
		}
	}

	private boolean hasMatachingFilters(FlowInstance flowInstance) throws SQLException {

		for (XMLDocumentFilter filter : xmlDocumentFilters) {

			if (filter.supportFlowInstance(flowInstance)) {

				return true;
			}
		}
		return false;
	}

	@RESTMethod(alias = "getinstance/{flowInstanceID}/pdf", method = "get")
	public void getFlowInstancePDF(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowInstanceID") Integer flowInstanceID) throws Throwable {

		FlowInstance flowInstance = getFlowInstance(flowInstanceID);

		if (flowInstance == null) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, flowInstance.getFlow().getFlowFamily(), user);

		log.info("User " + user + " requesting PDF for flow instance " + flowInstance);

		FlowInstanceEvent event = getLatestPDFEvent(flowInstance);

		if (event == null) {

			log.info("Unable to find any PDF events for flow instance " + flowInstance + " requested by user " + user);

			throw new URINotFoundException(uriParser);
		}

		File pdfFile = pdfProvider.getPDF(flowInstanceID, event.getEventID());

		if (pdfFile == null) {

			log.warn("Unable to find PDF file for event " + event + " for flow instance " + flowInstance + " requested by user " + user);

			throw new URINotFoundException(uriParser);
		}

		try {
			HTTPUtils.sendFile(pdfFile, "flowinstance-" + flowInstanceID + "-event-" + event.getEventID() + ".pdf", req, res, ContentDisposition.ATTACHMENT);
			
		} catch (IOException e) {
			
			log.info("Error sending PDF file for event " + event + " for flow instance " + flowInstance + " requested by user " + user);
		}
	}

	@RESTMethod(alias = "getstatus/{flowInstanceID}", method = "get")
	public Document getStatus(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowInstanceID") Integer flowInstanceID) throws Throwable {

		FlowInstance flowInstance = getFlowInstance(flowInstanceID, FlowInstance.STATUS_RELATION, FlowInstance.FLOW_RELATION, Flow.FLOW_FAMILY_RELATION);

		if (flowInstance == null) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, flowInstance.getFlow().getFlowFamily(), user);

		log.info("User " + user + " requesting status for flow instance " + flowInstance);

		Document doc = XMLUtils.createDomDocument();

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		
		generatorDocument.addIgnoredField(Status.INTERNAL_FIELDS);
		
		doc.appendChild(flowInstance.getStatus().toXML(generatorDocument));

		return doc;
	}

	@RESTMethod(alias = "getevents/family/{flowFamilyID}", method = "get")
	public Document getFlowInstanceEventsByFamily(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowFamilyID") Integer flowFamilyID) throws Throwable {

		FlowFamily family = getFlowFamily(flowFamilyID);

		if (family == null) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, family, user);

		Date fromDate = getDate(req.getParameter("fromDate"), true);
		Date toDate = getDate(req.getParameter("toDate"), false);

		Integer fromEventID = NumberUtils.toInt(req.getParameter("fromEventID"));
		Integer toEventID = NumberUtils.toInt(req.getParameter("toEventID"));
		
		log.info("User " + user + " requesting events for flow family " + family + " from date " + fromDate + " to date " + toDate + " from event ID " + fromEventID + " to event ID " + toEventID);

		List<FlowInstanceEvent> events = getFlowInstanceEvents(family, fromDate, toDate, fromEventID, toEventID);
		
		Document doc = XMLUtils.createDomDocument();

		Element eventsElement = doc.createElement("Events");
		doc.appendChild(eventsElement);

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(FlowInstance.SKIP_USER_ATTRIBUTES);
		generatorDocument.addFieldElementableListener(FlowInstance.class, MILLISECOND_LISTENER);
		generatorDocument.addIgnoredField(FlowInstanceEvent.SKIP_USER_ATTRIBUTES);
		generatorDocument.addRootElementableListener(FlowInstanceEvent.class, new FlowInstanceEventTimestampMillisecondElementableListener());

		XMLUtils.append(generatorDocument, eventsElement, events);

		return doc;
	}
	
	@RESTMethod(alias = "getevents/families", method = "get")
	public Document getFlowInstanceEventsByFamilies(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		String[] stringFlowFamilyIDs = req.getParameterValues("flowFamilyID");
		
		if(stringFlowFamilyIDs == null) {
			
			throw new URINotFoundException(uriParser);
		}
		
		List<Integer> flowFamilyIDs = NumberUtils.toInt(stringFlowFamilyIDs);
		
		if(flowFamilyIDs == null || flowFamilyIDs.size() != stringFlowFamilyIDs.length) {
			
			throw new URINotFoundException(uriParser);
		}

		List<FlowFamily> flowFamilies = new ArrayList<FlowFamily>(flowFamilyIDs.size());

		for (Integer flowFamilyID : flowFamilyIDs) {

			FlowFamily flowFamily = getFlowFamily(flowFamilyID);

			if (flowFamily == null) {

				throw new URINotFoundException(uriParser);
			}

			APIAccessUtils.accessCheck(apiAccessModule, flowFamily, user);

			flowFamilies.add(flowFamily);
		}

		Date fromDate = getDate(req.getParameter("fromDate"), true);
		Date toDate = getDate(req.getParameter("toDate"), false);

		Integer fromEventID = NumberUtils.toInt(req.getParameter("fromEventID"));
		Integer toEventID = NumberUtils.toInt(req.getParameter("toEventID"));
		
		log.info("User " + user + " requesting events for flow families " + flowFamilies + " from date " + fromDate + " to date " + toDate + " from event ID " + fromEventID + " to event ID " + toEventID);

		List<FlowInstanceEvent> events = getFlowInstanceEvents(flowFamilies, fromDate, toDate, fromEventID, toEventID);
		
		Document doc = XMLUtils.createDomDocument();

		Element eventsElement = doc.createElement("Events");
		doc.appendChild(eventsElement);

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(FlowInstance.SKIP_USER_ATTRIBUTES);
		generatorDocument.addFieldElementableListener(FlowInstance.class, MILLISECOND_LISTENER);
		generatorDocument.addIgnoredField(FlowInstanceEvent.SKIP_USER_ATTRIBUTES);
		generatorDocument.addRootElementableListener(FlowInstanceEvent.class, new FlowInstanceEventTimestampMillisecondElementableListener());

		XMLUtils.append(generatorDocument, eventsElement, events);

		return doc;
	}
	
	@RESTMethod(alias = "getevents/flow/{flowID}", method = "get")
	public Document getFlowInstanceEventsByFlow(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowID") Integer flowID) throws Throwable {

		Flow flow = getFlow(flowID);

		if (flow == null) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, flow.getFlowFamily(), user);

		Date fromDate = getDate(req.getParameter("fromDate"), true);
		Date toDate = getDate(req.getParameter("toDate"), false);

		Integer fromEventID = NumberUtils.toInt(req.getParameter("fromEventID"));
		Integer toEventID = NumberUtils.toInt(req.getParameter("toEventID"));
		
		log.info("User " + user + " requesting events for flow " + flow + " from date " + fromDate + " to date " + toDate + " from event ID " + fromEventID + " to event ID " + toEventID);

		List<FlowInstanceEvent> events = getFlowInstanceEvents(flow, fromDate, toDate, fromEventID, toEventID);
		
		Document doc = XMLUtils.createDomDocument();

		Element eventsElement = doc.createElement("Events");
		doc.appendChild(eventsElement);

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(FlowInstance.SKIP_USER_ATTRIBUTES);
		generatorDocument.addFieldElementableListener(FlowInstance.class, MILLISECOND_LISTENER);
		generatorDocument.addIgnoredField(FlowInstanceEvent.SKIP_USER_ATTRIBUTES);
		generatorDocument.addRootElementableListener(FlowInstanceEvent.class, new FlowInstanceEventTimestampMillisecondElementableListener());

		XMLUtils.append(generatorDocument, eventsElement, events);

		return doc;
	}
	
	@RESTMethod(alias = "getevents/{flowInstanceID}", method = "get")
	public Document getFlowInstanceEvents(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowInstanceID") Integer flowInstanceID) throws Throwable {

		FlowInstance flowInstance = getFlowInstance(flowInstanceID, FlowInstance.STATUS_RELATION, FlowInstanceEvent.ATTRIBUTES_RELATION, FlowInstance.FLOW_RELATION, Flow.FLOW_FAMILY_RELATION);

		if (flowInstance == null) {

			throw new URINotFoundException(uriParser);
		}
		
		APIAccessUtils.accessCheck(apiAccessModule, flowInstance.getFlow().getFlowFamily(), user);
		
		Date fromDate = getDate(req.getParameter("fromDate"), true);
		Date toDate = getDate(req.getParameter("toDate"), false);

		Integer fromEventID = NumberUtils.toInt(req.getParameter("fromEventID"));
		Integer toEventID = NumberUtils.toInt(req.getParameter("toEventID"));
		
		log.info("User " + user + " requesting events for flowInstanceID " + flowInstanceID + " from date " + fromDate + " to date " + toDate + " from event ID " + fromEventID + " to event ID " + toEventID);

		List<FlowInstanceEvent> events = getFlowInstanceEvents(flowInstanceID, fromDate, toDate, fromEventID, toEventID);
		
		Document doc = XMLUtils.createDomDocument();

		Element eventsElement = doc.createElement("Events");
		doc.appendChild(eventsElement);

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addIgnoredField(FlowInstanceEvent.SKIP_USER_ATTRIBUTES);
		generatorDocument.addRootElementableListener(FlowInstanceEvent.class, new FlowInstanceEventTimestampMillisecondElementableListener());

		XMLUtils.append(generatorDocument, eventsElement, events);

		return doc;
	}
	
	protected List<FlowInstanceEvent> getFlowInstanceEvents(FlowFamily flowFamily, Date fromDate, Date toDate, Integer fromEventID, Integer toEventID) throws SQLException {

		LowLevelQuery<FlowInstanceEvent> query = new LowLevelQuery<>();
		query.addChainedResultSetPopulator(CHAINED_FLOW_INSTANCE_EVENT_RESULT_SET_POPULATOR);
		query.addRelation(FlowInstanceEvent.ATTRIBUTES_RELATION);
			
		StringBuilder stringBuilder = new StringBuilder("SELECT e.* FROM flowengine_flow_instance_events e INNER JOIN flowengine_flow_instances fi ON e.flowInstanceID = fi.flowInstanceID INNER JOIN flowengine_flows f ON fi.flowID = f.flowID WHERE f.flowFamilyID = ? AND fi.firstSubmitted IS NOT NULL");
		
		query.addParameter(flowFamily.getFlowFamilyID());
		
		if (fromDate != null) {

			stringBuilder.append(" AND e.added >= ?");
			query.addParameter(new Timestamp(fromDate.getTime()));
		}

		if (toDate != null) {

			stringBuilder.append(" AND e.added <= ?");
			query.addParameter(new Timestamp(toDate.getTime()));
		}
		
		if (fromEventID != null) {

			stringBuilder.append(" AND e.eventID >= ?");
			query.addParameter(fromEventID);
		}

		if (toEventID != null) {

			stringBuilder.append(" AND e.eventID <= ?");
			query.addParameter(toEventID);
		}
				
		
		query.setSql(stringBuilder.toString());
		
		return flowInstanceEventDAO.getAll(query);
	}
	
	protected List<FlowInstanceEvent> getFlowInstanceEvents(List<FlowFamily> flowFamilies, Date fromDate, Date toDate, Integer fromEventID, Integer toEventID) throws SQLException {

		LowLevelQuery<FlowInstanceEvent> query = new LowLevelQuery<>();
		query.addRelation(FlowInstanceEvent.ATTRIBUTES_RELATION);
		query.addChainedResultSetPopulator(CHAINED_FLOW_INSTANCE_EVENT_RESULT_SET_POPULATOR, CHAINED_FLOW_INSTANCE_EVENT_FAMILY_RESULT_SET_POPULATOR);
			
		StringBuilder stringBuilder = new StringBuilder("SELECT e.*, f.flowFamilyID as flowFamilyID FROM flowengine_flow_instance_events e INNER JOIN flowengine_flow_instances fi ON e.flowInstanceID = fi.flowInstanceID INNER JOIN flowengine_flows f ON fi.flowID = f.flowID WHERE f.flowFamilyID IN (" + DBUtils.getPreparedStatementQuestionMarks(flowFamilies) + ") AND fi.firstSubmitted IS NOT NULL");
		
		for(FlowFamily flowFamily : flowFamilies) {
			
			query.addParameter(flowFamily.getFlowFamilyID());
		}
		
		if (fromDate != null) {

			stringBuilder.append(" AND e.added >= ?");
			query.addParameter(new Timestamp(fromDate.getTime()));
		}

		if (toDate != null) {

			stringBuilder.append(" AND e.added <= ?");
			query.addParameter(new Timestamp(toDate.getTime()));
		}
		
		if (fromEventID != null) {

			stringBuilder.append(" AND e.eventID >= ?");
			query.addParameter(fromEventID);
		}

		if (toEventID != null) {

			stringBuilder.append(" AND e.eventID <= ?");
			query.addParameter(toEventID);
		}
				
		
		query.setSql(stringBuilder.toString());
		
		return flowInstanceEventDAO.getAll(query);
	}
	
	protected List<FlowInstanceEvent> getFlowInstanceEvents(Flow flow, Date fromDate, Date toDate, Integer fromEventID, Integer toEventID) throws SQLException {

		LowLevelQuery<FlowInstanceEvent> query = new LowLevelQuery<>();
		query.addRelation(FlowInstanceEvent.ATTRIBUTES_RELATION);
		query.addChainedResultSetPopulator(CHAINED_FLOW_INSTANCE_EVENT_RESULT_SET_POPULATOR);
		
		query.addCachedRelation(FlowInstanceEvent.FLOW_INSTANCE_RELATION);
		query.addRelation(FlowInstanceEvent.FLOW_INSTANCE_RELATION);
		
		StringBuilder stringBuilder = new StringBuilder("SELECT e.* FROM flowengine_flow_instance_events e INNER JOIN flowengine_flow_instances fi ON e.flowInstanceID = fi.flowInstanceID WHERE fi.flowID = ? AND fi.firstSubmitted IS NOT NULL");
		query.addParameter(flow.getFlowID());
		
		if (fromDate != null) {

			stringBuilder.append(" AND e.added >= ?");
			query.addParameter(new Timestamp(fromDate.getTime()));
		}

		if (toDate != null) {

			stringBuilder.append(" AND e.added <= ?");
			query.addParameter(new Timestamp(toDate.getTime()));
		}
		
		if (fromEventID != null) {

			stringBuilder.append(" AND e.eventID >= ?");
			query.addParameter(fromEventID);
		}

		if (toEventID != null) {

			stringBuilder.append(" AND e.eventID <= ?");
			query.addParameter(toEventID);
		}
		
		query.setSql(stringBuilder.toString());
		
		return flowInstanceEventDAO.getAll(query);
	}
	
	protected List<FlowInstanceEvent> getFlowInstanceEvents(Integer flowInstanceID, Date fromDate, Date toDate, Integer fromEventID, Integer toEventID) throws SQLException {

		LowLevelQuery<FlowInstanceEvent> query = new LowLevelQuery<>();
		query.addRelation(FlowInstanceEvent.ATTRIBUTES_RELATION);
		query.addChainedResultSetPopulator(CHAINED_FLOW_INSTANCE_EVENT_RESULT_SET_POPULATOR);
		
		StringBuilder stringBuilder = new StringBuilder("SELECT * FROM flowengine_flow_instance_events e INNER JOIN flowengine_flow_instances fi ON e.flowInstanceID = fi.flowInstanceID WHERE e.flowInstanceID = ? AND fi.firstSubmitted IS NOT NULL");
		query.addParameter(flowInstanceID);
		
		if (fromDate != null) {

			stringBuilder.append(" AND added >= ?");
			query.addParameter(new Timestamp(fromDate.getTime()));
		}

		if (toDate != null) {

			stringBuilder.append(" AND added <= ?");
			query.addParameter(new Timestamp(toDate.getTime()));
		}
		
		if (fromEventID != null) {

			stringBuilder.append(" AND e.eventID >= ?");
			query.addParameter(fromEventID);
		}

		if (toEventID != null) {

			stringBuilder.append(" AND e.eventID <= ?");
			query.addParameter(toEventID);
		}
		
		query.setSql(stringBuilder.toString());
		
		return flowInstanceEventDAO.getAll(query);
	}

	protected FlowFamily getFlowFamily(Integer flowFamilyID) throws AccessDeniedException {

		return flowAdminModule.getFlowFamily(flowFamilyID);
	}

	protected Flow getFlow(Integer flowID) throws AccessDeniedException {

		return (Flow) flowAdminModule.getFlow(flowID);
	}

	protected List<FlowInstance> getFlowInstances(List<Flow> flows, Date fromDate, Date toDate) throws SQLException {

		HighLevelQuery<FlowInstance> query = new HighLevelQuery<FlowInstance>();

		query.addParameter(flowQueryParamFactory.getWhereInParameter(flows));

		addDateParameters(query, fromDate, toDate);

		return flowInstanceDAO.getAll(query);
	}
	
	protected List<FlowInstance> getFlowInstances(List<Flow> flows, List<Status> statuses, Date fromDate, Date toDate) throws SQLException {
		
		HighLevelQuery<FlowInstance> query = new HighLevelQuery<FlowInstance>();
		
		query.addParameter(flowQueryParamFactory.getWhereInParameter(flows));
		
		addDateParameters(query, null, null);
		
		addStatusParameters(query, statuses, fromDate, toDate);
		
		return flowInstanceDAO.getAll(query);
	}

	protected List<FlowInstance> getFlowInstances(Flow flow, Date fromDate, Date toDate) throws SQLException {

		HighLevelQuery<FlowInstance> query = new HighLevelQuery<FlowInstance>();

		query.addParameter(flowQueryParamFactory.getParameter(flow));

		addDateParameters(query, fromDate, toDate);

		return flowInstanceDAO.getAll(query);
	}
	
	protected List<FlowInstance> getFlowInstances(Flow flow, List<Status> statuses, Date fromDate, Date toDate) throws SQLException {

		HighLevelQuery<FlowInstance> query = new HighLevelQuery<FlowInstance>();

		query.addParameter(flowQueryParamFactory.getParameter(flow));

		addDateParameters(query, null, null);
		
		addStatusParameters(query, statuses, fromDate, toDate);

		return flowInstanceDAO.getAll(query);
	}

	protected List<FlowInstanceAttribute> getFlowInstanceAttributes(String attribute, String value, Field... relations) throws SQLException {

		HighLevelQuery<FlowInstanceAttribute> query = new HighLevelQuery<FlowInstanceAttribute>();

		if (relations != null) {
			query.addRelations(relations);
		}

		query.addParameter(attributeNameQueryParamFactory.getParameter(attribute));
		query.addParameter(attributeValueQueryParamFactory.getParameter(value));

		return flowInstanceAttributeDAO.getAll(query);
	}

	protected void addDateParameters(HighLevelQuery<FlowInstance> query, Date fromDate, Date toDate) {

		if (fromDate == null && toDate == null) {

			query.addParameter(firstSubmittedParamFactory.getIsNotNullParameter());

		} else {

			if (fromDate != null) {

				query.addParameter(firstSubmittedParamFactory.getParameter(new Timestamp(fromDate.getTime()), QueryOperators.BIGGER_THAN_OR_EQUALS));
			}

			if (toDate != null) {

				query.addParameter(firstSubmittedParamFactory.getParameter(new Timestamp(toDate.getTime()), QueryOperators.SMALLER_THAN_OR_EQUALS));
			}
		}
	}
	
	protected void addStatusParameters(HighLevelQuery<FlowInstance> query, List<Status> statuses, Date fromDate, Date toDate) {

		if (fromDate == null && toDate == null) {

			query.addParameter(lastStatusChangeParamFactory.getIsNotNullParameter());

		} else {

			if (fromDate != null) {

				query.addParameter(lastStatusChangeParamFactory.getParameter(new Timestamp(fromDate.getTime()), QueryOperators.BIGGER_THAN_OR_EQUALS));
			}

			if (toDate != null) {

				query.addParameter(lastStatusChangeParamFactory.getParameter(new Timestamp(toDate.getTime()), QueryOperators.SMALLER_THAN_OR_EQUALS));
			}
		}
		
		query.addParameter(flowInstanceStatusParamFactory.getWhereInParameter(statuses));
	}

	protected FlowInstance getFlowInstance(Integer flowInstanceID, Field... relations) throws SQLException, AccessDeniedException {

		HighLevelQuery<FlowInstance> query = new HighLevelQuery<FlowInstance>();

		if (relations.length > 0) {

			query.addRelations(relations);

		} else {

			query.addRelations(FlowInstance.EVENTS_RELATION, FlowInstanceEvent.ATTRIBUTES_RELATION, FlowInstance.FLOW_RELATION, Flow.FLOW_FAMILY_RELATION, FlowInstance.ATTRIBUTES_RELATION);
		}

		query.addParameter(flowInstanceIDParamFactory.getParameter(flowInstanceID));
		query.addParameter(firstSubmittedParamFactory.getIsNotNullParameter());

		return flowInstanceDAO.get(query);
	}

	private FlowInstanceEvent getLatestXMLEvent(FlowInstance flowInstance) {

		if (CollectionUtils.isEmpty(flowInstance.getEvents())) {

			return null;
		}

		for (int i = flowInstance.getEvents().size() - 1; i > -1; i--) {

			FlowInstanceEvent event = flowInstance.getEvents().get(i);

			if (Boolean.TRUE.equals(event.getAttributeHandler().getBoolean("xml"))) {

				return event;
			}
		}

		return null;
	}

	private FlowInstanceEvent getLatestPDFEvent(FlowInstance flowInstance) {

		if (CollectionUtils.isEmpty(flowInstance.getEvents())) {

			return null;
		}

		for (int i = flowInstance.getEvents().size() - 1; i > -1; i--) {

			FlowInstanceEvent event = flowInstance.getEvents().get(i);

			if (Boolean.TRUE.equals(event.getAttributeHandler().getBoolean("pdf"))) {

				return event;
			}
		}

		return null;
	}

	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		throw new URINotFoundException(uriParser);
	}

	public void addXMLDocumentFilter(XMLDocumentFilter filter) {

		if (filter == null) {
			throw new NullPointerException();
		}

		if (!xmlDocumentFilters.contains(filter)) {
			
			xmlDocumentFilters.add(filter);
		}
	}

	public void removeXMLDocumentFilter(XMLDocumentFilter filter) {

		xmlDocumentFilters.remove(filter);
	}
}
