package com.nordicpeak.flowengine.queries.treequery;

import java.io.File;
import java.util.HashMap;
import java.util.List;

import javax.sql.DataSource;

import org.apache.log4j.Level;

import se.unlogic.cron4jutils.CronStringValidator;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.beans.Group;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.io.FileUtils;
import se.unlogic.standardutils.serialization.SerializationUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.TimeUtils;

import com.nordicpeak.flowengine.queries.multitreequery.MultiTreeQueryProviderModule;
import com.nordicpeak.treequerydataprovider.SimpleTreeNode;
import com.nordicpeak.treequerydataprovider.TreeDataHandler;
import com.nordicpeak.treequerydataprovider.TreeDataProvider;
import com.nordicpeak.treequerydataprovider.TreeNode;

import it.sauronsoftware.cron4j.Scheduler;
import it.sauronsoftware.cron4j.Task;
import it.sauronsoftware.cron4j.TaskExecutionContext;

public class GroupTreeQueryDataProvider extends AnnotatedForegroundModule implements TreeDataProvider {

	private static final String GROUP_LDAP_UID_ATTRIBUTE = "ldap_group_uid";
	private static final String GROUP_PARENT_UID_ATTRIBUTE = "parentGroupUID";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Provider ID", description = "The unique identifier for this provider.", required = true)
	private String providerID = getClass().getName();

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Cache interval", description = "How often organizations should be cached (specified in crontab format)", required = true, formatValidator = CronStringValidator.class)
	private String cacheInterval = "0 0 * * *";

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "Cache filestore", description = "Optional filestore to improve restart times")
	private String cacheFilestore;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Group id attribute name", description = "The name of the group id attribute", required = true)
	private String groupIDAttribute = GROUP_LDAP_UID_ATTRIBUTE;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Parent group id attribute name", description = "The name of the parent group id attribute", required = true)
	private String parentGroupIDAttribute = GROUP_PARENT_UID_ATTRIBUTE;

	private File treeCacheFile;

	private Scheduler scheduler;

	protected TreeDataHandler treeDataHandler;
	protected MultiTreeQueryProviderModule multiTreeQueryProviderModule;

	protected SimpleTreeNode organizationTree;

	boolean firstStart = true;

	@Override
	public void update(ForegroundModuleDescriptor descriptor, DataSource dataSource) throws Exception {

		if (treeDataHandler != null) {
			treeDataHandler.removeTreeProvider(this);
		}
		
		if (multiTreeQueryProviderModule != null) {
			multiTreeQueryProviderModule.removeTreeProvider(this);
		}
		
		super.update(descriptor, dataSource);

		if (treeDataHandler != null) {
			treeDataHandler.addTreeProvider(this);
		}
		
		if (multiTreeQueryProviderModule != null) {
			multiTreeQueryProviderModule.addTreeProvider(this);
		}
	}

	@Override
	protected void moduleConfigured() throws Exception {

		stopScheduler();

		super.moduleConfigured();

		if (StringUtils.isEmpty(cacheFilestore)) {

			log.warn("Cache filestore not set");

		} else if (!FileUtils.isReadable(cacheFilestore)) {

			log.error("Cache filestore not readable");

		} else {

			treeCacheFile = new File(cacheFilestore + File.separator + this.getClass().getSimpleName() + "-tree.cache");
		}

		if (ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.ERROR)) {

			initScheduler();
		}
	}

	protected void initScheduler() {

		scheduler = new Scheduler(systemInterface.getApplicationName() + " - " + moduleDescriptor.toString());
		scheduler.setDaemon(true);

		Task cacheTask = new Task() {

			@Override
			public void execute(TaskExecutionContext context) throws RuntimeException {

				cacheTree(context);
			}

		};

		scheduler.schedule(cacheInterval, cacheTask);
		scheduler.start();

		if (firstStart) {

			if (treeCacheFile != null && treeCacheFile.exists()) {

				try {
					long startTime = System.currentTimeMillis();

					organizationTree = SerializationUtils.deserializeFromArray(SimpleTreeNode.class, FileUtils.getRawBytes(treeCacheFile));

					log.info("Loaded tree data from disk cache in " + TimeUtils.millisecondsToString(System.currentTimeMillis() - startTime) + " ms");

				} catch (Exception e) {

					organizationTree = null;

					log.error("Error loading tree data from disk cache", e);
				}
			}

			firstStart = false;

			if (organizationTree == null) {

				scheduler.launch(cacheTask);
			}

		} else {

			scheduler.launch(cacheTask);
		}
	}

	protected void stopScheduler() {

		try {
			if (scheduler != null) {

				scheduler.stop();
				scheduler = null;
			}

		} catch (IllegalStateException e) {
			log.error("Error stopping scheduler", e);
		}
	}

	@Override
	public void unload() throws Exception {

		stopScheduler();

		if (treeDataHandler != null) {
			treeDataHandler.removeTreeProvider(this);
		}
		
		if (multiTreeQueryProviderModule != null) {
			multiTreeQueryProviderModule.removeTreeProvider(this);
		}

		if (treeCacheFile != null && organizationTree != null) {

			log.info("Saving tree data to disk cache");

			FileUtils.writeFile(treeCacheFile, SerializationUtils.serializeToArray(organizationTree));
		}

		organizationTree = null;

		super.unload();
	}

	private synchronized void cacheTree(TaskExecutionContext context) {

		log.info("Start caching group tree");
		
		long startTime = System.currentTimeMillis();

		try {

			List<Group> groups = systemInterface.getGroupHandler().getGroupsByAttribute(groupIDAttribute, true);

			if (groups != null) {

				boolean hasMultipleRootNodes = false;

				HashMap<String, SimpleTreeNode> treeMap = new HashMap<String, SimpleTreeNode>();
				SimpleTreeNode rootNode = null;

				for (Group group : groups) {

					if(!group.isEnabled()) {
						
						continue;
					}
					
					String groupID = group.getAttributeHandler().getString(groupIDAttribute);
					
					if (treeMap.get(groupID) != null) {

						log.warn("Duplicated group with id " + groupID + " found, skipping..");
						continue;
					}

					SimpleTreeNode treeNode = new SimpleTreeNode(group.getName(), groupID);
					
					treeMap.put(groupID, treeNode);

					String parentGroupID = group.getAttributeHandler().getString(parentGroupIDAttribute);
					
					if (parentGroupID == null) {

						if(rootNode == null) {
							
							rootNode = treeMap.get(groupID);
							
						} else {
							
							log.warn("Has multiple root nodes");
							
							hasMultipleRootNodes = true;
							
						}
						
					}
					
				}
				
				if (hasMultipleRootNodes || rootNode == null) {

					rootNode = new SimpleTreeNode(getTreeName(), getTreeID());
				}
				
				for (Group group : groups) {
					
					if (!group.isEnabled()) {

						continue;
					}
					
					String groupID = group.getAttributeHandler().getString(groupIDAttribute);
					String parentGroupID = group.getAttributeHandler().getString(parentGroupIDAttribute);

					if (!StringUtils.isEmpty(parentGroupID)) {

						SimpleTreeNode parentNode = treeMap.get(parentGroupID);

						if (parentNode == null) {

							log.warn("Parent node " + parentGroupID + " not found specified on " + group + " (" + groupID + ")");
							continue;
						}

						if (parentNode.getKey().equals(groupID)) {

							log.warn("Group " + group + " (" + groupID + ") has itself as parent and will be skipped");
							continue;
						}

						parentNode.put(treeMap.get(groupID));
						
					} else if(!rootNode.getKey().equals(groupID)) {
						
						rootNode.put(treeMap.get(groupID));
						
					}

				}

				organizationTree = rootNode;

				log.info("Cached " + groups.size() + " organizations in " + TimeUtils.millisecondsToString(System.currentTimeMillis() - startTime) + " ms");

				return;

			}

		} catch (Exception e) {

			log.error("Unable to cache groups", e);
		}
	}

	@Override
	public TreeNode getTree() {

		return organizationTree;
	}

	@InstanceManagerDependency
	public void setTreeDataHandler(TreeDataHandler treeDataHandler) {

		if (treeDataHandler != null) {

			treeDataHandler.addTreeProvider(this);

		} else if (this.treeDataHandler != null) {

			this.treeDataHandler.removeTreeProvider(this);
		}

		this.treeDataHandler = treeDataHandler;
	}
	
	@InstanceManagerDependency
	public void setMultiTreeQueryProviderModule(MultiTreeQueryProviderModule multiTreeQueryProviderModule) {
		
		if (multiTreeQueryProviderModule != null) {
			
			multiTreeQueryProviderModule.addTreeProvider(this);
			
		} else if (this.multiTreeQueryProviderModule != null) {
			
			this.multiTreeQueryProviderModule.removeTreeProvider(this);
		}
		
		this.multiTreeQueryProviderModule = multiTreeQueryProviderModule;
	}

	@Override
	public String getTreeName() {

		return moduleDescriptor.getDescription();
	}

	@Override
	public String getTreeID() {

		return providerID;
	}

}
