package com.nordicpeak.flowengine.integration.http;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Level;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.LinkTag;
import se.unlogic.hierarchy.core.beans.ScriptTag;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.AccessUtils;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.core.utils.ModuleViewFragmentTransformer;
import se.unlogic.hierarchy.core.utils.ViewFragmentModule;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.hierarchy.foregroundmodules.staticcontent.StaticContentModule;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.io.CloseUtils;
import se.unlogic.standardutils.populators.StringPopulator;
import se.unlogic.standardutils.time.MillisecondTimeUnits;
import se.unlogic.standardutils.validation.NonNegativeStringIntegerValidator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowAdminExtensionShowView;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.enums.EventType;
import com.nordicpeak.flowengine.events.StatusChangedByManagerEvent;
import com.nordicpeak.flowengine.events.SubmitEvent;
import com.nordicpeak.flowengine.interfaces.FlowAdminExtensionViewProvider;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstanceEvent;
import com.nordicpeak.flowengine.interfaces.XMLProvider;
import com.nordicpeak.flowengine.managers.FlowInstanceManager;
import com.nordicpeak.flowengine.managers.ImmutableFlowInstanceManager;
import com.nordicpeak.flowengine.utils.FlowInstanceUtils;


public class HTTPIntegrationProvider extends AnnotatedForegroundModule implements FlowAdminExtensionViewProvider, ViewFragmentModule<ForegroundModuleDescriptor>{

	private static final AnnotatedRequestPopulator<FlowFamilyIntegrationSettings> POPULATOR = new AnnotatedRequestPopulator<FlowFamilyIntegrationSettings>(FlowFamilyIntegrationSettings.class);
	
	@XSLVariable(name = "i18n.Name")
	private String paramNameDisplayName;

	@XSLVariable(name = "i18n.Value")
	private String paramValueDisplayName;
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name="Provider priority", description="The priority of this provider", required=true)
	private Integer priority = 100;
	
	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Enable fragment XML debug", description = "Enables debugging of fragment XML")
	private boolean debugFragmententXML;
	
	@ModuleSetting(allowsNull=true)
	@TextAreaSettingDescriptor(name="Supported actionID's", description="The action ID's which will trigger export XML to be generated and stored when a submit event is detected", required=true)
	private List<String> supportedActionIDs;
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "Connection timeout in seconds", formatValidator = NonNegativeStringIntegerValidator.class, required = true)
	private Integer connectionTimeout = 5;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Request timeout", description = "Request timeout in seconds", formatValidator = NonNegativeStringIntegerValidator.class, required = true)
	private Integer requestTimeout = 10;
	
	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Force UTF-8 encoding", description = "Forces UTF-8 encoding to be used instead of the system default encoding when adding integration parameters.")
	protected boolean forceUTF;
	
	@InstanceManagerDependency(required=true)
	private StaticContentModule staticContentModule;
	
	@InstanceManagerDependency
	protected XMLProvider xmlProvider;
	
	private FlowAdminModule flowAdminModule;
	
	private AnnotatedDAO<FlowFamilyIntegrationSettings> settingsDAO;
	private QueryParameterFactory<FlowFamilyIntegrationSettings, Integer> flowFamilyIDParamFactory;
	
	private ModuleViewFragmentTransformer<ForegroundModuleDescriptor> viewFragmentTransformer;
	
	protected boolean configured;
	
	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, HTTPIntegrationProvider.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}
		
		settingsDAO = new AnnotatedDAO<FlowFamilyIntegrationSettings>(dataSource, FlowFamilyIntegrationSettings.class, new SimpleAnnotatedDAOFactory(dataSource));
		
		flowFamilyIDParamFactory = settingsDAO.getParamFactory("flowFamilyID", Integer.class);
	}
	
	@Override
	protected void moduleConfigured() throws Exception {

		this.configured = ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.WARN);
		
		this.viewFragmentTransformer = new ModuleViewFragmentTransformer<ForegroundModuleDescriptor>(sectionInterface.getForegroundModuleXSLTCache(), this, systemInterface.getEncoding());

		this.viewFragmentTransformer.setDebugXML(debugFragmententXML);
	}
	
	@Override
	public void unload() throws Exception {

		if(flowAdminModule != null){
			
			setFlowAdminModule(null);
		}
		
		super.unload();
	}
	
	@Override
	public FlowAdminExtensionShowView getShowView(Flow flow, HttpServletRequest req, User user, URIParser uriParser) throws TransformerConfigurationException, TransformerException, SQLException {

		if (!AccessUtils.checkAccess(user, this.moduleDescriptor)) {

			return null;
		}
		
		Document doc = createDocument(req, uriParser, user);

		Element showViewElement = doc.createElement("ShowIntegrationSettings");
		doc.getDocumentElement().appendChild(showViewElement);

		FlowFamilyIntegrationSettings integrationSettings = getIntegrationSettings(flow);
		
		boolean enabled = false;
		
		if(integrationSettings != null) {
		
			XMLUtils.append(doc, showViewElement, integrationSettings);
			enabled = integrationSettings.isEnabled();
		}

		showViewElement.appendChild(flow.toXML(doc));

		return new FlowAdminExtensionShowView(viewFragmentTransformer.createViewFragment(doc, true), enabled);
	}

	private FlowFamilyIntegrationSettings getIntegrationSettings(Flow flow) throws SQLException {

		HighLevelQuery<FlowFamilyIntegrationSettings> query = new HighLevelQuery<FlowFamilyIntegrationSettings>();

		query.addParameter(flowFamilyIDParamFactory.getParameter(flow.getFlowFamily().getFlowFamilyID()));

		return settingsDAO.get(query);
	}

	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {

		Document doc = XMLUtils.createDomDocument();
		Element document = doc.createElement("Document");
		
		XMLUtils.appendNewElement(doc, document, "ModuleURI", req.getContextPath() +  getFullAlias());
		XMLUtils.appendNewElement(doc, document, "StaticContentURL", staticContentModule.getModuleContentURL(moduleDescriptor));

		doc.appendChild(document);

		return doc;
	}

	@EventListener(channel = FlowFamily.class)
	public void processEvent(CRUDEvent<FlowFamily> event, EventSource eventSource) {

		if (event.getAction() == CRUDAction.DELETE) {

			for (FlowFamily flowFamily : event.getBeans()) {

				log.info("Deleting http integration settings for flowFamilyID: " + flowFamily.getFlowFamilyID());

				try {
					
					HighLevelQuery<FlowFamilyIntegrationSettings> query = new HighLevelQuery<FlowFamilyIntegrationSettings>();
					
					query.addParameter(flowFamilyIDParamFactory.getParameter(flowFamily.getFlowFamilyID()));

					settingsDAO.delete(query);
					
				} catch (SQLException e) {

					log.error("Error deleting http integration settings for flowFamilyID: " + flowFamily.getFlowFamilyID(), e);
				}
			}
		}
	}
	
	@EventListener(channel=FlowInstanceManager.class, priority = 200)
	public void processEvent(SubmitEvent event, EventSource eventSource) throws SQLException{
		
		if(!configured){
			
			log.warn("Module " + moduleDescriptor + " is not properly configured, ignoring event " + event);
			
			return;
		}
		
		if(event.getEvent().getEventType() != EventType.SUBMITTED || event.getActionID() == null || !supportedActionIDs.contains(event.getActionID())){

			return;
		}
		
		processEvent(event.getFlowInstanceManager().getFlowInstance(), event.getEvent(), false);
		
	}
	
	@EventListener(channel = FlowInstance.class)
	public void processEvent(StatusChangedByManagerEvent event, EventSource eventSource) throws SQLException {

		if(!configured){
			
			log.warn("Module " + moduleDescriptor + " is not properly configured, ignoring event " + event);
			
			return;
		}
		
		processEvent(event.getFlowInstance(), event.getEvent(), true);
	}
	
	public void processEvent(ImmutableFlowInstance flowInstance, FlowInstanceEvent event, boolean generateXML) throws SQLException {
		
		if(xmlProvider == null){
			
			log.warn("No XMLProvider set, unable to send XML for flow instance " + flowInstance);
			
			return;
		}
		
		FlowFamilyIntegrationSettings integrationSettings = getIntegrationSettings((Flow) flowInstance.getFlow());
		
		if(integrationSettings == null || !integrationSettings.isEnabled()){
			
			return;
		}
		
		if(event.getEventType() == EventType.STATUS_UPDATED && !integrationSettings.isTriggerOnStatusChanges()){
			
			return;
		}
		
		File xmlFile = xmlProvider.getXML(flowInstance.getFlowInstanceID(), event.getEventID());
		
		File tempFile = null;
		
		try{
			if(xmlFile == null){
				
				if(!generateXML){
					
					log.warn("No XML file found for event " + event + " belonging to flow instance " + flowInstance);
					
					return;
				}
				
				try {
					ImmutableFlowInstanceManager instanceManager = flowAdminModule.getImmutableFlowInstanceManager(flowInstance.getFlowInstanceID());
					
					tempFile = createTempFile(flowInstance, event);
					
					ImmutableFlowInstanceEvent lastSubmitted = FlowInstanceUtils.getLatestSubmitEvent(instanceManager.getFlowInstance());
					
					xmlProvider.generateXML(instanceManager.getFlowInstance(), instanceManager, event, lastSubmitted.getAdded(), tempFile);
				
				} catch (Exception e) {
					
					log.error("Error generating XML file for event " + event + " belonging to flow instance " + flowInstance, e);
					
					return;
				}
			}

			if (integrationSettings.getParameters() != null) {

				try {
					Document doc;

					if(tempFile != null){

						doc = parseXML(tempFile);
						
					}else{
						
						doc = parseXML(xmlFile);
					}
					
					Element parametersElement = doc.createElement("Parameters");
					doc.getDocumentElement().appendChild(parametersElement);

					for (IntegrationParameter parameter : integrationSettings.getParameters()) {

						Element parameterElement = doc.createElement("Parameter");

						parameterElement.setAttribute("Name", parameter.getName());
						parameterElement.setAttribute("Value", parameter.getValue());

						parametersElement.appendChild(parameterElement);
					}

					if(tempFile == null){
						
						tempFile = createTempFile(flowInstance, event);
					}
					
					if(forceUTF){
						
						XMLUtils.writeXMLFile(doc, tempFile, true, "UTF-8", "1.1");
						
					}else{
						
						XMLUtils.writeXMLFile(doc, tempFile, true, systemInterface.getEncoding());
					}
					
				} catch (Exception e) {

					log.error("Error adding integration parameters to XML file for event " + event + " belonging to flow instance " + flowInstance, e);

					return;
				}

			}
			
			FileInputStream inputStream;
			
			try {
				if(tempFile != null){

					inputStream = new FileInputStream(tempFile);
					
				}else{
					
					inputStream = new FileInputStream(xmlFile);
				}
				
			} catch (FileNotFoundException e) {

				log.error("Error reading XML file for event " + event + " from flow instance " + flowInstance, e);
				
				return;
			}
			
			try {
				
				log.info("Sending XML of event " + event + " from flow instance " + flowInstance + " to " + integrationSettings.getUrl());
				
				SimpleRequest request = new SimpleRequest(integrationSettings.getUrl());
				request.setUsername(integrationSettings.getUsername());
				request.setPassword(integrationSettings.getPassword());
				request.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
				request.setReadTimeout(requestTimeout * MillisecondTimeUnits.SECOND);
				request.addHeader("Content-Type", "application/xml");
				
				String response = HTTPUtils.sendHTTPPostRequest(request, inputStream, Charset.forName(systemInterface.getEncoding()));
				
				log.info("XML successfully sent with response: " + response);
				
			} catch (Exception e) {

				log.warn("Error sending XML file for event " + event + " belonging to flow instance " + event.getFlowInstance() + " to URL " + integrationSettings.getUrl(), e);
			
			}finally{
			
				CloseUtils.close(inputStream);
			}
			
		}finally {
			
			if(tempFile != null){
				tempFile.delete();
			}
		}
	}
	
	private File createTempFile(ImmutableFlowInstance flowInstance, FlowInstanceEvent event) throws IOException {

		return File.createTempFile("integration-xml" + flowInstance.getFlowInstanceID() + "-" + event.getEventID(), ".xml");
	}

	private Document parseXML(File xmlFile) throws SAXException, IOException, ParserConfigurationException {

		if(forceUTF){
			
			InputStream inputStream = null;
			
			try {
				inputStream = new FileInputStream(xmlFile);
				
				return XMLUtils.parseXML(inputStream, false, false, "UTF-8");
				
			} finally{
				
				CloseUtils.close(inputStream);
			}

		}else{
		
			return XMLUtils.parseXMLFile(xmlFile, false, false);
		}
	}
	
	@WebPublic(alias="update")
	public ForegroundModuleResponse updateFlowIntegrationSettings(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
	
		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);

		if (flow == null) {

			return flowAdminModule.list(req, res, user, uriParser, new ValidationError("FlowNotFound"));

		} else if (!flowAdminModule.hasFlowAccess(user, flow)) {

			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}

		FlowFamilyIntegrationSettings integrationSettings = getIntegrationSettings(flow);

		ValidationException validationException = null;

		if (req.getMethod().equalsIgnoreCase("POST")) {

			if (integrationSettings == null) {

				integrationSettings = new FlowFamilyIntegrationSettings();
				integrationSettings.setFlowFamilyID(flow.getFlowFamily().getFlowFamilyID());
			}

			try {
				POPULATOR.populate(integrationSettings, req);

				String[] parameterIDs = req.getParameterValues("parameterID");

				if (parameterIDs != null) {

					List<IntegrationParameter> parameters = new ArrayList<IntegrationParameter>(parameterIDs.length);

					List<ValidationError> validationErrors = new ArrayList<ValidationError>();

					for (String parameterID : parameterIDs) {

						String name = ValidationUtils.validateParameter("name-" + parameterID, paramNameDisplayName, req, true, 1, 255, StringPopulator.getPopulator(), validationErrors);
						String value = ValidationUtils.validateParameter("value-" + parameterID, paramValueDisplayName, req, true, 1, 255, StringPopulator.getPopulator(), validationErrors);

						if (validationErrors.isEmpty()) {

							IntegrationParameter parameter = new IntegrationParameter();
							parameter.setName(name);
							parameter.setValue(value);
							parameters.add(parameter);
						}
					}

					if (!validationErrors.isEmpty()) {

						throw new ValidationException(validationErrors);
					}

					integrationSettings.setParameters(parameters);

				} else {

					integrationSettings.setParameters(null);
				}

				log.info("User " + user + " updating integration settings for flow " + flow);

				if(integrationSettings.isPopulated()){
					
					settingsDAO.addOrUpdate(integrationSettings, null);
					
				}else{
					
					settingsDAO.delete(integrationSettings);
				}
				

				flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + "#httpintegrationsettings");

				return null;

			} catch (ValidationException e) {

				validationException = e;
			}
		}
		
		log.info("User " + user + " requesting update integration settings form for flow " + flow);

		Document doc = createDocument(req, uriParser, user);

		Element integrationSettingsElement = doc.createElement("UpdateIntegrationSettings");
		doc.getDocumentElement().appendChild(integrationSettingsElement);

		integrationSettingsElement.appendChild(flow.toXML(doc));

		XMLUtils.append(doc, integrationSettingsElement, integrationSettings);

		if (validationException != null) {

			integrationSettingsElement.appendChild(validationException.toXML(doc));
			integrationSettingsElement.appendChild(RequestUtils.getRequestParameters(req, doc));
		}

		return new SimpleForegroundModuleResponse(doc, this.getDefaultBreadcrumb());
	}
	
	@WebPublic(alias="delete")
	public ForegroundModuleResponse deleteFlowIntegrationSettings(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		if(!HTTPUtils.isPost(req)) {
			
			throw new AccessDeniedException("Delete requests using method " + req.getMethod() + " are not allowed.");
		}		
		
		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);

		if (flow == null) {

			return flowAdminModule.list(req, res, user, uriParser, new ValidationError("FlowNotFound"));

		} else if (!flowAdminModule.hasFlowAccess(user, flow)) {

			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}

		FlowFamilyIntegrationSettings integrationSettings = getIntegrationSettings(flow);
		
		if(integrationSettings != null){
			
			log.info("User " + user + " deleting integration settings for flow " + flow);
			
			settingsDAO.delete(integrationSettings);
			
			//TODO add event to flow family history
			
		}else{
			
			log.warn("User " + user + " trying to deleting integration settings for flow " + flow + " which has no integration settings");
		}
		
		flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + "#httpintegrationsettings");
		
		return null;
	}
	
	@Override
	public String getExtensionViewTitle() {

		return moduleDescriptor.getName();
	}
	
	@Override
	public int getPriority() {

		return priority;
	}

	@Override
	public ForegroundModuleDescriptor getModuleDescriptor() {

		return moduleDescriptor;
	}

	@Override
	public List<LinkTag> getLinkTags() {

		return links;
	}

	@Override
	public List<ScriptTag> getScriptTags() {

		return scripts;
	}
	
	@InstanceManagerDependency(required = true)
	public void setFlowAdminModule(FlowAdminModule flowAdminModule) {

		if(flowAdminModule == null && this.flowAdminModule != null){
			
			this.flowAdminModule.removeExtensionViewProvider(this);
		}
		
		this.flowAdminModule = flowAdminModule;

		if (this.flowAdminModule != null) {

			this.flowAdminModule.addExtensionViewProvider(this);
		}
	}
	
	@Override
	public int getModuleID() {

		return moduleDescriptor.getModuleID();
	}
}
