package com.nordicpeak.flowengine.queries.internaluserquery;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.handlers.UserHandler;
import se.unlogic.hierarchy.core.utils.UserUtils;
import se.unlogic.hierarchy.foregroundmodules.usersessionadmin.UserNameComparator;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.json.JsonUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.interfaces.QueryRequestProcessor;
import com.nordicpeak.flowengine.queries.internaluserqueries.InternalUserQueriesConstants;
import com.nordicpeak.flowengine.queries.internaluserquery.beans.InternalUserQuery;
import com.nordicpeak.flowengine.queries.internaluserquery.beans.InternalUserQueryInstance;
import com.nordicpeak.flowengine.queries.internaluserquery.beans.SelectedManagerAttribute;
import com.nordicpeak.flowengine.queries.internaluserquery.beans.SelectedUserAttribute;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;

public class InternalUserQueryRequestProcessor implements QueryRequestProcessor {

	protected final Logger log = Logger.getLogger(this.getClass());

	private final String encoding;
	private final InternalUserQueryInstance queryInstance;
	private final UserHandler userHandler;
	private final InternalUserQueryProviderModule internalUserQueryProviderModule;
	private final User poster;

	public InternalUserQueryRequestProcessor(InternalUserQueryInstance queryInstance, User poster, UserHandler userHandler, String encoding, InternalUserQueryProviderModule internalUserQueryProviderModule) {

		super();

		this.queryInstance = queryInstance;
		this.poster = poster;
		this.userHandler = userHandler;
		this.encoding = encoding;
		this.internalUserQueryProviderModule = internalUserQueryProviderModule;
	}

	protected static void sendEmptyJSONResponse(HttpServletResponse res) throws IOException {

		JsonObject jsonObject = new JsonObject(1);
		jsonObject.putField("hitCount", "0");
		HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);
	}

	protected void sendJSONResponse(JsonArray jsonArray, HttpServletResponse res) throws IOException {

		JsonObject jsonObject = new JsonObject(2);
		jsonObject.putField("hitCount", Integer.toString(jsonArray.size()));
		jsonObject.putField("hits", jsonArray);
		HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);
	}

	@Override
	public void processRequest(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		String searchString = req.getParameter("q");

		if (searchString != null && !encoding.equalsIgnoreCase("UTF-8")) {
			try {
				searchString = URLDecoder.decode(searchString, "UTF-8");

			} catch (UnsupportedEncodingException | IllegalArgumentException e) {

				log.warn("Unable to parse request from user " + user, e);

				sendEmptyJSONResponse(res);
				return;
			}
		}

		if (StringUtils.isEmpty(searchString)) {

			sendEmptyJSONResponse(res);
			return;
		}

		InternalUserQuery query = queryInstance.getQuery();
		List<Integer> groupIDs = queryInstance.getGroupIDs();

		if (groupIDs == null) {

			sendEmptyJSONResponse(res);
			return;
		}

		List<User> users = userHandler.searchUsers(searchString, false, true, internalUserQueryProviderModule.getSearchResultLimit(), groupIDs, query.isExcludeInactiveUsers());

		if (query.isSearchCitizenID() && searchString.matches("[0-9-]+")) {

			List<User> citizenIDUsers = null;

			citizenIDUsers = CollectionUtils.addAndInstantiateIfNeeded(citizenIDUsers, userHandler.getUsersByAttribute(Constants.MANAGER_CITIZEN_IDENTIFIER_ATTRIBUTE, searchString, true, true));
			citizenIDUsers = CollectionUtils.addAndInstantiateIfNeeded(citizenIDUsers, userHandler.getUsersByAttribute(Constants.USER_CITIZEN_IDENTIFIER_ATTRIBUTE, searchString, true, true));

			if (citizenIDUsers != null) {

				Iterator<User> it = citizenIDUsers.iterator();
				while (it.hasNext()) {
					User nextUser = it.next();

					if (CollectionUtils.isEmpty(CollectionUtils.conjunction(groupIDs, UserUtils.getUserGroupIDs(nextUser)))) {
						it.remove();
					}
				}

				if (!citizenIDUsers.isEmpty()) {
					@SuppressWarnings("unchecked")
					HashSet<User> usersSet = CollectionUtils.combineAsSet(users, citizenIDUsers);

					users = new ArrayList<>(usersSet);
				}
			}
		}

		if (query.isSearchDepartment()) {

			List<User> departmentUsers = userHandler.getUsersByAttributeWildcard("department", searchString, true, true);

			if (departmentUsers != null) {

				Iterator<User> it = departmentUsers.iterator();
				while (it.hasNext()) {
					User nextUser = it.next();

					if (CollectionUtils.isEmpty(CollectionUtils.conjunction(groupIDs, UserUtils.getUserGroupIDs(nextUser)))) {
						it.remove();
					}
				}

				if (!departmentUsers.isEmpty()) {
					@SuppressWarnings("unchecked")
					HashSet<User> usersSet = CollectionUtils.combineAsSet(users, departmentUsers);

					users = new ArrayList<>(usersSet);
				}
			}
		}

		if (query.isSearchTitle()) {

			List<User> titleUsers = userHandler.getUsersByAttributeWildcard("title", searchString, true, true);

			if (titleUsers != null) {

				Iterator<User> it = titleUsers.iterator();
				while (it.hasNext()) {
					User nextUser = it.next();

					if (CollectionUtils.isEmpty(CollectionUtils.conjunction(groupIDs, UserUtils.getUserGroupIDs(nextUser)))) {
						it.remove();
					}
				}

				if (!titleUsers.isEmpty()) {
					@SuppressWarnings("unchecked")
					HashSet<User> usersSet = CollectionUtils.combineAsSet(users, titleUsers);

					users = new ArrayList<>(usersSet);
				}
			}
		}

		if (users != null && (query.isLimitSearchToMyEmployees() || query.isBlockPosterSelection())) {

			users = filterUsers(users, poster, query);
		}
		
		log.info("User " + user + " searching for users using query " + searchString + ", found " + CollectionUtils.getSize(users) + " hits");

		if (users == null) {

			sendEmptyJSONResponse(res);
			return;
		}

		if (query.isSearchCitizenID() || query.isSearchDepartment() || query.isSearchTitle()) {
			Collections.sort(users, UserNameComparator.getInstance());
		}

		JsonArray jsonArray = new JsonArray();

		for (User currentUser : users) {

			JsonObject instance = new JsonObject(4);
			instance.putField("ID", currentUser.getUserID());
			instance.putField("Enabled", currentUser.isEnabled());

			if (currentUser.getUsername() != null) {
				instance.putField("Username", internalUserQueryProviderModule.formatUsername(currentUser.getUsername()));
			}

			String citizenID = CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(currentUser);

			if (citizenID != null) {
				
				if (queryInstance.getQuery().isMaskCitizenIDLastFourDigits()) {

					citizenID = CitizenIdentifierUtils.maskCitizenIDLastFourDigits(citizenID);
				}
				
				instance.putField("CitizenID", citizenID);
			}

			StringBuilder label = new StringBuilder();
			label.append(currentUser.getFirstname());
			label.append(" ");
			label.append(currentUser.getLastname());

			String username = currentUser.getUsername();
			String department = currentUser.getAttributeHandler().getString("department");
			String title = currentUser.getAttributeHandler().getString("title");

			if (username != null || citizenID != null && query.isSearchCitizenID() || department != null && query.isSearchDepartment() || title != null && query.isSearchTitle()) {

				label.append(" (");
				boolean prev = false;

				if (currentUser.getUsername() != null) {
					label.append(currentUser.getUsername());
					prev = true;
				}

				if (department != null && query.isSearchDepartment()) {
					if (prev) {
						label.append(", ");
					}

					label.append(department);
					prev = true;
				}

				if (title != null && query.isSearchTitle()) {
					if (prev) {
						label.append(", ");
					}

					label.append(title);
					prev = true;
				}

				if (citizenID != null && query.isSearchCitizenID()) {
					if (prev) {
						label.append(", ");
					}

					label.append(citizenID);
					prev = true;
				}

				label.append(")");
			}

			instance.putField("Label", label.toString());

			instance.putField("Name", currentUser.getFirstname() + " " + currentUser.getLastname());

			if (query.isShowEmail() && currentUser.getUsername() != null) {

				instance.putField("Email", currentUser.getEmail());
			}

			if (query.getSelectedUserAttributes() != null) {

				JsonObject attributes = new JsonObject(query.getSelectedUserAttributes().size());

				for (SelectedUserAttribute selectedAttribute : queryInstance.getQuery().getSelectedUserAttributes()) {

					String name = selectedAttribute.getName();
					String value = currentUser.getAttributeHandler().getString(name);

					if (value != null) {

						if (CitizenIdentifierUtils.isCitizenIdentifierAttribute(name) && queryInstance.getQuery().isMaskCitizenIDLastFourDigits()) {

							value = CitizenIdentifierUtils.maskCitizenIDLastFourDigits(value);
						}

						attributes.putField(name, internalUserQueryProviderModule.formatAttribute(selectedAttribute, value));
					}
				}

				instance.putField("Attributes", attributes);
			}

			if (query.isShowManagerInfo()) {

				String managerUsername = currentUser.getAttributeHandler().getString(InternalUserQueriesConstants.USER_MANAGER_USERNAME_ATTRIBUTE);

				if (managerUsername != null) {

					User manager = userHandler.getUserByUsername(managerUsername, false, true);

					if (manager != null) {

						JsonObject managerJson = new JsonObject();

						if (query.isShowManagerName()) {
							managerJson.putField("Name", manager.getFirstname() + " " + manager.getLastname());
						}

						if (query.isShowManagerUsername() && manager.getUsername() != null) {
							managerJson.putField("Username", manager.getUsername());
						}

						if (query.isShowManagerEmail()) {
							managerJson.putField("Email", manager.getEmail());
						}

						if (query.getSelectedManagerAttributes() != null) {

							JsonObject attributes = new JsonObject(query.getSelectedManagerAttributes().size());

							for (SelectedManagerAttribute selectedAttribute : queryInstance.getQuery().getSelectedManagerAttributes()) {

								String name = selectedAttribute.getName();
								String value = manager.getAttributeHandler().getString(name);

								if (value != null) {

									if (CitizenIdentifierUtils.isCitizenIdentifierAttribute(name) && queryInstance.getQuery().isMaskCitizenIDLastFourDigits()) {

										value = CitizenIdentifierUtils.maskCitizenIDLastFourDigits(value);
									}

									attributes.putField(name, internalUserQueryProviderModule.formatAttribute(selectedAttribute, value));
								}
							}

							managerJson.putField("Attributes", attributes);
						}

						instance.putField("Manager", managerJson);
					}
				}
			}

			jsonArray.addNode(instance);
		}

		sendJSONResponse(jsonArray, res);
	}

	private List<User> filterUsers(List<User> users, User poster, InternalUserQuery query) {
		
		if (poster == null || poster.getUsername() == null) {
			
			return null;
		}
		
		Iterator<User> it = users.iterator();
		String username = poster.getUsername();

		while (it.hasNext()) {

			User currentUser = it.next();
			
			if (query.isLimitSearchToMyEmployees() && !username.equals(currentUser.getAttributeHandler().getString(InternalUserQueriesConstants.USER_MANAGER_USERNAME_ATTRIBUTE))) {
				
				it.remove();
				continue;
			}
			
			if (query.isBlockPosterSelection() && poster.equals(currentUser)) {
				
				it.remove();
			}
		}
		
		return users;
	}

	@Override
	public void close() throws Exception {

	}
}
