package com.nordicpeak.flowengine.queue;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.sql.DataSource;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.handlers.UserHandler;
import se.unlogic.hierarchy.core.utils.UserUtils;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.querys.ArrayListQuery;
import se.unlogic.standardutils.dao.querys.BooleanQuery;
import se.unlogic.standardutils.dao.querys.ObjectQuery;
import se.unlogic.standardutils.populators.IntegerPopulator;
import se.unlogic.standardutils.populators.TimeStampPopulator;
import se.unlogic.standardutils.string.StringUtils;

import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.queue.beans.Queue;
import com.nordicpeak.flowengine.queue.beans.QueueSlot;
import com.nordicpeak.flowengine.queue.beans.QueueSlotEvent;
import com.nordicpeak.flowengine.queue.enums.QueueSlotEventType;

public class QueueUtils {

	public static void getCalculatedQueuePosition(DataSource dataSource, QueueSlot queueSlot) throws SQLException {

		ObjectQuery<Integer> query = new ObjectQuery<>(dataSource, "SELECT COUNT(*) FROM queue_slots WHERE queueID = ? AND archived = 0 AND queueIndex <= ?", IntegerPopulator.getPopulator());
		query.setInt(1, queueSlot.getQueue().getQueueID());
		query.setInt(2, queueSlot.getQueueIndex());

		queueSlot.setCalculatedPosition(query.executeQuery());
	}

	public static boolean isActiveMemberOfQueue(DataSource dataSource, Queue queue, String citizenIdentifier) throws SQLException {

		//TODO don't select *
		BooleanQuery query = new BooleanQuery(dataSource, "SELECT * FROM queue_slot_contacts INNER JOIN queue_slots ON queue_slot_contacts.queueSlotID = queue_slots.queueSlotID WHERE citizenIdentifier = ? AND queueID = ? AND archived = 0");

		query.setString(1, citizenIdentifier);
		query.setInt(2, queue.getQueueID());

		return query.executeQuery();
	}
	
	public static List<Timestamp> getWaitingPeriods(DataSource dataSource, Queue queue, String citizenIdentifier) throws SQLException {

		ArrayListQuery<Timestamp> query = new ArrayListQuery<>(dataSource, "SELECT waitingPeriodTime FROM queue_slots INNER JOIN queues ON queue_slots.queueID=queues.queueID INNER JOIN queue_slot_contacts ON queue_slot_contacts.queueSlotID = queue_slots.queueSlotID WHERE citizenIdentifier = ? AND queue_slots.queueID = ? AND queues.useWaitingPeriod='1' order by waitingPeriodTime desc", TimeStampPopulator.getPopulator());

		query.setString(1, citizenIdentifier);
		query.setInt(2, queue.getQueueID());

		return query.executeQuery();
	}
	
	public static LocalDate getLatestWaitingDate(DataSource dataSource, Queue queue, String citizenIdentifier) throws SQLException {

		if(queue != null && queue.isUseWaitingPeriod()) {
			List<Timestamp> waitingPeriods = getWaitingPeriods(dataSource, queue, citizenIdentifier);
			
			if(waitingPeriods != null) {
				waitingPeriods.removeAll(Collections.singleton(null));
			}
			
			if(!CollectionUtils.isEmpty(waitingPeriods)) {
				return LocalDateTime.ofInstant(waitingPeriods.get(0).toInstant(), ZoneId.systemDefault()).toLocalDate();
			}
		}
		
		return null;
		
	}
	
	public static List<Integer> getQueueSlotIDs(DataSource dataSource, Queue queue, String citizenIdentifier) throws SQLException {

		ArrayListQuery<Integer> query = new ArrayListQuery<>(dataSource, "SELECT queueSlotID FROM queue_slots INNER JOIN queue_slot_contacts ON queue_slot_contacts.queueSlotID = queue_slots.queueSlotID WHERE citizenIdentifier = ? AND queueID = ?", IntegerPopulator.getPopulator());

		query.setString(1, citizenIdentifier);
		query.setInt(2, queue.getQueueID());

		return query.executeQuery();
	}
	
	public static boolean isDigitalMember(QueueSlot queueSlot, UserHandler userHandler) {

		User poster = queueSlot.getPoster(userHandler);

		if (poster != null && (!StringUtils.isEmpty(poster.getEmail()) || !StringUtils.isEmpty(queueSlot.getPrimaryContact().getEmail()))) {

			return true;
		}

		return false;
	}

	public static Integer getActiveQueueSlotCount(Queue queue, AnnotatedDAO<Queue> queueDAO) throws SQLException {

		ObjectQuery<Integer> query = new ObjectQuery<>(queueDAO.getDataSource(), "SELECT COUNT(*) FROM queue_slots WHERE queueID = ? AND archived = 0", IntegerPopulator.getPopulator());

		query.setInt(1, queue.getQueueID());

		return query.executeQuery();
	}

	public static Integer getArchivedQueueSlotCount(Queue queue, AnnotatedDAO<Queue> queueDAO) throws SQLException {

		ObjectQuery<Integer> query = new ObjectQuery<>(queueDAO.getDataSource(), "SELECT COUNT(*) FROM queue_slots WHERE queueID = ? AND archived = 1", IntegerPopulator.getPopulator());

		query.setInt(1, queue.getQueueID());

		return query.executeQuery();
	}

	public static QueueSlotEvent getLatestPaymentEvent(QueueSlot queueSlot) {

		if (CollectionUtils.isEmpty(queueSlot.getEvents())) {

			return null;
		}

		for (int i = queueSlot.getEvents().size() - 1; i > -1; i--) {

			QueueSlotEvent event = queueSlot.getEvents().get(i);

			if (event.getEventType() == QueueSlotEventType.PAYED) {

				return event;
			}
		}

		return null;
	}

	public static void setQueueSlotUsers(Queue queue, List<QueueSlot> queueSlots, UserHandler userHandler) {

		if (queueSlots == null) {

			return;
		}

		Set<String> citizenIdentifiers = new HashSet<>();

		for (QueueSlot slot : queueSlots) {

			citizenIdentifiers.add(slot.getPrimaryContact().getCitizenIdentifier());
		}

		if (citizenIdentifiers.isEmpty()) {

			return;
		}

		List<User> externalUsers = userHandler.getUsersByAttribute(Constants.USER_CITIZEN_IDENTIFIER_ATTRIBUTE, citizenIdentifiers, false, true);
		List<User> internalUsers = userHandler.getUsersByAttribute(Constants.MANAGER_CITIZEN_IDENTIFIER_ATTRIBUTE, citizenIdentifiers, false, true);

		if (CollectionUtils.isEmpty(externalUsers) && CollectionUtils.isEmpty(internalUsers)) {

			return;
		}

		Map<String, User> externalUserMap = UserUtils.getUserAttributeMap(externalUsers, Constants.USER_CITIZEN_IDENTIFIER_ATTRIBUTE);
		Map<String, User> internalUserMap = UserUtils.getUserAttributeMap(internalUsers, Constants.MANAGER_CITIZEN_IDENTIFIER_ATTRIBUTE);

		for (QueueSlot slot : queueSlots) {

			String citizenIdentifier;

			if (slot.getPrimaryContact() == null || (citizenIdentifier = slot.getPrimaryContact().getCitizenIdentifier()) == null) {

				continue;
			}

			if (queue.isUseQueueForInternalUsers() && !CollectionUtils.isEmpty(queue.getInternalGroupIDs()) && !CollectionUtils.isEmpty(internalUserMap)) {

				// For internal calendars, fetch citizen identifier from internalUserMap
				slot.setPosterFromPrimaryContactCitizenIdentifier(internalUserMap.get(citizenIdentifier));

			} else {

				if (!CollectionUtils.isEmpty(externalUserMap) && externalUserMap.get(citizenIdentifier) != null) {

					slot.setPosterFromPrimaryContactCitizenIdentifier(externalUserMap.get(citizenIdentifier));

				} else if (!CollectionUtils.isEmpty(internalUserMap) && internalUserMap.get(citizenIdentifier) != null) {

					slot.setPosterFromPrimaryContactCitizenIdentifier(internalUserMap.get(citizenIdentifier));
				}
			}
		}
	}

}
