package com.nordicpeak.flowengine.queue.cruds;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.hierarchy.core.utils.AccessUtils;
import se.unlogic.hierarchy.core.utils.crud.FragmentLinkScriptFilter;
import se.unlogic.hierarchy.core.utils.crud.IntegerBeanIDParser;
import se.unlogic.hierarchy.core.utils.crud.ModularCRUD;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.populators.StringPopulator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.queue.QueueAdminModule;
import com.nordicpeak.flowengine.queue.QueueUtils;
import com.nordicpeak.flowengine.queue.beans.Queue;

public class QueueCRUD extends ModularCRUD<Queue, Integer, User, QueueAdminModule> {

	protected final AnnotatedDAOWrapper<Queue, Integer> queueDAOWrapper;

	public QueueCRUD(AnnotatedDAOWrapper<Queue, Integer> queueDAOWrapper, QueueAdminModule callback) {

		super(IntegerBeanIDParser.getInstance(), queueDAOWrapper, new AnnotatedRequestPopulator<>(Queue.class), "Queue", "queue", "", callback);

		this.queueDAOWrapper = queueDAOWrapper;

		setRequirePostForDelete(true);
	}

	@Override
	protected void appendListFormData(Document doc, Element listTypeElement, User user, HttpServletRequest req, URIParser uriParser, List<ValidationError> validationError) throws SQLException {

		super.appendListFormData(doc, listTypeElement, user, req, uriParser, validationError);

		if (callback.isEnableProfileSupport() && callback.getSiteProfileHandler() != null) {

			XMLUtils.append(doc, listTypeElement, "Profiles", callback.getSiteProfileHandler().getProfiles());
		}
	}

	@Override
	protected void appendAddFormData(Document doc, Element addTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		super.appendAddFormData(doc, addTypeElement, user, req, uriParser);

		appendFormData(null, doc, addTypeElement, user, req, uriParser);
	}

	@Override
	protected void appendUpdateFormData(Queue queue, Document doc, Element updateTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		super.appendUpdateFormData(queue, doc, updateTypeElement, user, req, uriParser);

		AccessUtils.appendAllowedGroupsAndUsers(doc, updateTypeElement, queue, callback.getUserHandler(), callback.getGroupHandler());

		appendInternalGroups(doc, updateTypeElement, queue);

		appendFormData(queue, doc, updateTypeElement, user, req, uriParser);
	}

	private void appendInternalGroups(Document doc, Element targetElement, Queue queue) {

		Collection<Integer> internalGroupIDs = queue.getInternalGroupIDs();

		if (!CollectionUtils.isEmpty(internalGroupIDs)) {

			XMLUtils.append(doc, targetElement, "internalGroups", callback.getGroupHandler().getGroups(internalGroupIDs, false));
		}
	}

	protected void appendFormData(Queue queue, Document doc, Element element, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		XMLUtils.appendNewElement(doc, element, "DefaultEventPaymentInvoiceMessage", callback.getDefaultEventPaymentInvoiceMessage());

		if (callback.getPaymentProvider() != null) {

			try {

				String paymentConfigurationID = null;

				if (queue != null) {

					paymentConfigurationID = queue.getPaymentConfigurationID();
				}

				ViewFragment paymentSettings = callback.getPaymentProvider().showConfigurationForm(paymentConfigurationID, req, user, uriParser, (ValidationException) req.getAttribute("paymentValidationException"));

				if (paymentSettings != null) {

					element.appendChild(paymentSettings.toXML(doc));
					FragmentLinkScriptFilter.addViewFragment(paymentSettings, req);
				}

			} catch (Exception e) {

				log.error("Exception while showing payment configuration", e);
			}
		}

		if (callback.isEnableProfileSupport() && callback.getSiteProfileHandler() != null) {

			XMLUtils.append(doc, element, "Profiles", callback.getSiteProfileHandler().getProfiles());
		}
	}

	@Override
	protected Queue populateFromAddRequest(HttpServletRequest req, User user, URIParser uriParser) throws ValidationException, Exception {

		ArrayList<ValidationError> validationErrors = new ArrayList<>();

		Queue queue = null;

		try {
			queue = super.populateFromAddRequest(req, user, uriParser);

		} catch (ValidationException e) {

			validationErrors.addAll(e.getErrors());
		}

		if (callback.getPaymentProvider() != null && queue != null && queue.getPrice() > 0) {

			queue.setEventPaymentInvoiceMessage(ValidationUtils.validateParameter("eventPaymentInvoiceMessage", req, true, StringPopulator.getPopulator(), validationErrors));

			try {
				callback.getPaymentProvider().processPaymentConfiguration(null, req, user, uriParser);

			} catch (ValidationException e) {

				req.setAttribute("paymentValidationException", e);
				validationErrors.add(new ValidationError("PaymentProviderConfigurationError"));
			}
		}

		if (callback.isEnableProfileSupport() && queue != null && queue.getProfileID() == null) {

			validationErrors.add(new ValidationError("NoProfileChosen"));
		}

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		return queue;
	}

	@Override
	protected ForegroundModuleResponse filteredBeanAdded(Queue queue, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		ForegroundModuleResponse response = super.filteredBeanAdded(queue, req, res, user, uriParser);

		if (callback.getPaymentProvider() != null) {

			callback.getPaymentProvider().processPaymentConfiguration(queue.getPaymentConfigurationID(), req, user, uriParser);
		}

		return response;
	}

	@Override
	protected ForegroundModuleResponse filteredBeanDeleted(Queue queue, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		ForegroundModuleResponse response = super.filteredBeanDeleted(queue, req, res, user, uriParser);

		if (callback.getPaymentProvider() != null) {

			callback.getPaymentProvider().deletePaymentConfiguration(queue.getPaymentConfigurationID());
		}

		return response;
	}

	@Override
	protected Queue populateFromUpdateRequest(Queue oldQueue, HttpServletRequest req, User user, URIParser uriParser) throws ValidationException, Exception {

		ArrayList<ValidationError> validationErrors = new ArrayList<>();

		Queue queue = oldQueue;

		try {
			queue = super.populateFromUpdateRequest(oldQueue, req, user, uriParser);

		} catch (ValidationException e) {

			validationErrors.addAll(e.getErrors());
		}

		if (callback.getPaymentProvider() != null) {

			if (queue.getPrice() > 0) {

				queue.setEventPaymentInvoiceMessage(ValidationUtils.validateParameter("eventPaymentInvoiceMessage", req, true, StringPopulator.getPopulator(), validationErrors));

				try {
					callback.getPaymentProvider().processPaymentConfiguration(queue.getPaymentConfigurationID(), req, user, uriParser);

				} catch (ValidationException e) {

					req.setAttribute("paymentValidationException", e);
					validationErrors.add(new ValidationError("PaymentProviderConfigurationError"));
				}

			} else {

				callback.getPaymentProvider().deletePaymentConfiguration(queue.getPaymentConfigurationID());
			}
		}

		if (callback.isEnableProfileSupport() && queue.getProfileID() == null) {

			validationErrors.add(new ValidationError("NoProfileChosen"));
		}

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		return queue;
	}

	@Override
	protected List<Queue> getAllBeans(User user, HttpServletRequest req, URIParser uriParser) throws SQLException {

		List<Queue> queues = super.getAllBeans(user, req, uriParser);

		if (!CollectionUtils.isEmpty(queues)) {
			for (Queue queue : queues) {

				queue.setActiveSlots(QueueUtils.getActiveQueueSlotCount(queue, queueDAOWrapper.getAnnotatedDAO()));
				queue.setArchivedSlots(QueueUtils.getArchivedQueueSlotCount(queue, queueDAOWrapper.getAnnotatedDAO()));
			}
		}

		return queues;
	}

}
