package com.nordicpeak.flowengine.reservation;

import java.io.IOException;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.cron4jutils.CronStringValidator;
import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.listeners.ServerStartupListener;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.hierarchy.foregroundmodules.rest.RESTMethod;
import se.unlogic.hierarchy.foregroundmodules.rest.URIParam;
import se.unlogic.hierarchy.foregroundmodules.staticcontent.StaticContentModule;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonNode;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.json.JsonUtils;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.BaseFlowModule;
import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.FlowBrowserModule;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.enums.EventType;
import com.nordicpeak.flowengine.events.SubmitEvent;
import com.nordicpeak.flowengine.interfaces.ReservationProvider;
import com.nordicpeak.flowengine.interfaces.ReservationQueryInstance;
import com.nordicpeak.flowengine.managers.FlowInstanceManager;
import com.nordicpeak.flowengine.reservation.beans.Reservation;
import com.nordicpeak.flowengine.reservation.daos.ReservationDAO;

import it.sauronsoftware.cron4j.Scheduler;

public class ReservationModule extends AnnotatedRESTModule implements ReservationProvider, Runnable, ServerStartupListener {

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Interval to delete expired reservations", description = "How often this module should delete expired reservations (specified in crontab format)", required = true, formatValidator = CronStringValidator.class)
	private String deleteExpiredReservationsInterval = "* * * * *";

	@InstanceManagerDependency(required = true)
	private StaticContentModule staticContentModule;

	@InstanceManagerDependency(required = true)
	private FlowAdminModule flowAdminModule;

	private ReservationDAO reservationDAO;

	private Scheduler scheduler;
	private String deleteExpiredReservationsScheduleID;
	private boolean running;

	private AbortedReservationListener abortedReservationListener = new AbortedReservationListener(this);

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, ReservationModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("daos/DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.warn(upgradeResult.toString());
		}

		HierarchyAnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface);

		reservationDAO = new ReservationDAO(dataSource, Reservation.class, daoFactory);
	}

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);

		if (!systemInterface.getInstanceHandler().addInstance(ReservationProvider.class, this)) {

			throw new RuntimeException("Unable to register module " + moduleDescriptor + " in global instance handler using key " + ReservationProvider.class.getSimpleName() + ", another instance is already registered using this key.");
		}

		systemInterface.addServerStartupListener(this);

		abortedReservationListener.init(systemInterface);
	}

	@Override
	public void update(ForegroundModuleDescriptor descriptor, DataSource dataSource) throws Exception {

		super.update(descriptor, dataSource);

		scheduler.reschedule(deleteExpiredReservationsScheduleID, deleteExpiredReservationsInterval);
	}

	@Override
	public void unload() throws Exception {

		systemInterface.getInstanceHandler().removeInstance(ReservationProvider.class, this);

		stopScheduler();

		abortedReservationListener.unload(systemInterface);

		super.unload();
	}

	@Override
	public void serverStarted() throws Exception {

		initScheduler();
	}

	@Override
	public int getPriority() {

		return 0;
	}

	protected synchronized void initScheduler() {

		if (scheduler != null) {

			log.warn("Invalid state, scheduler already running!");
			stopScheduler();
		}

		scheduler = new Scheduler(systemInterface.getApplicationName() + " - " + moduleDescriptor.toString());
		scheduler.setDaemon(true);

		deleteExpiredReservationsScheduleID = scheduler.schedule(deleteExpiredReservationsInterval, this);

		scheduler.start();
	}

	protected synchronized void stopScheduler() {

		try {

			if (scheduler != null) {

				scheduler.stop();
				scheduler = null;
			}

		} catch (IllegalStateException e) {

			log.error("Error stopping scheduler", e);
		}
	}

	@Override
	public void run() {

		if (running) {

			return;
		}

		running = true;

		try {

			Collection<Integer> flowFamilyIDs = reservationDAO.getExpiredFlowFamilyIDs();

			if (flowFamilyIDs == null) {

				return;
			}

			for (Integer flowFamilyID : flowFamilyIDs) {

				synchronized (BaseFlowModule.getSubmitLockMutexKey(flowFamilyID)) {

					List<Reservation> expiredReservations = reservationDAO.getExpiredByFlowFamilyID(flowFamilyID);

					if (expiredReservations == null) {

						continue;
					}

					try (TransactionHandler transactionHandler = reservationDAO.createTransaction()) {

						for (Reservation reservation : expiredReservations) {

							log.info("Deleting expired reservation " + reservation);

							reservationDAO.delete(reservation, transactionHandler);
						}

						transactionHandler.commit();
					}
				}
			}

		} catch (Throwable t) {

			log.error("Error deleting expired reservations", t);

		} finally {

			running = false;
		}
	}

	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws URINotFoundException, SQLException {

		Document doc = createDocument(req, uriParser, user);
		Element listReservationsElement = doc.createElement("ListReservations");
		doc.getFirstChild().appendChild(listReservationsElement);

		log.warn("User " + user + " listing all reservations");

		return new SimpleForegroundModuleResponse(doc);
	}

	@RESTMethod(alias = "reservations", method = { "get" }, requireLogin = true)
	public ForegroundModuleResponse getReservations(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws SQLException, IOException {

		log.info("User " + user + " requesting reservations");
		
		List<Reservation> reservations = reservationDAO.getAll();

		JsonObject jsonObject = new JsonObject(1);

		if (!CollectionUtils.isEmpty(reservations)) {

			JsonArray reservationsArray = new JsonArray(reservations.size());

			for (Reservation reservation : reservations) {

				Flow flow = getFlow(reservation);

				JsonNode json = reservation.toJSON(flow);

				reservationsArray.addNode(json);
			}

			jsonObject.putField("data", reservationsArray);

		} else {

			jsonObject.putField("data", new JsonArray());
		}

		HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);

		return null;
	}

	private Flow getFlow(Reservation reservation) {

		FlowFamily flowFamily = flowAdminModule.getFlowFamily(reservation.getFlowFamilyID());

		if (flowFamily == null) {

			return null;
		}

		return flowAdminModule.getLatestFlowVersion(flowFamily);
	}

	@RESTMethod(alias = "reservation/{reservationID}/delete", method = { "post" }, requireLogin = true)
	public ForegroundModuleResponse deleteReservation(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "reservationID") String reservationID) throws Throwable {

		Reservation reservation = reservationDAO.getByReservationID(reservationID);

		if (reservation == null) {

			throw new URINotFoundException(uriParser);
		}

		Integer flowFamilyID = reservation.getFlowFamilyID();

		synchronized (flowFamilyID) {

			reservation = reservationDAO.getByReservationID(reservationID);

			if (reservation != null) {

				log.info("User " + user + " deleting reservation " + reservation);

				reservationDAO.delete(reservation);
			}
		}

		redirectToDefaultMethod(req, res);
		return null;
	}

	@Override
	public List<String> getReservedObjectIDs(List<String> objectIDs, Integer flowFamilyID, List<String> excludedReservationIDs) throws SQLException {

		if (CollectionUtils.isEmpty(objectIDs)) {

			return null;
		}

		Objects.requireNonNull(flowFamilyID);

		List<Reservation> reservations = reservationDAO.getByObjectIDsAndFlowFamilyID(objectIDs, flowFamilyID, excludedReservationIDs);

		return CollectionUtils.map(reservations, Reservation::getObjectID);
	}

	@Override
	public String createReservation(String objectID, String oldReservationID, Integer flowFamilyID, Integer flowInstanceID, String flowInstanceManagerID, Integer reservationMinutes) throws SQLException {

		Objects.requireNonNull(objectID);
		Objects.requireNonNull(flowFamilyID);
		Objects.requireNonNull(reservationMinutes);

		if (flowInstanceID != null) {

			flowInstanceManagerID = null;
		}

		synchronized (BaseFlowModule.getSubmitLockMutexKey(flowFamilyID)) {

			Reservation existingReservation = reservationDAO.getByObjectIDAndFlowFamilyID(objectID, flowFamilyID);

			Timestamp now = TimeUtils.getCurrentTimestamp();
			Timestamp expiration = Timestamp.valueOf(now.toLocalDateTime().plusMinutes(reservationMinutes));

			if (existingReservation != null) {

				return updateReservation(existingReservation, oldReservationID, flowInstanceManagerID, flowInstanceID, now, expiration);
			}

			Reservation reservation = new Reservation(objectID, flowFamilyID, flowInstanceManagerID, flowInstanceID, now, expiration);

			log.info("Creating new reservation " + reservation);

			reservationDAO.add(reservation);

			return reservation.getReservationID();
		}
	}

	private String updateReservation(Reservation existingReservation, String oldReservationID, String flowInstanceManagerID, Integer flowInstanceID, Timestamp now, Timestamp expiration) throws SQLException {

		if (oldReservationID == null || !oldReservationID.equals(existingReservation.getReservationID())) {

			log.info("Old reservationID (" + oldReservationID + ") does not match existing reservation " + existingReservation);
			return null;
		}

		if (existingReservation.getExpires() == null) {

			log.info("Existing reservation " + existingReservation + " has no expiration, not updating.");
			return existingReservation.getReservationID();
		}

		existingReservation.setFlowInstanceManagerID(flowInstanceManagerID);
		existingReservation.setFlowInstanceID(flowInstanceID);
		existingReservation.setAdded(now);
		existingReservation.setExpires(expiration);

		log.info("Updating existing reservation " + existingReservation);

		reservationDAO.update(existingReservation);

		return existingReservation.getReservationID();
	}

	@Override
	public boolean deleteReservation(String reservationID, Integer flowFamilyID) throws Exception {

		return deleteReservations(Collections.singletonList(reservationID), flowFamilyID) > 0;
	}

	@Override
	public int deleteReservations(List<String> reservationIDs, Integer flowFamilyID) throws Exception {

		if (reservationIDs == null) {

			return 0;
		}

		Objects.requireNonNull(flowFamilyID);

		int deletedCount = 0;

		synchronized (BaseFlowModule.getSubmitLockMutexKey(flowFamilyID)) {

			try (TransactionHandler transactionHandler = reservationDAO.createTransaction()) {

				List<Reservation> reservations = reservationDAO.getByReservationIDs(reservationIDs, transactionHandler);

				if (reservations == null) {

					return 0;
				}

				for (Reservation reservation : reservations) {

					log.info("Deleting reservation " + reservation);

					reservationDAO.delete(reservation, transactionHandler);

					deletedCount++;
				}

				transactionHandler.commit();
			}
		}

		return deletedCount;
	}

	public int deleteReservationsByFlowInstanceManagerID(String flowInstanceManagerID) throws Exception {

		if (flowInstanceManagerID == null) {

			return 0;
		}

		Collection<Integer> flowFamilyIDs = reservationDAO.getFlowFamilyIDsByFlowInstanceManagerID(flowInstanceManagerID);

		if (flowFamilyIDs == null) {

			return 0;
		}

		int deletedCount = 0;

		for (Integer flowFamilyID : flowFamilyIDs) {

			synchronized (BaseFlowModule.getSubmitLockMutexKey(flowFamilyID)) {

				try (TransactionHandler transactionHandler = reservationDAO.createTransaction()) {

					List<Reservation> reservations = reservationDAO.getByFlowInstanceManagerIDAndFlowFamilyID(flowInstanceManagerID, flowFamilyID, transactionHandler);

					if (reservations == null) {

						continue;
					}

					for (Reservation reservation : reservations) {

						log.info("Deleting reservation " + reservation);

						reservationDAO.delete(reservation, transactionHandler);

						deletedCount++;
					}

					transactionHandler.commit();
				}
			}
		}

		return deletedCount;
	}

	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {

		Document doc = XMLUtils.createDomDocument();
		Element document = doc.createElement("Document");
		document.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		document.appendChild(this.moduleDescriptor.toXML(doc));

		XMLUtils.appendNewElement(doc, document, "ModuleURI", req.getContextPath() + getFullAlias());
		XMLUtils.appendNewElement(doc, document, "StaticContentURL", staticContentModule.getModuleContentURL(moduleDescriptor));

		doc.appendChild(document);

		return doc;
	}

	@EventListener(channel = FlowInstanceManager.class)
	public void processEvent(SubmitEvent event, EventSource source) {

		if (event.getEvent().getEventType() != EventType.SUBMITTED || !FlowBrowserModule.SUBMIT_ACTION_ID.equals(event.getActionID())) {

			return;
		}

		List<ReservationQueryInstance> queryInstances = event.getFlowInstanceManager().getQueries(ReservationQueryInstance.class);

		if (queryInstances == null) {

			return;
		}

		Integer flowInstanceID = event.getFlowInstanceManager().getFlowInstanceID();

		try (TransactionHandler transactionHandler = reservationDAO.createTransaction()) {

			List<String> reservationIDs = CollectionUtils.flatMap(queryInstances, ReservationQueryInstance::getReservationIDs);

			if (reservationIDs == null) {

				return;
			}

			List<Reservation> reservations = reservationDAO.getByReservationIDs(reservationIDs, transactionHandler);

			if (reservations == null) {

				return;
			}

			for (Reservation reservation : reservations) {

				reservation.setFlowInstanceManagerID(null);
				reservation.setFlowInstanceID(flowInstanceID);
				reservation.setExpires(null);
			}

			reservationDAO.update(reservations, transactionHandler, null);

			transactionHandler.commit();

		} catch (Exception e) {

			log.error("Error handling submit event " + event, e);

		}
	}

	@EventListener(channel = FlowInstance.class)
	public void processFlowInstanceEvent(CRUDEvent<FlowInstance> event, EventSource eventSource) {

		if (event.getAction() != CRUDAction.DELETE) {

			return;
		}

		for (FlowInstance flowInstance : event.getBeans()) {

			log.info("Deleting reservations for " + flowInstance);

			try {

				reservationDAO.deleteByFlowInstanceID(flowInstance);

			} catch (SQLException e) {

				log.info("Error deleting reservations for " + flowInstance, e);
			}
		}
	}

}