package com.nordicpeak.flowengine.reservation.daos;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.sql.DataSource;

import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.AnnotatedDAOFactory;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryOperators;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.time.TimeUtils;

import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.reservation.beans.Reservation;

public class ReservationDAO extends AnnotatedDAO<Reservation> {

	private QueryParameterFactory<Reservation, String> reservationIDParamFactory;
	private QueryParameterFactory<Reservation, String> objectIDParamFactory;
	private QueryParameterFactory<Reservation, Integer> flowFamilyIDParamFactory;
	private QueryParameterFactory<Reservation, String> flowInstanceManagerIDParamFactory;
	private QueryParameterFactory<Reservation, Integer> flowInstanceIDParamFactory;
	private QueryParameterFactory<Reservation, Timestamp> expiresParamFactory;

	public ReservationDAO(DataSource dataSource, Class<Reservation> beanClass, AnnotatedDAOFactory daoFactory) {

		super(dataSource, beanClass, daoFactory);

		reservationIDParamFactory = this.getParamFactory(Reservation.FIELD_RESERVATION_ID, String.class);
		objectIDParamFactory = this.getParamFactory(Reservation.FIELD_OBJECT_ID, String.class);
		flowFamilyIDParamFactory = this.getParamFactory(Reservation.FIELD_FLOW_FAMILY_ID, Integer.class);
		flowInstanceManagerIDParamFactory = this.getParamFactory(Reservation.FIELD_FLOW_INSTANCE_MANAGER_ID, String.class);
		flowInstanceIDParamFactory = this.getParamFactory(Reservation.FIELD_FLOW_INSTANCE_ID, Integer.class);
		expiresParamFactory = this.getParamFactory(Reservation.FIELD_EXPIRES, Timestamp.class);
	}

	public Reservation getByReservationID(String reservationID) throws SQLException {

		HighLevelQuery<Reservation> query = new HighLevelQuery<>();

		query.addParameter(reservationIDParamFactory.getParameter(reservationID));

		return get(query);
	}

	public List<Reservation> getByReservationIDs(List<String> reservationIDs, TransactionHandler transactionHandler) throws SQLException {

		HighLevelQuery<Reservation> query = new HighLevelQuery<>();

		query.addParameter(reservationIDParamFactory.getWhereInParameter(reservationIDs));

		return getAll(query, transactionHandler);
	}

	public Reservation getByObjectIDAndFlowFamilyID(String objectID, Integer flowFamilyID) throws SQLException {

		HighLevelQuery<Reservation> query = new HighLevelQuery<>();

		query.addParameter(objectIDParamFactory.getParameter(objectID));
		query.addParameter(flowFamilyIDParamFactory.getParameter(flowFamilyID));

		return get(query);
	}

	public List<Reservation> getByObjectIDsAndFlowFamilyID(List<String> objectIDs, Integer flowFamilyID, List<String> excludedReservationIDs) throws SQLException {

		HighLevelQuery<Reservation> query = new HighLevelQuery<>();

		query.addParameter(objectIDParamFactory.getWhereInParameter(objectIDs));
		query.addParameter(flowFamilyIDParamFactory.getParameter(flowFamilyID));

		if (excludedReservationIDs != null) {

			query.addParameter(reservationIDParamFactory.getWhereNotInParameter(excludedReservationIDs));
		}

		return getAll(query);
	}

	public Collection<Integer> getFlowFamilyIDsByFlowInstanceManagerID(String flowInstanceManagerID) throws SQLException {

		HighLevelQuery<Reservation> query = new HighLevelQuery<>();

		query.addParameter(flowInstanceManagerIDParamFactory.getParameter(flowInstanceManagerID));

		return getFlowFamilyIDs(getAll(query));
	}

	public List<Reservation> getByFlowInstanceManagerIDAndFlowFamilyID(String flowInstanceManagerID, Integer flowFamilyID, TransactionHandler transactionHandler) throws SQLException {

		HighLevelQuery<Reservation> query = new HighLevelQuery<>();

		query.addParameter(flowInstanceManagerIDParamFactory.getParameter(flowInstanceManagerID));
		query.addParameter(flowFamilyIDParamFactory.getParameter(flowFamilyID));

		return getAll(query, transactionHandler);
	}

	public Collection<Integer> getExpiredFlowFamilyIDs() throws SQLException {

		HighLevelQuery<Reservation> query = new HighLevelQuery<>();

		Timestamp now = TimeUtils.getCurrentTimestamp();

		query.addParameter(expiresParamFactory.getParameter(now, QueryOperators.SMALLER_THAN));

		return getFlowFamilyIDs(getAll(query));
	}

	public List<Reservation> getExpiredByFlowFamilyID(Integer flowFamilyID) throws SQLException {

		HighLevelQuery<Reservation> query = new HighLevelQuery<>();

		Timestamp now = TimeUtils.getCurrentTimestamp();

		query.addParameter(expiresParamFactory.getParameter(now, QueryOperators.SMALLER_THAN));
		query.addParameter(flowFamilyIDParamFactory.getParameter(flowFamilyID));

		return getAll(query);
	}

	public int deleteByReservationID(String reservationID) throws SQLException {

		HighLevelQuery<Reservation> query = new HighLevelQuery<>();

		query.addParameter(reservationIDParamFactory.getParameter(reservationID));

		return delete(query);
	}

	public void deleteByFlowInstanceID(FlowInstance flowInstance) throws SQLException {

		HighLevelQuery<Reservation> query = new HighLevelQuery<>();

		query.addParameter(flowInstanceIDParamFactory.getParameter(flowInstance.getFlowInstanceID()));

		delete(query, 10);
	}

	private Collection<Integer> getFlowFamilyIDs(List<Reservation> reservations) {

		if (reservations == null) {

			return null;
		}

		Set<Integer> flowFamilyIDs = new HashSet<>();

		reservations.forEach(reservation -> flowFamilyIDs.add(reservation.getFlowFamilyID()));

		return flowFamilyIDs;
	}

}
