package com.nordicpeak.flowengine.evaluators.userage;

import java.sql.SQLException;
import java.util.Calendar;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.FlowInstanceAdminModule;
import com.nordicpeak.flowengine.beans.EvaluationResponse;
import com.nordicpeak.flowengine.evaluators.basequerystateevaluator.BaseQueryStateEvaluationProviderModule;
import com.nordicpeak.flowengine.interfaces.EvaluationCallback;
import com.nordicpeak.flowengine.interfaces.Evaluator;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.MutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.Query;
import com.nordicpeak.flowengine.interfaces.QueryInstance;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;


public class UserAgeQueryStateEvaluationProviderModule extends BaseQueryStateEvaluationProviderModule<UserAgeQueryStateEvaluator> {

	protected AnnotatedDAO<UserAgeQueryStateEvaluator> evaluatorDAO;
	
	protected QueryParameterFactory<UserAgeQueryStateEvaluator, Integer> evaluatorIDParamFactory;
	
	protected UserAgeQueryStateEvaluatorCRUD evaluatorCRUD;
	
	@InstanceManagerDependency(required = true)
	protected FlowInstanceAdminModule flowInstanceAdminModule;
	
	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {
		
		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, UserAgeQueryStateEvaluationProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));
		
		if (upgradeResult.isUpgrade()) {
			
			log.info(upgradeResult.toString());
		}
		
		evaluatorDAO = new SimpleAnnotatedDAOFactory(dataSource).getDAO(UserAgeQueryStateEvaluator.class);
		evaluatorIDParamFactory = evaluatorDAO.getParamFactory("evaluatorID", Integer.class);
		
		evaluatorCRUD = new UserAgeQueryStateEvaluatorCRUD(UserAgeQueryStateEvaluator.class, evaluatorDAO.getWrapper("evaluatorID", Integer.class), this);
	}
	
	@WebPublic(alias = "config")
	public ForegroundModuleResponse configureEvaluator(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		return evaluatorCRUD.update(req, res, user, uriParser);
	}	
	
	@Override
	public EvaluationResponse evaluate(QueryInstance queryInstance, User user, User poster, UserAgeQueryStateEvaluator evaluator, EvaluationCallback callback, boolean hasQueryValidationErrors, MutableAttributeHandler attributeHandler) {

		if(evaluator.getMinAge() == null && evaluator.getMaxAge() == null){
			
			return null;
		}
		
		String citizenIdentifier = getCitizenIdentifier(poster);
		
		if(citizenIdentifier == null){
			
			return restoreDefaultQueryStates(queryInstance, evaluator, callback, attributeHandler);
		}
		
		Integer age = getAge(citizenIdentifier, evaluator.isCalculateBasedOnYearOnly());
		
		if(age == null){
			
			return restoreDefaultQueryStates(queryInstance, evaluator, callback, attributeHandler);
		}
		
		if((evaluator.getMinAge() == null || age >= evaluator.getMinAge()) && (evaluator.getMaxAge() == null || age <= evaluator.getMaxAge())){
			
			if (evaluator.isSetAttributeIfAgeInRange()) {
				
				attributeHandler.setAttribute(evaluator.getAttributeName(), evaluator.getAttributeValue());
			}
			
			return applyEvaluatorQueryStates(queryInstance, evaluator, callback, attributeHandler);
		
		}else{
			
			return restoreDefaultQueryStates(queryInstance, evaluator, callback, attributeHandler);
		}
	}

	private Integer getAge(String citizenIdentifier, boolean calculateBasedOnYearOnly) {

		if(calculateBasedOnYearOnly){
			
			if(citizenIdentifier.length() < 4){
				
				return null;
			}
			
			Integer birthYear = NumberUtils.toInt(citizenIdentifier.substring(0, 4));
			
			if(birthYear == null){
				
				return null;
			}
			
			Integer currentYear = Calendar.getInstance().get(Calendar.YEAR);
			
			return currentYear - birthYear;
			
		}else{
			
			if(citizenIdentifier.length() < 8){
				
				return null;
			}
			
			Integer year = NumberUtils.toInt(citizenIdentifier.substring(0, 4));
			Integer month = NumberUtils.toInt(citizenIdentifier.substring(4, 6));
			Integer day = NumberUtils.toInt(citizenIdentifier.substring(6, 8));
			
			if(year == null || month == null || day == null){
				
				return null;
			}
			
			Calendar calendar = Calendar.getInstance();
			calendar.set(Calendar.YEAR, year);
			calendar.set(Calendar.MONTH, month-1); //month is 0-based
			calendar.set(Calendar.DAY_OF_MONTH, day);
			
			return DateUtils.getYearsBetween(calendar, Calendar.getInstance());
		}
	}

	private String getCitizenIdentifier(User poster) {

		return CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(poster);
	}

	@Override
	public Evaluator createEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		UserAgeQueryStateEvaluator evaluator = new UserAgeQueryStateEvaluator();
		
		evaluator.setEvaluatorID(descriptor.getEvaluatorID());
		
		this.evaluatorDAO.add(evaluator, transactionHandler, null);
		
		evaluator.init(descriptor, getFullAlias() + "/config/" + descriptor.getEvaluatorID());
		
		return evaluator;
	}

	@Override
	public boolean supportsQueryType(Class<? extends Query> queryClass) {

		return true;
	}

	@Override
	public void copyEvaluator(MutableEvaluatorDescriptor sourceEvaluatorDescriptor, MutableEvaluatorDescriptor copyEvaluatorDescriptor, Query sourceQuery, Query copyQuery, Map<Integer, ImmutableStatus> statusConversionMap, TransactionHandler transactionHandler) throws SQLException {

		UserAgeQueryStateEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<UserAgeQueryStateEvaluator>(evaluatorIDParamFactory.getParameter(sourceEvaluatorDescriptor.getEvaluatorID())));
		
		evaluator.setEvaluatorID(copyEvaluatorDescriptor.getEvaluatorID());
		
		this.evaluatorDAO.add(evaluator, transactionHandler, null);
	}

	@Override
	protected Class<UserAgeQueryStateEvaluator> getEvaluatorClass() {

		return UserAgeQueryStateEvaluator.class;
	}

	@Override
	protected AnnotatedDAO<UserAgeQueryStateEvaluator> getEvaluatorDAO() {

		return evaluatorDAO;
	}

	@Override
	protected QueryParameterFactory<UserAgeQueryStateEvaluator, Integer> getEvaluatorIDParameterFactory() {

		return this.evaluatorIDParamFactory;
	}

	@Override
	protected List<Integer> getTargetQueryIDs(UserAgeQueryStateEvaluator evaluator, EvaluationCallback callback) {

		return Collections.singletonList(callback.getQueryInstance().getQueryInstanceDescriptor().getQueryDescriptor().getQueryID());
	}

	@Override
	public UserAgeQueryStateEvaluator importEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler, Query query, Map<Integer, ImmutableStatus> statusConversionMap) throws Throwable {

		UserAgeQueryStateEvaluator evaluator = importEvaluatorDescriptor(descriptor, transactionHandler, query, statusConversionMap);
		
		evaluatorDAO.add(evaluator, transactionHandler, null);
		
		return evaluator;
	}
}
