package com.nordicpeak.flowengine.usersubscriptions;

import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map.Entry;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.HTMLEditorSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.AccessInterface;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.utils.extensionlinks.ExtensionLink;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.standardutils.datatypes.SimpleEntry;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.CharsetValidator;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPResponse;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.UserFlowInstanceMenuModule;
import com.nordicpeak.flowengine.interfaces.UserMenuProvider;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;

public class UserSubscriptionModule extends AnnotatedRESTModule implements UserMenuProvider {

	@XSLVariable(prefix = "java.")
	private String userMenuTabTitle = "My subscriptions";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Title", description = "The title displayed above the subscriptions list", required = true)
	private String title = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Description", description = "The descripton dsiplayed above the subscriptions list", required = true)
	private String description = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Help text", description = "The help text for the subscriptions list", required = true)
	private String helpText = "Not set";

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "Subscription list API URL", description = "URL to the subscription list API, $citizenIdentifier will be replaced with the users citizen identifier.")
	private String subscriptionsURL;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "API username", description = "")
	private String apiUsername;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "API password", description = "")
	private String apiPassword;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Encoding", description = "Encoding to use when parsing the API response", required = true, formatValidator = CharsetValidator.class)
	private String encoding = "UTF-8";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "The maximum time in seconds to wait for a connection", required = true)
	private Integer connectionTimeout = 15;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Read timeout", description = "The maximum time in seconds to wait for a response", required = true)
	private Integer readTimeout = 15;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "User menu item slot", description = "User menu item slot")
	protected String userMenuExtensionLinkSlot = "55";

	protected UserFlowInstanceMenuModule userFlowInstanceMenuModule;

	protected ExtensionLink userMenuLink;

	@Override
	protected void moduleConfigured() throws Exception {

		super.moduleConfigured();

		this.userMenuLink = new ExtensionLink(userMenuTabTitle, getFullAlias(), ">", userMenuExtensionLinkSlot);

		if (userFlowInstanceMenuModule != null) {

			userFlowInstanceMenuModule.sortProviders();
		}
	}

	@InstanceManagerDependency
	public void setUserFlowInstanceMenuModule(UserFlowInstanceMenuModule userFlowInstanceMenuModule) {

		if (userFlowInstanceMenuModule == null && this.userFlowInstanceMenuModule != null) {

			this.userFlowInstanceMenuModule.removeUserMenuProvider(this);
		}

		this.userFlowInstanceMenuModule = userFlowInstanceMenuModule;

		if (this.userFlowInstanceMenuModule != null) {

			this.userFlowInstanceMenuModule.addUserMenuProvider(this);
		}
	}

	@Override
	public void unload() throws Exception {

		if (userFlowInstanceMenuModule != null) {

			setUserFlowInstanceMenuModule(null);
		}

		super.unload();
	}

	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		Document doc = createDocument(req, uriParser);

		Element userInvoiceListElement = XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "SubscriptionList");
		
		XMLUtils.append(doc, userInvoiceListElement, "Subscriptions", getSubscriptions(user));
		
		XMLUtils.appendNewElement(doc, userInvoiceListElement, "title", title);
		XMLUtils.appendNewElement(doc, userInvoiceListElement, "description", description);
		XMLUtils.appendNewElement(doc, userInvoiceListElement, "helpText", helpText);
		
		return new SimpleForegroundModuleResponse(doc, userMenuTabTitle); 
	}
	

	private List<Subscription> getSubscriptions(User user) {

		String citizenID = CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(user);

		if (subscriptionsURL != null && citizenID != null) {

			log.info("Getting subscriptions for user " + user + "...");

			SimpleRequest simpleRequest = new SimpleRequest(subscriptionsURL.replace("$citizenIdentifier", citizenID));

			simpleRequest.setFollowRedirects(false);
			simpleRequest.setConnectionTimeout(connectionTimeout * 1000);
			simpleRequest.setReadTimeout(readTimeout * 1000);

			if (apiUsername != null && apiPassword != null) {

				simpleRequest.setUsername(apiUsername);
				simpleRequest.setPassword(apiPassword);
			}

			List<Entry<String, String>> headerEntries = new ArrayList<Entry<String, String>>(2);

			headerEntries.add(new SimpleEntry<String, String>("NoLoginRedirect", "NoLoginRedirect"));
			simpleRequest.setHeaders(headerEntries);

			try {
				HTTPResponse response = HTTPUtils.sendHTTPGetRequest(simpleRequest, Charset.forName(encoding));

				try {
					XMLParser parser = new XMLParser(XMLUtils.parseXML(response.getValue(), false, false));

					List<XMLParser> items = parser.getNodes("/Subscriptions/Subscription");

					if (items.size() > 0) {

						List<Subscription> subscriptions = new ArrayList<Subscription>(items.size());

						for (XMLParser item : items) {

							String invoiceID = item.getString("SubscriptionID");
							String cost = item.getString("Cost");
							String shortDescription = item.getString("ShortDescription");
							String longDescription = item.getString("LongDescription");
							boolean active = item.getPrimitiveBoolean("Active");

							Date invoiceDate = null;
							Date dueDate = null;

							String startDate = item.getString("StartDate");
							String endDate = item.getString("EndDate");

							if (!StringUtils.isEmpty(startDate)) {
								invoiceDate = DateUtils.DATE_FORMATTER.parse(startDate);
							}

							if (!StringUtils.isEmpty(endDate)) {
								dueDate = DateUtils.DATE_FORMATTER.parse(endDate);
							}
							
							subscriptions.add(new Subscription(invoiceID, cost, invoiceDate, dueDate, active, shortDescription, longDescription));
						}

						log.info("Got " + subscriptions.size() + " subscriptions for user " + user);
						
						return subscriptions;
					}

					log.info("Got no subscriptions for user " + user);
					
				} catch (Exception e) {

					log.warn("Error parsing subscription API response for user " + user, e);
				}

			} catch (Throwable e) {

				log.warn("Error connecting to subscription API for user " + user, e);
			}
		}
		
		return null;
	}

	public Document createDocument(HttpServletRequest req, URIParser uriParser) {

		Document doc = XMLUtils.createDomDocument();
		Element documentElement = doc.createElement("Document");
		doc.appendChild(documentElement);

		XMLUtils.appendNewElement(doc, documentElement, "BaseURL", uriParser.getContextPath() + this.getFullAlias());
		
		return doc;
	}

	@Override
	public AccessInterface getAccessInterface() {

		return moduleDescriptor;
	}

	@Override
	public String getUserMenuAlias() {

		return getFullAlias();
	}

	@Override
	public ExtensionLink getUserMenuExtensionLink(User user) {

		return userMenuLink;
	}
	
	@Override
	public String getUserMenuPriority() {

		return userMenuExtensionLinkSlot;
	}
}
