package com.nordicpeak.wmsproxy;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.datatypes.SimpleEntry;
import se.unlogic.standardutils.streams.StreamUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.URIParser;

public class WMSProxyModule extends AnnotatedForegroundModule {

	public static final String ORTO_ALIAS = "orto";
	private static final Map<String, String> ALIASES_TO_LAYER_NAMES = new HashMap<>(1);

	static {
		ALIASES_TO_LAYER_NAMES.put(ORTO_ALIAS, "Ortofoto_0.25");
	}

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Log requests", description = "Controls if requests handled by the proxy should be logged")
	protected boolean logRequests = false;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "WMS Server username", description = "The username to use when autenticate to WMS server (if required)", required = false)
	protected String username;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "WMS Server password", description = "The password to use when autenticate to WMS server (if required)", required = false)
	protected String password;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "The connection timeout for proxy requests")
	protected Integer connectionTimeout = 5000;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Read timeout", description = "The read timeout for proxy requests")
	protected Integer readTimeout = 10000;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "WMS server URL parameter name", description = "The name of the parameter for WMS server alias", required = true)
	protected String wmsURLParam = "wms";

	@ModuleSetting
	@TextAreaSettingDescriptor(name = "WMS servers", description = "WMS servers proxied by this module specified in format [alias]=http://example.com/wms (without brackets)", required = true)
	private List<String> wmsServers;

	private Map<String, String> cachedWMSServers;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);

		systemInterface.getInstanceHandler().addInstance(WMSProxyModule.class, this);
	}

	@Override
	public void unload() throws Exception {

		systemInterface.getInstanceHandler().removeInstance(WMSProxyModule.class, this);

		super.unload();
	}

	@Override
	protected void moduleConfigured() throws Exception {

		if (wmsServers == null) {

			log.warn("No WMS servers specified in module settings.");

			return;
		}

		log.info("Caching WMS servers");

		cachedWMSServers = new HashMap<>();

		for (String wmsServer : wmsServers) {

			String[] parts = wmsServer.split("=");

			if (parts.length != 2) {

				log.warn("Invalid format of WMS server definition " + wmsServer + ", skipping");

				continue;
			}

			if (!HTTPUtils.isValidURL(parts[1])) {

				log.warn("Invalid HTTP url specified for WMS server definition " + wmsServer + ", skipping");

				continue;
			}

			cachedWMSServers.put(parts[0], parts[1]);
		}
	}

	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		String wmsAlias = req.getParameter(wmsURLParam);

		if (!StringUtils.isEmpty(wmsAlias) && cachedWMSServers.containsKey(wmsAlias)) {

			String endPoint = cachedWMSServers.get(wmsAlias);

			try {
				List<Entry<String, String>> forwardedParams = null;

				@SuppressWarnings("rawtypes")
				Map paramMap = req.getParameterMap();

				if (!paramMap.isEmpty()) {

					forwardedParams = new ArrayList<>();

					for (Object paramName : paramMap.keySet()) {

						String[] paramValues = (String[]) paramMap.get(paramName);

						for (String paramValue : paramValues) {

							forwardedParams.add(new SimpleEntry<>((String) paramName, paramValue));
						}
					}
				}

				if (req.getMethod().equalsIgnoreCase("GET")) {

					if (logRequests) {

						log.info("Proxying request " + req.getRequestURI() + " to " + endPoint);

					}

					proxyHTTPGetRequest(endPoint, forwardedParams, res, username, password);

					return null;

				}

			} catch (IOException e) {

				log.warn("Caught exception " + e + " while sending request " + endPoint);

			}

			return null;
		}

		throw new URINotFoundException(uriParser);

	}

	public static void proxyHTTPGetRequest(String endpoint, List<? extends Entry<String, String>> requestParameters, HttpServletResponse res, String username, String password) throws IOException {

		HttpURLConnection connection = null;

		try {
			connection = HTTPUtils.getHttpURLConnection(endpoint, requestParameters, false);

			HTTPUtils.setBasicAuthentication(connection, username, password);

			res.setContentType(connection.getContentType());
			res.setContentLength(connection.getContentLength());
			res.setStatus(connection.getResponseCode());

			res.flushBuffer();

			StreamUtils.transfer(connection.getInputStream(), res.getOutputStream());

		} finally {

			if (connection != null) {
				connection.disconnect();
			}
		}
	}

	public String getWMSServerByAlias(String alias) {

		if (cachedWMSServers.containsKey(alias)) {

			return cachedWMSServers.get(alias);
		}

		for (Entry<String, String> entry : cachedWMSServers.entrySet()) {

			if (alias.startsWith(entry.getKey())) {

				return entry.getValue();
			}
		}

		return null;
	}

	public static String getLayerName(String alias) {

		return ALIASES_TO_LAYER_NAMES.getOrDefault(alias, alias);
	}
}
