package com.nordicpeak.flowengine.evaluators.weightapi;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.utils.crud.IntegerBeanIDParser;
import se.unlogic.hierarchy.core.utils.crud.ModularCRUD;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationErrorType;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.BeanRequestPopulator;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.evaluators.weightapi.beans.WeightAPISource;
import com.nordicpeak.flowengine.evaluators.weightapi.exceptions.UnableToDeleteWeightAPISourceInUseException;

public class WeightAPISourceCRUD extends ModularCRUD<WeightAPISource, Integer, User, WeightAPIAdminModule> {

	public WeightAPISourceCRUD(AnnotatedDAOWrapper<WeightAPISource, Integer> queryDAO, BeanRequestPopulator<WeightAPISource> populator, WeightAPIAdminModule callback) {

		super(IntegerBeanIDParser.getInstance(), queryDAO, populator, "APISource", "API source", "/", callback);

		setRequirePostForDelete(true);
	}

	@Override
	protected void appendShowFormData(WeightAPISource source, Document doc, Element showTypeElement, User user, HttpServletRequest req, HttpServletResponse res, URIParser uriParser) throws SQLException, IOException, Exception {

		super.appendShowFormData(source, doc, showTypeElement, user, req, res, uriParser);

		XMLUtils.append(doc, showTypeElement, "Events", callback.getRecentListEvents(source));
	}

	@Override
	protected void appendAddFormData(Document doc, Element addTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		appendFormData(null, doc, addTypeElement, user, req, uriParser);
		
		XMLUtils.append(doc, addTypeElement, "FlowFamilyList", callback.getFlowFamilyListConnector().getFlowFamilyListItems(NumberUtils.toInt(req.getParameterValues("flowfamilyID"))));
	}

	@Override
	protected void appendUpdateFormData(WeightAPISource bean, Document doc, Element updateTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		appendFormData(bean, doc, updateTypeElement, user, req, uriParser);
		
		XMLUtils.append(doc, updateTypeElement, "FlowFamilyList", callback.getFlowFamilyListConnector().getFlowFamilyListItems(bean.getFlowFamilies()));
	}

	protected void appendFormData(WeightAPISource bean, Document doc, Element updateTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		XMLUtils.append(doc, updateTypeElement, "AllowedEncodings", "Encoding", callback.getAllowedEncodings());
	}

	@Override
	protected void validateAddPopulation(WeightAPISource source, HttpServletRequest req, User user, URIParser uriParser) throws ValidationException, SQLException, Exception {

		validate(source, uriParser);
	}

	@Override
	protected void validateUpdatePopulation(WeightAPISource source, HttpServletRequest req, User user, URIParser uriParser) throws ValidationException, SQLException, Exception {

		validate(source, uriParser);
		
		if(source.isUseAccessRestriction()) {
			
			List<ValidationError> missingFlowFamilies = callback.checkMissingFlowFamilies(source.getSourceID(), source.getFlowFamilies());
			
			if(missingFlowFamilies != null) {
				
				throw new ValidationException(missingFlowFamilies);
			}
		}
	}

	protected void validate(WeightAPISource source, URIParser uriParser) throws ValidationException, Exception {

		List<ValidationError> validationErrors = new ArrayList<>();

		if (source.getEncoding() != null) {

			if (!callback.getAllowedEncodings().contains(source.getEncoding())) {
				throw new ValidationException(new ValidationError("encoding", ValidationErrorType.InvalidFormat));
			}
		}

		WeightAPISource existingSource = callback.getSource(source.getName());

		if (existingSource != null) {

			if (!existingSource.getSourceID().equals(source.getSourceID())) {

				validationErrors.add(new ValidationError("APISourceNameAlreadyInUse"));
			}
		}
		
		if ((source.getName() == null || source.getName().isEmpty())) {
			validationErrors.add(new ValidationError("name", ValidationErrorType.RequiredField));
		}

		if ((source.getUrl() == null || source.getUrl().isEmpty())) {
			validationErrors.add(new ValidationError("url", ValidationErrorType.RequiredField));
		}

		if ((source.getUsername() == null || source.getUsername().isEmpty()) && (source.getPassword() != null && !source.getPassword().isEmpty())) {
			validationErrors.add(new ValidationError("username", ValidationErrorType.RequiredField));

		}

		if ((source.getPassword() == null || source.getPassword().isEmpty()) && (source.getUsername() != null && !source.getUsername().isEmpty())) {
			validationErrors.add(new ValidationError("password", ValidationErrorType.RequiredField));

		}

		if (!validationErrors.isEmpty()) {
			throw new ValidationException(validationErrors);
		}
	}

	@Override
	protected void checkDeleteAccess(WeightAPISource source, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		super.checkDeleteAccess(source, user, req, uriParser);

		if (!CollectionUtils.isEmpty(callback.getEvaluatorsUsingAPISource(source))) {

			throw new UnableToDeleteWeightAPISourceInUseException();
		}
	}

	@Override
	protected ForegroundModuleResponse filteredBeanAdded(WeightAPISource source, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		callback.addListEvent(callback.getEventSourceAddedMessage(), source, user);

		return super.filteredBeanAdded(source, req, res, user, uriParser);
	}

	@Override
	protected ForegroundModuleResponse filteredBeanUpdated(WeightAPISource source, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		callback.addListEvent(callback.getEventSourceUpdatedMessage(), source, user);

		return super.filteredBeanUpdated(source, req, res, user, uriParser);
	}
	
	@Override
	public ForegroundModuleResponse list(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, List<ValidationError> validationErrors) throws Exception {

		if(callback.getAPISourceHandler() != null) {
			
			res.sendRedirect(uriParser.getContextPath() + callback.getAPISourceHandler().getFullAlias());
			return null;
			
		}else {
			
			return super.list(req, res, user, uriParser, validationErrors);
		}
	}
}
